/* Automatically @generated by tree-sitter */

#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#define LANGUAGE_VERSION 15
#define STATE_COUNT 566
#define LARGE_STATE_COUNT 2
#define SYMBOL_COUNT 163
#define ALIAS_COUNT 2
#define TOKEN_COUNT 89
#define EXTERNAL_TOKEN_COUNT 5
#define FIELD_COUNT 19
#define MAX_ALIAS_SEQUENCE_LENGTH 9
#define MAX_RESERVED_WORD_SET_SIZE 0
#define PRODUCTION_ID_COUNT 59
#define SUPERTYPE_COUNT 0

enum ts_symbol_identifiers {
  sym_identifier = 1,
  anon_sym_LPAREN = 2,
  anon_sym_RPAREN = 3,
  anon_sym_COLON_EQ = 4,
  anon_sym_COMMA = 5,
  anon_sym_alias = 6,
  anon_sym_COLON_COLON = 7,
  anon_sym_unexport = 8,
  anon_sym_eager = 9,
  anon_sym_export = 10,
  anon_sym_import = 11,
  anon_sym_QMARK = 12,
  anon_sym_mod = 13,
  anon_sym_set = 14,
  anon_sym_true = 15,
  anon_sym_false = 16,
  anon_sym_SLASH = 17,
  anon_sym_PLUS_PLUS = 18,
  anon_sym_PLUS = 19,
  anon_sym_EQ_EQ = 20,
  anon_sym_BANG_EQ = 21,
  anon_sym_EQ_TILDE = 22,
  anon_sym_BANG_TILDE = 23,
  anon_sym_AMP_AMP = 24,
  anon_sym_PIPE_PIPE = 25,
  anon_sym_if = 26,
  anon_sym_else = 27,
  anon_sym_LBRACE = 28,
  anon_sym_RBRACE = 29,
  anon_sym_BANG = 30,
  anon_sym_LBRACK = 31,
  anon_sym_RBRACK = 32,
  anon_sym_assert = 33,
  anon_sym_COLON = 34,
  anon_sym_EQ = 35,
  anon_sym_AT = 36,
  anon_sym_DOLLAR = 37,
  anon_sym_STAR = 38,
  anon_sym_AT_DASH_QMARK = 39,
  anon_sym_AT_QMARK_DASH = 40,
  anon_sym_DASH_AT_QMARK = 41,
  anon_sym_DASH_QMARK_AT = 42,
  anon_sym_QMARK_AT_DASH = 43,
  anon_sym_QMARK_DASH_AT = 44,
  anon_sym_AT_DASH = 45,
  anon_sym_AT_QMARK = 46,
  anon_sym_DASH_AT = 47,
  anon_sym_DASH_QMARK = 48,
  anon_sym_QMARK_AT = 49,
  anon_sym_QMARK_DASH = 50,
  anon_sym_DASH = 51,
  aux_sym_shebang_token1 = 52,
  aux_sym__shebang_with_lang_token1 = 53,
  anon_sym_env = 54,
  aux_sym__shebang_with_lang_token2 = 55,
  aux_sym__shebang_with_lang_token3 = 56,
  sym__opaque_shebang = 57,
  aux_sym_string_token1 = 58,
  aux_sym_string_token2 = 59,
  anon_sym_f_DQUOTE = 60,
  aux_sym__format_string_token1 = 61,
  aux_sym__format_string_token2 = 62,
  anon_sym_DQUOTE = 63,
  anon_sym_f_DQUOTE_DQUOTE_DQUOTE = 64,
  anon_sym_DQUOTE_DQUOTE_DQUOTE = 65,
  anon_sym_f_SQUOTE = 66,
  aux_sym__format_raw_string_token1 = 67,
  anon_sym_SQUOTE = 68,
  anon_sym_f_SQUOTE_SQUOTE_SQUOTE = 69,
  anon_sym_SQUOTE_SQUOTE_SQUOTE = 70,
  aux_sym__raw_string_indented_token1 = 71,
  aux_sym__string_token1 = 72,
  anon_sym_x_DQUOTE = 73,
  anon_sym_x_DQUOTE_DQUOTE_DQUOTE = 74,
  aux_sym__shell_expanded_string_indented_token1 = 75,
  anon_sym_x_SQUOTE_SQUOTE_SQUOTE = 76,
  sym_escape_sequence = 77,
  anon_sym_BQUOTE = 78,
  anon_sym_BQUOTE_BQUOTE_BQUOTE = 79,
  anon_sym_LBRACE_LBRACE = 80,
  anon_sym_RBRACE_RBRACE = 81,
  sym_numeric_error = 82,
  sym_comment = 83,
  sym__indent = 84,
  sym__dedent = 85,
  sym__newline = 86,
  sym_text = 87,
  sym_error_recovery = 88,
  sym_source_file = 89,
  sym__item = 90,
  sym_function_definition = 91,
  sym_function_parameters = 92,
  sym_alias = 93,
  sym_module_path = 94,
  sym__name = 95,
  sym_assignment = 96,
  sym__assignment = 97,
  sym_eager = 98,
  sym_export = 99,
  sym_unexport = 100,
  sym_import = 101,
  sym_module = 102,
  sym_setting = 103,
  sym_boolean = 104,
  sym_expression = 105,
  sym__expression_inner = 106,
  sym_if_expression = 107,
  sym_else_if_clause = 108,
  sym_else_clause = 109,
  sym__braced_expr = 110,
  sym_condition = 111,
  sym_regex_literal = 112,
  sym_value = 113,
  sym_not_expression = 114,
  sym_list_literal = 115,
  sym_list_elements = 116,
  sym_assert_expression = 117,
  sym_function_call = 118,
  sym_external_command = 119,
  sym_sequence = 120,
  sym_attribute = 121,
  sym_attribute_named_param = 122,
  sym_recipe = 123,
  sym_recipe_header = 124,
  sym_parameters = 125,
  sym_parameter = 126,
  sym_variadic_parameter = 127,
  sym_dependencies = 128,
  sym_dependency = 129,
  sym_dependency_expression = 130,
  sym_starred_dependency_argument = 131,
  sym_recipe_body = 132,
  sym_recipe_line = 133,
  sym_recipe_line_prefix = 134,
  sym_shebang = 135,
  sym__shebang_with_lang = 136,
  sym_string = 137,
  sym_format_string = 138,
  sym__backticked = 139,
  sym__indented_backticked = 140,
  sym_command_body = 141,
  sym_interpolation = 142,
  aux_sym_source_file_repeat1 = 143,
  aux_sym_function_definition_repeat1 = 144,
  aux_sym_function_parameters_repeat1 = 145,
  aux_sym_module_path_repeat1 = 146,
  aux_sym_if_expression_repeat1 = 147,
  aux_sym_list_elements_repeat1 = 148,
  aux_sym_attribute_repeat1 = 149,
  aux_sym_attribute_repeat2 = 150,
  aux_sym_parameters_repeat1 = 151,
  aux_sym_dependencies_repeat1 = 152,
  aux_sym_dependency_expression_repeat1 = 153,
  aux_sym_recipe_body_repeat1 = 154,
  aux_sym_recipe_line_repeat1 = 155,
  aux_sym__shebang_with_lang_repeat1 = 156,
  aux_sym__format_string_repeat1 = 157,
  aux_sym__format_raw_string_repeat1 = 158,
  aux_sym__raw_string_indented_repeat1 = 159,
  aux_sym__string_repeat1 = 160,
  aux_sym__shell_expanded_string_indented_repeat1 = 161,
  aux_sym_command_body_repeat1 = 162,
  anon_alias_sym_expression = 163,
  alias_sym_language = 164,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [sym_identifier] = "identifier",
  [anon_sym_LPAREN] = "(",
  [anon_sym_RPAREN] = ")",
  [anon_sym_COLON_EQ] = ":=",
  [anon_sym_COMMA] = ",",
  [anon_sym_alias] = "alias",
  [anon_sym_COLON_COLON] = "::",
  [anon_sym_unexport] = "unexport",
  [anon_sym_eager] = "eager",
  [anon_sym_export] = "export",
  [anon_sym_import] = "import",
  [anon_sym_QMARK] = "\?",
  [anon_sym_mod] = "mod",
  [anon_sym_set] = "set",
  [anon_sym_true] = "true",
  [anon_sym_false] = "false",
  [anon_sym_SLASH] = "/",
  [anon_sym_PLUS_PLUS] = "++",
  [anon_sym_PLUS] = "+",
  [anon_sym_EQ_EQ] = "==",
  [anon_sym_BANG_EQ] = "!=",
  [anon_sym_EQ_TILDE] = "=~",
  [anon_sym_BANG_TILDE] = "!~",
  [anon_sym_AMP_AMP] = "&&",
  [anon_sym_PIPE_PIPE] = "||",
  [anon_sym_if] = "if",
  [anon_sym_else] = "else",
  [anon_sym_LBRACE] = "{",
  [anon_sym_RBRACE] = "}",
  [anon_sym_BANG] = "!",
  [anon_sym_LBRACK] = "[",
  [anon_sym_RBRACK] = "]",
  [anon_sym_assert] = "identifier",
  [anon_sym_COLON] = ":",
  [anon_sym_EQ] = "=",
  [anon_sym_AT] = "@",
  [anon_sym_DOLLAR] = "$",
  [anon_sym_STAR] = "*",
  [anon_sym_AT_DASH_QMARK] = "@-\?",
  [anon_sym_AT_QMARK_DASH] = "@\?-",
  [anon_sym_DASH_AT_QMARK] = "-@\?",
  [anon_sym_DASH_QMARK_AT] = "-\?@",
  [anon_sym_QMARK_AT_DASH] = "\?@-",
  [anon_sym_QMARK_DASH_AT] = "\?-@",
  [anon_sym_AT_DASH] = "@-",
  [anon_sym_AT_QMARK] = "@\?",
  [anon_sym_DASH_AT] = "-@",
  [anon_sym_DASH_QMARK] = "-\?",
  [anon_sym_QMARK_AT] = "\?@",
  [anon_sym_QMARK_DASH] = "\?-",
  [anon_sym_DASH] = "-",
  [aux_sym_shebang_token1] = "shebang_token1",
  [aux_sym__shebang_with_lang_token1] = "_shebang_with_lang_token1",
  [anon_sym_env] = "env",
  [aux_sym__shebang_with_lang_token2] = "_shebang_with_lang_token2",
  [aux_sym__shebang_with_lang_token3] = "_shebang_with_lang_token3",
  [sym__opaque_shebang] = "_opaque_shebang",
  [aux_sym_string_token1] = "string_token1",
  [aux_sym_string_token2] = "string_token2",
  [anon_sym_f_DQUOTE] = "f\"",
  [aux_sym__format_string_token1] = "_format_string_token1",
  [aux_sym__format_string_token2] = "_format_string_token2",
  [anon_sym_DQUOTE] = "\"",
  [anon_sym_f_DQUOTE_DQUOTE_DQUOTE] = "f\"\"\"",
  [anon_sym_DQUOTE_DQUOTE_DQUOTE] = "\"\"\"",
  [anon_sym_f_SQUOTE] = "f'",
  [aux_sym__format_raw_string_token1] = "_format_raw_string_token1",
  [anon_sym_SQUOTE] = "'",
  [anon_sym_f_SQUOTE_SQUOTE_SQUOTE] = "f'''",
  [anon_sym_SQUOTE_SQUOTE_SQUOTE] = "'''",
  [aux_sym__raw_string_indented_token1] = "_raw_string_indented_token1",
  [aux_sym__string_token1] = "_string_token1",
  [anon_sym_x_DQUOTE] = "x\"",
  [anon_sym_x_DQUOTE_DQUOTE_DQUOTE] = "x\"\"\"",
  [aux_sym__shell_expanded_string_indented_token1] = "_shell_expanded_string_indented_token1",
  [anon_sym_x_SQUOTE_SQUOTE_SQUOTE] = "x'''",
  [sym_escape_sequence] = "escape_sequence",
  [anon_sym_BQUOTE] = "`",
  [anon_sym_BQUOTE_BQUOTE_BQUOTE] = "```",
  [anon_sym_LBRACE_LBRACE] = "{{",
  [anon_sym_RBRACE_RBRACE] = "}}",
  [sym_numeric_error] = "numeric_error",
  [sym_comment] = "comment",
  [sym__indent] = "_indent",
  [sym__dedent] = "_dedent",
  [sym__newline] = "_newline",
  [sym_text] = "text",
  [sym_error_recovery] = "error_recovery",
  [sym_source_file] = "source_file",
  [sym__item] = "_item",
  [sym_function_definition] = "function_definition",
  [sym_function_parameters] = "function_parameters",
  [sym_alias] = "alias",
  [sym_module_path] = "module_path",
  [sym__name] = "_name",
  [sym_assignment] = "assignment",
  [sym__assignment] = "_assignment",
  [sym_eager] = "eager",
  [sym_export] = "export",
  [sym_unexport] = "unexport",
  [sym_import] = "import",
  [sym_module] = "module",
  [sym_setting] = "setting",
  [sym_boolean] = "boolean",
  [sym_expression] = "expression",
  [sym__expression_inner] = "_expression_inner",
  [sym_if_expression] = "if_expression",
  [sym_else_if_clause] = "else_if_clause",
  [sym_else_clause] = "else_clause",
  [sym__braced_expr] = "_braced_expr",
  [sym_condition] = "condition",
  [sym_regex_literal] = "regex_literal",
  [sym_value] = "value",
  [sym_not_expression] = "not_expression",
  [sym_list_literal] = "list_literal",
  [sym_list_elements] = "list_elements",
  [sym_assert_expression] = "assert_expression",
  [sym_function_call] = "function_call",
  [sym_external_command] = "external_command",
  [sym_sequence] = "sequence",
  [sym_attribute] = "attribute",
  [sym_attribute_named_param] = "attribute_named_param",
  [sym_recipe] = "recipe",
  [sym_recipe_header] = "recipe_header",
  [sym_parameters] = "parameters",
  [sym_parameter] = "parameter",
  [sym_variadic_parameter] = "variadic_parameter",
  [sym_dependencies] = "dependencies",
  [sym_dependency] = "dependency",
  [sym_dependency_expression] = "dependency_expression",
  [sym_starred_dependency_argument] = "starred_dependency_argument",
  [sym_recipe_body] = "recipe_body",
  [sym_recipe_line] = "recipe_line",
  [sym_recipe_line_prefix] = "recipe_line_prefix",
  [sym_shebang] = "shebang",
  [sym__shebang_with_lang] = "_shebang_with_lang",
  [sym_string] = "string",
  [sym_format_string] = "format_string",
  [sym__backticked] = "_backticked",
  [sym__indented_backticked] = "_indented_backticked",
  [sym_command_body] = "command_body",
  [sym_interpolation] = "interpolation",
  [aux_sym_source_file_repeat1] = "source_file_repeat1",
  [aux_sym_function_definition_repeat1] = "function_definition_repeat1",
  [aux_sym_function_parameters_repeat1] = "function_parameters_repeat1",
  [aux_sym_module_path_repeat1] = "module_path_repeat1",
  [aux_sym_if_expression_repeat1] = "if_expression_repeat1",
  [aux_sym_list_elements_repeat1] = "list_elements_repeat1",
  [aux_sym_attribute_repeat1] = "attribute_repeat1",
  [aux_sym_attribute_repeat2] = "attribute_repeat2",
  [aux_sym_parameters_repeat1] = "parameters_repeat1",
  [aux_sym_dependencies_repeat1] = "dependencies_repeat1",
  [aux_sym_dependency_expression_repeat1] = "dependency_expression_repeat1",
  [aux_sym_recipe_body_repeat1] = "recipe_body_repeat1",
  [aux_sym_recipe_line_repeat1] = "recipe_line_repeat1",
  [aux_sym__shebang_with_lang_repeat1] = "_shebang_with_lang_repeat1",
  [aux_sym__format_string_repeat1] = "_format_string_repeat1",
  [aux_sym__format_raw_string_repeat1] = "_format_raw_string_repeat1",
  [aux_sym__raw_string_indented_repeat1] = "_raw_string_indented_repeat1",
  [aux_sym__string_repeat1] = "_string_repeat1",
  [aux_sym__shell_expanded_string_indented_repeat1] = "_shell_expanded_string_indented_repeat1",
  [aux_sym_command_body_repeat1] = "command_body_repeat1",
  [anon_alias_sym_expression] = "expression",
  [alias_sym_language] = "language",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [sym_identifier] = sym_identifier,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [anon_sym_COLON_EQ] = anon_sym_COLON_EQ,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [anon_sym_alias] = anon_sym_alias,
  [anon_sym_COLON_COLON] = anon_sym_COLON_COLON,
  [anon_sym_unexport] = anon_sym_unexport,
  [anon_sym_eager] = anon_sym_eager,
  [anon_sym_export] = anon_sym_export,
  [anon_sym_import] = anon_sym_import,
  [anon_sym_QMARK] = anon_sym_QMARK,
  [anon_sym_mod] = anon_sym_mod,
  [anon_sym_set] = anon_sym_set,
  [anon_sym_true] = anon_sym_true,
  [anon_sym_false] = anon_sym_false,
  [anon_sym_SLASH] = anon_sym_SLASH,
  [anon_sym_PLUS_PLUS] = anon_sym_PLUS_PLUS,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_EQ_EQ] = anon_sym_EQ_EQ,
  [anon_sym_BANG_EQ] = anon_sym_BANG_EQ,
  [anon_sym_EQ_TILDE] = anon_sym_EQ_TILDE,
  [anon_sym_BANG_TILDE] = anon_sym_BANG_TILDE,
  [anon_sym_AMP_AMP] = anon_sym_AMP_AMP,
  [anon_sym_PIPE_PIPE] = anon_sym_PIPE_PIPE,
  [anon_sym_if] = anon_sym_if,
  [anon_sym_else] = anon_sym_else,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [anon_sym_BANG] = anon_sym_BANG,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [anon_sym_assert] = sym_identifier,
  [anon_sym_COLON] = anon_sym_COLON,
  [anon_sym_EQ] = anon_sym_EQ,
  [anon_sym_AT] = anon_sym_AT,
  [anon_sym_DOLLAR] = anon_sym_DOLLAR,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_AT_DASH_QMARK] = anon_sym_AT_DASH_QMARK,
  [anon_sym_AT_QMARK_DASH] = anon_sym_AT_QMARK_DASH,
  [anon_sym_DASH_AT_QMARK] = anon_sym_DASH_AT_QMARK,
  [anon_sym_DASH_QMARK_AT] = anon_sym_DASH_QMARK_AT,
  [anon_sym_QMARK_AT_DASH] = anon_sym_QMARK_AT_DASH,
  [anon_sym_QMARK_DASH_AT] = anon_sym_QMARK_DASH_AT,
  [anon_sym_AT_DASH] = anon_sym_AT_DASH,
  [anon_sym_AT_QMARK] = anon_sym_AT_QMARK,
  [anon_sym_DASH_AT] = anon_sym_DASH_AT,
  [anon_sym_DASH_QMARK] = anon_sym_DASH_QMARK,
  [anon_sym_QMARK_AT] = anon_sym_QMARK_AT,
  [anon_sym_QMARK_DASH] = anon_sym_QMARK_DASH,
  [anon_sym_DASH] = anon_sym_DASH,
  [aux_sym_shebang_token1] = aux_sym_shebang_token1,
  [aux_sym__shebang_with_lang_token1] = aux_sym__shebang_with_lang_token1,
  [anon_sym_env] = anon_sym_env,
  [aux_sym__shebang_with_lang_token2] = aux_sym__shebang_with_lang_token2,
  [aux_sym__shebang_with_lang_token3] = aux_sym__shebang_with_lang_token3,
  [sym__opaque_shebang] = sym__opaque_shebang,
  [aux_sym_string_token1] = aux_sym_string_token1,
  [aux_sym_string_token2] = aux_sym_string_token2,
  [anon_sym_f_DQUOTE] = anon_sym_f_DQUOTE,
  [aux_sym__format_string_token1] = aux_sym__format_string_token1,
  [aux_sym__format_string_token2] = aux_sym__format_string_token2,
  [anon_sym_DQUOTE] = anon_sym_DQUOTE,
  [anon_sym_f_DQUOTE_DQUOTE_DQUOTE] = anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
  [anon_sym_DQUOTE_DQUOTE_DQUOTE] = anon_sym_DQUOTE_DQUOTE_DQUOTE,
  [anon_sym_f_SQUOTE] = anon_sym_f_SQUOTE,
  [aux_sym__format_raw_string_token1] = aux_sym__format_raw_string_token1,
  [anon_sym_SQUOTE] = anon_sym_SQUOTE,
  [anon_sym_f_SQUOTE_SQUOTE_SQUOTE] = anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
  [anon_sym_SQUOTE_SQUOTE_SQUOTE] = anon_sym_SQUOTE_SQUOTE_SQUOTE,
  [aux_sym__raw_string_indented_token1] = aux_sym__raw_string_indented_token1,
  [aux_sym__string_token1] = aux_sym__string_token1,
  [anon_sym_x_DQUOTE] = anon_sym_x_DQUOTE,
  [anon_sym_x_DQUOTE_DQUOTE_DQUOTE] = anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
  [aux_sym__shell_expanded_string_indented_token1] = aux_sym__shell_expanded_string_indented_token1,
  [anon_sym_x_SQUOTE_SQUOTE_SQUOTE] = anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
  [sym_escape_sequence] = sym_escape_sequence,
  [anon_sym_BQUOTE] = anon_sym_BQUOTE,
  [anon_sym_BQUOTE_BQUOTE_BQUOTE] = anon_sym_BQUOTE_BQUOTE_BQUOTE,
  [anon_sym_LBRACE_LBRACE] = anon_sym_LBRACE_LBRACE,
  [anon_sym_RBRACE_RBRACE] = anon_sym_RBRACE_RBRACE,
  [sym_numeric_error] = sym_numeric_error,
  [sym_comment] = sym_comment,
  [sym__indent] = sym__indent,
  [sym__dedent] = sym__dedent,
  [sym__newline] = sym__newline,
  [sym_text] = sym_text,
  [sym_error_recovery] = sym_error_recovery,
  [sym_source_file] = sym_source_file,
  [sym__item] = sym__item,
  [sym_function_definition] = sym_function_definition,
  [sym_function_parameters] = sym_function_parameters,
  [sym_alias] = sym_alias,
  [sym_module_path] = sym_module_path,
  [sym__name] = sym__name,
  [sym_assignment] = sym_assignment,
  [sym__assignment] = sym__assignment,
  [sym_eager] = sym_eager,
  [sym_export] = sym_export,
  [sym_unexport] = sym_unexport,
  [sym_import] = sym_import,
  [sym_module] = sym_module,
  [sym_setting] = sym_setting,
  [sym_boolean] = sym_boolean,
  [sym_expression] = sym_expression,
  [sym__expression_inner] = sym__expression_inner,
  [sym_if_expression] = sym_if_expression,
  [sym_else_if_clause] = sym_else_if_clause,
  [sym_else_clause] = sym_else_clause,
  [sym__braced_expr] = sym__braced_expr,
  [sym_condition] = sym_condition,
  [sym_regex_literal] = sym_regex_literal,
  [sym_value] = sym_value,
  [sym_not_expression] = sym_not_expression,
  [sym_list_literal] = sym_list_literal,
  [sym_list_elements] = sym_list_elements,
  [sym_assert_expression] = sym_assert_expression,
  [sym_function_call] = sym_function_call,
  [sym_external_command] = sym_external_command,
  [sym_sequence] = sym_sequence,
  [sym_attribute] = sym_attribute,
  [sym_attribute_named_param] = sym_attribute_named_param,
  [sym_recipe] = sym_recipe,
  [sym_recipe_header] = sym_recipe_header,
  [sym_parameters] = sym_parameters,
  [sym_parameter] = sym_parameter,
  [sym_variadic_parameter] = sym_variadic_parameter,
  [sym_dependencies] = sym_dependencies,
  [sym_dependency] = sym_dependency,
  [sym_dependency_expression] = sym_dependency_expression,
  [sym_starred_dependency_argument] = sym_starred_dependency_argument,
  [sym_recipe_body] = sym_recipe_body,
  [sym_recipe_line] = sym_recipe_line,
  [sym_recipe_line_prefix] = sym_recipe_line_prefix,
  [sym_shebang] = sym_shebang,
  [sym__shebang_with_lang] = sym__shebang_with_lang,
  [sym_string] = sym_string,
  [sym_format_string] = sym_format_string,
  [sym__backticked] = sym__backticked,
  [sym__indented_backticked] = sym__indented_backticked,
  [sym_command_body] = sym_command_body,
  [sym_interpolation] = sym_interpolation,
  [aux_sym_source_file_repeat1] = aux_sym_source_file_repeat1,
  [aux_sym_function_definition_repeat1] = aux_sym_function_definition_repeat1,
  [aux_sym_function_parameters_repeat1] = aux_sym_function_parameters_repeat1,
  [aux_sym_module_path_repeat1] = aux_sym_module_path_repeat1,
  [aux_sym_if_expression_repeat1] = aux_sym_if_expression_repeat1,
  [aux_sym_list_elements_repeat1] = aux_sym_list_elements_repeat1,
  [aux_sym_attribute_repeat1] = aux_sym_attribute_repeat1,
  [aux_sym_attribute_repeat2] = aux_sym_attribute_repeat2,
  [aux_sym_parameters_repeat1] = aux_sym_parameters_repeat1,
  [aux_sym_dependencies_repeat1] = aux_sym_dependencies_repeat1,
  [aux_sym_dependency_expression_repeat1] = aux_sym_dependency_expression_repeat1,
  [aux_sym_recipe_body_repeat1] = aux_sym_recipe_body_repeat1,
  [aux_sym_recipe_line_repeat1] = aux_sym_recipe_line_repeat1,
  [aux_sym__shebang_with_lang_repeat1] = aux_sym__shebang_with_lang_repeat1,
  [aux_sym__format_string_repeat1] = aux_sym__format_string_repeat1,
  [aux_sym__format_raw_string_repeat1] = aux_sym__format_raw_string_repeat1,
  [aux_sym__raw_string_indented_repeat1] = aux_sym__raw_string_indented_repeat1,
  [aux_sym__string_repeat1] = aux_sym__string_repeat1,
  [aux_sym__shell_expanded_string_indented_repeat1] = aux_sym__shell_expanded_string_indented_repeat1,
  [aux_sym_command_body_repeat1] = aux_sym_command_body_repeat1,
  [anon_alias_sym_expression] = anon_alias_sym_expression,
  [alias_sym_language] = alias_sym_language,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [sym_identifier] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_alias] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_unexport] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_eager] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_export] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_import] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_mod] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_set] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_true] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_false] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_if] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_else] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_assert] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AT_DASH_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AT_QMARK_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_AT_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_QMARK_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_AT_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_DASH_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AT_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AT_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_shebang_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__shebang_with_lang_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_env] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__shebang_with_lang_token2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__shebang_with_lang_token3] = {
    .visible = false,
    .named = false,
  },
  [sym__opaque_shebang] = {
    .visible = false,
    .named = true,
  },
  [aux_sym_string_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_token2] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_f_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__format_string_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__format_string_token2] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_f_DQUOTE_DQUOTE_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DQUOTE_DQUOTE_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_f_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__format_raw_string_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_f_SQUOTE_SQUOTE_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SQUOTE_SQUOTE_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__raw_string_indented_token1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__string_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_x_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_x_DQUOTE_DQUOTE_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [aux_sym__shell_expanded_string_indented_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_x_SQUOTE_SQUOTE_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [sym_escape_sequence] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_BQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BQUOTE_BQUOTE_BQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [sym_numeric_error] = {
    .visible = true,
    .named = true,
  },
  [sym_comment] = {
    .visible = true,
    .named = true,
  },
  [sym__indent] = {
    .visible = false,
    .named = true,
  },
  [sym__dedent] = {
    .visible = false,
    .named = true,
  },
  [sym__newline] = {
    .visible = false,
    .named = true,
  },
  [sym_text] = {
    .visible = true,
    .named = true,
  },
  [sym_error_recovery] = {
    .visible = true,
    .named = true,
  },
  [sym_source_file] = {
    .visible = true,
    .named = true,
  },
  [sym__item] = {
    .visible = false,
    .named = true,
  },
  [sym_function_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_function_parameters] = {
    .visible = true,
    .named = true,
  },
  [sym_alias] = {
    .visible = true,
    .named = true,
  },
  [sym_module_path] = {
    .visible = true,
    .named = true,
  },
  [sym__name] = {
    .visible = false,
    .named = true,
  },
  [sym_assignment] = {
    .visible = true,
    .named = true,
  },
  [sym__assignment] = {
    .visible = false,
    .named = true,
  },
  [sym_eager] = {
    .visible = true,
    .named = true,
  },
  [sym_export] = {
    .visible = true,
    .named = true,
  },
  [sym_unexport] = {
    .visible = true,
    .named = true,
  },
  [sym_import] = {
    .visible = true,
    .named = true,
  },
  [sym_module] = {
    .visible = true,
    .named = true,
  },
  [sym_setting] = {
    .visible = true,
    .named = true,
  },
  [sym_boolean] = {
    .visible = true,
    .named = true,
  },
  [sym_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__expression_inner] = {
    .visible = false,
    .named = true,
  },
  [sym_if_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_else_if_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_else_clause] = {
    .visible = true,
    .named = true,
  },
  [sym__braced_expr] = {
    .visible = false,
    .named = true,
  },
  [sym_condition] = {
    .visible = true,
    .named = true,
  },
  [sym_regex_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_value] = {
    .visible = true,
    .named = true,
  },
  [sym_not_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_list_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_list_elements] = {
    .visible = true,
    .named = true,
  },
  [sym_assert_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_function_call] = {
    .visible = true,
    .named = true,
  },
  [sym_external_command] = {
    .visible = true,
    .named = true,
  },
  [sym_sequence] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_named_param] = {
    .visible = true,
    .named = true,
  },
  [sym_recipe] = {
    .visible = true,
    .named = true,
  },
  [sym_recipe_header] = {
    .visible = true,
    .named = true,
  },
  [sym_parameters] = {
    .visible = true,
    .named = true,
  },
  [sym_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_variadic_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_dependencies] = {
    .visible = true,
    .named = true,
  },
  [sym_dependency] = {
    .visible = true,
    .named = true,
  },
  [sym_dependency_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_starred_dependency_argument] = {
    .visible = true,
    .named = true,
  },
  [sym_recipe_body] = {
    .visible = true,
    .named = true,
  },
  [sym_recipe_line] = {
    .visible = true,
    .named = true,
  },
  [sym_recipe_line_prefix] = {
    .visible = true,
    .named = true,
  },
  [sym_shebang] = {
    .visible = true,
    .named = true,
  },
  [sym__shebang_with_lang] = {
    .visible = false,
    .named = true,
  },
  [sym_string] = {
    .visible = true,
    .named = true,
  },
  [sym_format_string] = {
    .visible = true,
    .named = true,
  },
  [sym__backticked] = {
    .visible = false,
    .named = true,
  },
  [sym__indented_backticked] = {
    .visible = false,
    .named = true,
  },
  [sym_command_body] = {
    .visible = true,
    .named = true,
  },
  [sym_interpolation] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_source_file_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_function_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_function_parameters_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_module_path_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_if_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_list_elements_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_attribute_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_attribute_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_parameters_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_dependencies_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_dependency_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_recipe_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_recipe_line_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__shebang_with_lang_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__format_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__format_raw_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__raw_string_indented_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__shell_expanded_string_indented_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_command_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [anon_alias_sym_expression] = {
    .visible = true,
    .named = false,
  },
  [alias_sym_language] = {
    .visible = true,
    .named = true,
  },
};

enum ts_field_identifiers {
  field_alternative = 1,
  field_argument = 2,
  field_arguments = 3,
  field_body = 4,
  field_condition = 5,
  field_consequence = 6,
  field_default = 7,
  field_elements = 8,
  field_kleene = 9,
  field_left = 10,
  field_map = 11,
  field_message = 12,
  field_name = 13,
  field_parameter = 14,
  field_parameters = 15,
  field_right = 16,
  field_shebang = 17,
  field_star = 18,
  field_value = 19,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_alternative] = "alternative",
  [field_argument] = "argument",
  [field_arguments] = "arguments",
  [field_body] = "body",
  [field_condition] = "condition",
  [field_consequence] = "consequence",
  [field_default] = "default",
  [field_elements] = "elements",
  [field_kleene] = "kleene",
  [field_left] = "left",
  [field_map] = "map",
  [field_message] = "message",
  [field_name] = "name",
  [field_parameter] = "parameter",
  [field_parameters] = "parameters",
  [field_right] = "right",
  [field_shebang] = "shebang",
  [field_star] = "star",
  [field_value] = "value",
};

static const TSMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [2] = {.index = 0, .length = 2},
  [3] = {.index = 2, .length = 2},
  [4] = {.index = 4, .length = 1},
  [5] = {.index = 4, .length = 1},
  [6] = {.index = 5, .length = 1},
  [7] = {.index = 2, .length = 2},
  [8] = {.index = 6, .length = 1},
  [9] = {.index = 7, .length = 1},
  [10] = {.index = 8, .length = 1},
  [11] = {.index = 5, .length = 1},
  [12] = {.index = 9, .length = 1},
  [13] = {.index = 10, .length = 2},
  [14] = {.index = 12, .length = 2},
  [15] = {.index = 14, .length = 2},
  [16] = {.index = 16, .length = 2},
  [18] = {.index = 18, .length = 2},
  [19] = {.index = 20, .length = 2},
  [20] = {.index = 22, .length = 1},
  [21] = {.index = 23, .length = 1},
  [22] = {.index = 24, .length = 2},
  [23] = {.index = 26, .length = 3},
  [24] = {.index = 29, .length = 1},
  [26] = {.index = 30, .length = 1},
  [27] = {.index = 31, .length = 2},
  [28] = {.index = 33, .length = 2},
  [29] = {.index = 35, .length = 1},
  [32] = {.index = 36, .length = 2},
  [33] = {.index = 38, .length = 2},
  [34] = {.index = 40, .length = 2},
  [35] = {.index = 42, .length = 2},
  [36] = {.index = 44, .length = 1},
  [38] = {.index = 45, .length = 2},
  [39] = {.index = 47, .length = 1},
  [40] = {.index = 48, .length = 3},
  [41] = {.index = 51, .length = 3},
  [42] = {.index = 54, .length = 2},
  [43] = {.index = 56, .length = 2},
  [44] = {.index = 58, .length = 1},
  [45] = {.index = 59, .length = 3},
  [46] = {.index = 62, .length = 2},
  [47] = {.index = 64, .length = 3},
  [48] = {.index = 67, .length = 1},
  [49] = {.index = 68, .length = 1},
  [50] = {.index = 69, .length = 4},
  [51] = {.index = 73, .length = 2},
  [52] = {.index = 75, .length = 2},
  [53] = {.index = 77, .length = 2},
  [54] = {.index = 79, .length = 2},
  [55] = {.index = 81, .length = 3},
  [56] = {.index = 84, .length = 3},
  [57] = {.index = 87, .length = 3},
  [58] = {.index = 90, .length = 1},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_left, 0, .inherited = true},
    {field_right, 0, .inherited = true},
  [2] =
    {field_left, 1, .inherited = true},
    {field_right, 1, .inherited = true},
  [4] =
    {field_name, 0},
  [5] =
    {field_name, 1},
  [6] =
    {field_kleene, 0},
  [7] =
    {field_name, 2},
  [8] =
    {field_left, 1},
  [9] =
    {field_parameter, 0},
  [10] =
    {field_left, 2, .inherited = true},
    {field_right, 2, .inherited = true},
  [12] =
    {field_left, 1},
    {field_right, 3},
  [14] =
    {field_default, 2},
    {field_name, 0},
  [16] =
    {field_argument, 0, .inherited = true},
    {field_argument, 1, .inherited = true},
  [18] =
    {field_parameter, 0},
    {field_parameter, 1, .inherited = true},
  [20] =
    {field_left, 0},
    {field_right, 2},
  [22] =
    {field_name, 3},
  [23] =
    {field_left, 2},
  [24] =
    {field_default, 3},
    {field_name, 1},
  [26] =
    {field_left, 1},
    {field_right, 2},
    {field_right, 3},
  [29] =
    {field_argument, 2, .inherited = true},
  [30] =
    {field_parameter, 1},
  [31] =
    {field_parameter, 0, .inherited = true},
    {field_parameter, 1, .inherited = true},
  [33] =
    {field_body, 2, .inherited = true},
    {field_consequence, 2},
  [35] =
    {field_elements, 1},
  [36] =
    {field_left, 2},
    {field_right, 4},
  [38] =
    {field_argument, 1},
    {field_star, 0},
  [40] =
    {field_map, 0},
    {field_name, 2},
  [42] =
    {field_name, 0},
    {field_value, 2},
  [44] =
    {field_argument, 3},
  [45] =
    {field_body, 4},
    {field_name, 0},
  [47] =
    {field_alternative, 0},
  [48] =
    {field_alternative, 3},
    {field_body, 2, .inherited = true},
    {field_consequence, 2},
  [51] =
    {field_alternative, 3, .inherited = true},
    {field_body, 2, .inherited = true},
    {field_consequence, 2},
  [54] =
    {field_condition, 2},
    {field_name, 0},
  [56] =
    {field_arguments, 2},
    {field_name, 0},
  [58] =
    {field_shebang, 1},
  [59] =
    {field_left, 2},
    {field_right, 3},
    {field_right, 4},
  [62] =
    {field_argument, 3},
    {field_argument, 4, .inherited = true},
  [64] =
    {field_body, 5},
    {field_name, 0},
    {field_parameters, 2},
  [67] =
    {field_body, 1},
  [68] =
    {field_body, 1, .inherited = true},
  [69] =
    {field_alternative, 3, .inherited = true},
    {field_alternative, 4},
    {field_body, 2, .inherited = true},
    {field_consequence, 2},
  [73] =
    {field_alternative, 0, .inherited = true},
    {field_alternative, 1, .inherited = true},
  [75] =
    {field_body, 5},
    {field_name, 1},
  [77] =
    {field_argument, 3},
    {field_argument, 5, .inherited = true},
  [79] =
    {field_argument, 3},
    {field_argument, 4},
  [81] =
    {field_condition, 2},
    {field_message, 4},
    {field_name, 0},
  [84] =
    {field_body, 6},
    {field_name, 1},
    {field_parameters, 3},
  [87] =
    {field_argument, 3},
    {field_argument, 4},
    {field_argument, 6, .inherited = true},
  [90] =
    {field_body, 3, .inherited = true},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [1] = {
    [0] = sym_identifier,
  },
  [3] = {
    [1] = sym_assignment,
  },
  [4] = {
    [0] = sym_identifier,
  },
  [11] = {
    [1] = sym_identifier,
  },
  [13] = {
    [2] = sym_assignment,
  },
  [17] = {
    [1] = alias_sym_language,
  },
  [25] = {
    [2] = alias_sym_language,
  },
  [30] = {
    [0] = anon_alias_sym_expression,
    [2] = anon_alias_sym_expression,
  },
  [31] = {
    [0] = anon_alias_sym_expression,
  },
  [37] = {
    [3] = alias_sym_language,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  sym__assignment, 2,
    sym__assignment,
    sym_assignment,
  sym__expression_inner, 2,
    sym__expression_inner,
    anon_alias_sym_expression,
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 3,
  [4] = 4,
  [5] = 5,
  [6] = 6,
  [7] = 7,
  [8] = 8,
  [9] = 9,
  [10] = 10,
  [11] = 11,
  [12] = 12,
  [13] = 13,
  [14] = 14,
  [15] = 15,
  [16] = 16,
  [17] = 17,
  [18] = 18,
  [19] = 15,
  [20] = 15,
  [21] = 16,
  [22] = 16,
  [23] = 23,
  [24] = 24,
  [25] = 25,
  [26] = 26,
  [27] = 27,
  [28] = 28,
  [29] = 29,
  [30] = 30,
  [31] = 31,
  [32] = 32,
  [33] = 33,
  [34] = 34,
  [35] = 35,
  [36] = 36,
  [37] = 37,
  [38] = 38,
  [39] = 39,
  [40] = 35,
  [41] = 36,
  [42] = 27,
  [43] = 35,
  [44] = 36,
  [45] = 27,
  [46] = 46,
  [47] = 47,
  [48] = 48,
  [49] = 49,
  [50] = 50,
  [51] = 51,
  [52] = 52,
  [53] = 53,
  [54] = 50,
  [55] = 50,
  [56] = 46,
  [57] = 57,
  [58] = 58,
  [59] = 57,
  [60] = 47,
  [61] = 61,
  [62] = 50,
  [63] = 50,
  [64] = 46,
  [65] = 65,
  [66] = 57,
  [67] = 47,
  [68] = 68,
  [69] = 50,
  [70] = 50,
  [71] = 71,
  [72] = 71,
  [73] = 71,
  [74] = 71,
  [75] = 75,
  [76] = 76,
  [77] = 77,
  [78] = 78,
  [79] = 79,
  [80] = 75,
  [81] = 77,
  [82] = 78,
  [83] = 79,
  [84] = 75,
  [85] = 85,
  [86] = 86,
  [87] = 87,
  [88] = 88,
  [89] = 89,
  [90] = 90,
  [91] = 76,
  [92] = 92,
  [93] = 93,
  [94] = 94,
  [95] = 95,
  [96] = 77,
  [97] = 78,
  [98] = 79,
  [99] = 75,
  [100] = 100,
  [101] = 101,
  [102] = 102,
  [103] = 103,
  [104] = 104,
  [105] = 105,
  [106] = 106,
  [107] = 107,
  [108] = 108,
  [109] = 77,
  [110] = 78,
  [111] = 76,
  [112] = 79,
  [113] = 76,
  [114] = 114,
  [115] = 115,
  [116] = 116,
  [117] = 117,
  [118] = 118,
  [119] = 119,
  [120] = 120,
  [121] = 121,
  [122] = 122,
  [123] = 123,
  [124] = 124,
  [125] = 122,
  [126] = 122,
  [127] = 122,
  [128] = 86,
  [129] = 85,
  [130] = 130,
  [131] = 87,
  [132] = 132,
  [133] = 114,
  [134] = 134,
  [135] = 135,
  [136] = 90,
  [137] = 137,
  [138] = 138,
  [139] = 139,
  [140] = 140,
  [141] = 141,
  [142] = 142,
  [143] = 143,
  [144] = 144,
  [145] = 145,
  [146] = 146,
  [147] = 147,
  [148] = 148,
  [149] = 149,
  [150] = 150,
  [151] = 117,
  [152] = 152,
  [153] = 95,
  [154] = 94,
  [155] = 92,
  [156] = 88,
  [157] = 93,
  [158] = 108,
  [159] = 106,
  [160] = 105,
  [161] = 107,
  [162] = 89,
  [163] = 100,
  [164] = 104,
  [165] = 102,
  [166] = 101,
  [167] = 103,
  [168] = 168,
  [169] = 116,
  [170] = 119,
  [171] = 118,
  [172] = 115,
  [173] = 120,
  [174] = 5,
  [175] = 5,
  [176] = 176,
  [177] = 4,
  [178] = 178,
  [179] = 4,
  [180] = 180,
  [181] = 181,
  [182] = 182,
  [183] = 11,
  [184] = 138,
  [185] = 185,
  [186] = 186,
  [187] = 137,
  [188] = 11,
  [189] = 189,
  [190] = 142,
  [191] = 191,
  [192] = 192,
  [193] = 193,
  [194] = 194,
  [195] = 195,
  [196] = 14,
  [197] = 197,
  [198] = 198,
  [199] = 199,
  [200] = 200,
  [201] = 201,
  [202] = 202,
  [203] = 203,
  [204] = 204,
  [205] = 205,
  [206] = 206,
  [207] = 207,
  [208] = 208,
  [209] = 18,
  [210] = 210,
  [211] = 211,
  [212] = 212,
  [213] = 213,
  [214] = 214,
  [215] = 17,
  [216] = 18,
  [217] = 14,
  [218] = 117,
  [219] = 219,
  [220] = 220,
  [221] = 221,
  [222] = 222,
  [223] = 223,
  [224] = 17,
  [225] = 117,
  [226] = 226,
  [227] = 227,
  [228] = 228,
  [229] = 229,
  [230] = 103,
  [231] = 231,
  [232] = 232,
  [233] = 39,
  [234] = 234,
  [235] = 235,
  [236] = 236,
  [237] = 120,
  [238] = 92,
  [239] = 93,
  [240] = 94,
  [241] = 95,
  [242] = 108,
  [243] = 85,
  [244] = 86,
  [245] = 87,
  [246] = 246,
  [247] = 115,
  [248] = 29,
  [249] = 249,
  [250] = 120,
  [251] = 251,
  [252] = 252,
  [253] = 100,
  [254] = 254,
  [255] = 101,
  [256] = 102,
  [257] = 31,
  [258] = 32,
  [259] = 33,
  [260] = 31,
  [261] = 261,
  [262] = 34,
  [263] = 116,
  [264] = 32,
  [265] = 33,
  [266] = 39,
  [267] = 38,
  [268] = 118,
  [269] = 119,
  [270] = 90,
  [271] = 271,
  [272] = 114,
  [273] = 116,
  [274] = 118,
  [275] = 119,
  [276] = 276,
  [277] = 277,
  [278] = 278,
  [279] = 279,
  [280] = 107,
  [281] = 34,
  [282] = 88,
  [283] = 105,
  [284] = 115,
  [285] = 29,
  [286] = 106,
  [287] = 89,
  [288] = 38,
  [289] = 289,
  [290] = 290,
  [291] = 291,
  [292] = 104,
  [293] = 293,
  [294] = 294,
  [295] = 295,
  [296] = 296,
  [297] = 297,
  [298] = 298,
  [299] = 299,
  [300] = 138,
  [301] = 301,
  [302] = 302,
  [303] = 303,
  [304] = 304,
  [305] = 304,
  [306] = 137,
  [307] = 307,
  [308] = 301,
  [309] = 302,
  [310] = 303,
  [311] = 304,
  [312] = 312,
  [313] = 301,
  [314] = 302,
  [315] = 315,
  [316] = 316,
  [317] = 303,
  [318] = 299,
  [319] = 319,
  [320] = 320,
  [321] = 321,
  [322] = 322,
  [323] = 323,
  [324] = 324,
  [325] = 325,
  [326] = 326,
  [327] = 327,
  [328] = 142,
  [329] = 329,
  [330] = 320,
  [331] = 321,
  [332] = 322,
  [333] = 325,
  [334] = 325,
  [335] = 326,
  [336] = 326,
  [337] = 321,
  [338] = 338,
  [339] = 329,
  [340] = 322,
  [341] = 341,
  [342] = 320,
  [343] = 329,
  [344] = 323,
  [345] = 345,
  [346] = 346,
  [347] = 347,
  [348] = 348,
  [349] = 349,
  [350] = 350,
  [351] = 351,
  [352] = 352,
  [353] = 352,
  [354] = 354,
  [355] = 355,
  [356] = 347,
  [357] = 357,
  [358] = 358,
  [359] = 359,
  [360] = 360,
  [361] = 351,
  [362] = 362,
  [363] = 363,
  [364] = 364,
  [365] = 365,
  [366] = 366,
  [367] = 367,
  [368] = 368,
  [369] = 369,
  [370] = 370,
  [371] = 371,
  [372] = 372,
  [373] = 373,
  [374] = 347,
  [375] = 375,
  [376] = 376,
  [377] = 377,
  [378] = 347,
  [379] = 366,
  [380] = 380,
  [381] = 369,
  [382] = 370,
  [383] = 383,
  [384] = 384,
  [385] = 366,
  [386] = 380,
  [387] = 369,
  [388] = 370,
  [389] = 380,
  [390] = 390,
  [391] = 391,
  [392] = 392,
  [393] = 393,
  [394] = 394,
  [395] = 395,
  [396] = 396,
  [397] = 397,
  [398] = 398,
  [399] = 399,
  [400] = 400,
  [401] = 401,
  [402] = 402,
  [403] = 403,
  [404] = 404,
  [405] = 405,
  [406] = 406,
  [407] = 347,
  [408] = 347,
  [409] = 347,
  [410] = 410,
  [411] = 411,
  [412] = 412,
  [413] = 390,
  [414] = 414,
  [415] = 415,
  [416] = 416,
  [417] = 417,
  [418] = 418,
  [419] = 419,
  [420] = 401,
  [421] = 421,
  [422] = 422,
  [423] = 390,
  [424] = 416,
  [425] = 425,
  [426] = 426,
  [427] = 416,
  [428] = 428,
  [429] = 401,
  [430] = 430,
  [431] = 431,
  [432] = 432,
  [433] = 433,
  [434] = 432,
  [435] = 435,
  [436] = 436,
  [437] = 437,
  [438] = 433,
  [439] = 432,
  [440] = 440,
  [441] = 436,
  [442] = 442,
  [443] = 443,
  [444] = 444,
  [445] = 445,
  [446] = 446,
  [447] = 447,
  [448] = 437,
  [449] = 433,
  [450] = 450,
  [451] = 451,
  [452] = 452,
  [453] = 436,
  [454] = 437,
  [455] = 455,
  [456] = 456,
  [457] = 457,
  [458] = 458,
  [459] = 459,
  [460] = 460,
  [461] = 461,
  [462] = 462,
  [463] = 463,
  [464] = 464,
  [465] = 465,
  [466] = 466,
  [467] = 467,
  [468] = 468,
  [469] = 469,
  [470] = 470,
  [471] = 471,
  [472] = 472,
  [473] = 473,
  [474] = 474,
  [475] = 475,
  [476] = 476,
  [477] = 477,
  [478] = 478,
  [479] = 479,
  [480] = 480,
  [481] = 481,
  [482] = 482,
  [483] = 483,
  [484] = 484,
  [485] = 485,
  [486] = 486,
  [487] = 487,
  [488] = 488,
  [489] = 489,
  [490] = 469,
  [491] = 491,
  [492] = 492,
  [493] = 493,
  [494] = 494,
  [495] = 492,
  [496] = 496,
  [497] = 497,
  [498] = 498,
  [499] = 499,
  [500] = 500,
  [501] = 501,
  [502] = 502,
  [503] = 341,
  [504] = 504,
  [505] = 493,
  [506] = 506,
  [507] = 507,
  [508] = 508,
  [509] = 509,
  [510] = 510,
  [511] = 511,
  [512] = 494,
  [513] = 513,
  [514] = 514,
  [515] = 515,
  [516] = 516,
  [517] = 517,
  [518] = 518,
  [519] = 519,
  [520] = 520,
  [521] = 521,
  [522] = 522,
  [523] = 523,
  [524] = 488,
  [525] = 489,
  [526] = 469,
  [527] = 527,
  [528] = 492,
  [529] = 493,
  [530] = 494,
  [531] = 498,
  [532] = 532,
  [533] = 497,
  [534] = 498,
  [535] = 535,
  [536] = 536,
  [537] = 537,
  [538] = 502,
  [539] = 489,
  [540] = 489,
  [541] = 489,
  [542] = 489,
  [543] = 502,
  [544] = 544,
  [545] = 545,
  [546] = 546,
  [547] = 547,
  [548] = 548,
  [549] = 497,
  [550] = 523,
  [551] = 551,
  [552] = 552,
  [553] = 553,
  [554] = 554,
  [555] = 555,
  [556] = 556,
  [557] = 557,
  [558] = 523,
  [559] = 559,
  [560] = 560,
  [561] = 489,
  [562] = 562,
  [563] = 557,
  [564] = 557,
  [565] = 488,
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(65);
      ADVANCE_MAP(
        '!', 85,
        '"', 131,
        '#', 173,
        '$', 94,
        '&', 25,
        '\'', 140,
        '(', 66,
        ')', 67,
        '*', 95,
        '+', 75,
        ',', 69,
        '-', 108,
        '/', 73,
        ':', 89,
        '=', 91,
        '?', 72,
        '@', 93,
        '[', 86,
      );
      if (lookahead == '\\') SKIP(51);
      if (lookahead == ']') ADVANCE(87);
      if (lookahead == '`') ADVANCE(164);
      if (lookahead == 'f') ADVANCE(168);
      if (lookahead == 'x') ADVANCE(169);
      if (lookahead == '{') ADVANCE(83);
      if (lookahead == '|') ADVANCE(41);
      if (lookahead == '}') ADVANCE(84);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(0);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(171);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(170);
      END_STATE();
    case 1:
      if (lookahead == '\n') SKIP(13);
      END_STATE();
    case 2:
      if (lookahead == '\n') SKIP(13);
      if (lookahead == '\r') SKIP(1);
      END_STATE();
    case 3:
      if (lookahead == '\n') ADVANCE(160);
      END_STATE();
    case 4:
      ADVANCE_MAP(
        '\n', 160,
        '\r', 3,
        'u', 40,
        '"', 159,
        '\\', 159,
        'n', 159,
        'r', 159,
        't', 159,
      );
      END_STATE();
    case 5:
      if (lookahead == '\n') SKIP(5);
      if (lookahead == '#') ADVANCE(143);
      if (lookahead == '\\') ADVANCE(143);
      if (lookahead == '`') ADVANCE(163);
      if (lookahead == '{') ADVANCE(149);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(145);
      if (lookahead != 0 &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(143);
      END_STATE();
    case 6:
      if (lookahead == '\n') SKIP(6);
      if (lookahead == '#') ADVANCE(143);
      if (lookahead == '\'') ADVANCE(147);
      if (lookahead == '\\') ADVANCE(143);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(144);
      if (lookahead != 0 &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(143);
      END_STATE();
    case 7:
      if (lookahead == '\n') SKIP(7);
      if (lookahead == '#') ADVANCE(119);
      if (lookahead == '/') ADVANCE(110);
      if (lookahead == '\\') ADVANCE(116);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(117);
      if (lookahead != 0 &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(118);
      END_STATE();
    case 8:
      if (lookahead == '\n') ADVANCE(161);
      END_STATE();
    case 9:
      ADVANCE_MAP(
        '\n', 161,
        '\r', 8,
        'u', 40,
        '"', 159,
        '\\', 159,
        'n', 159,
        'r', 159,
        't', 159,
      );
      END_STATE();
    case 10:
      if (lookahead == '\n') SKIP(10);
      if (lookahead == '#') ADVANCE(143);
      if (lookahead == '\\') ADVANCE(143);
      if (lookahead == '`') ADVANCE(148);
      if (lookahead == '{') ADVANCE(149);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(146);
      if (lookahead != 0 &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(143);
      END_STATE();
    case 11:
      if (lookahead == '\n') ADVANCE(162);
      END_STATE();
    case 12:
      ADVANCE_MAP(
        '\n', 162,
        '\r', 11,
        'u', 40,
        '"', 159,
        '\\', 159,
        'n', 159,
        'r', 159,
        't', 159,
      );
      END_STATE();
    case 13:
      ADVANCE_MAP(
        '!', 85,
        '"', 131,
        '#', 175,
        '$', 94,
        '&', 25,
        '\'', 29,
        '(', 66,
        ')', 67,
        '*', 95,
        '+', 75,
        ',', 69,
        '-', 108,
        '/', 73,
        ':', 88,
        '=', 37,
        '?', 72,
        '@', 93,
        '[', 86,
      );
      if (lookahead == '\\') SKIP(2);
      if (lookahead == ']') ADVANCE(87);
      if (lookahead == '`') ADVANCE(164);
      if (lookahead == 'f') ADVANCE(168);
      if (lookahead == 'x') ADVANCE(169);
      if (lookahead == '{') ADVANCE(39);
      if (lookahead == '|') ADVANCE(41);
      if (lookahead == '}') ADVANCE(84);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(13);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(171);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(170);
      END_STATE();
    case 14:
      if (lookahead == '"') ADVANCE(16);
      END_STATE();
    case 15:
      if (lookahead == '"') ADVANCE(16);
      if (lookahead != 0 &&
          lookahead != '\\' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(157);
      END_STATE();
    case 16:
      if (lookahead == '"') ADVANCE(133);
      END_STATE();
    case 17:
      if (lookahead == '"') ADVANCE(132);
      END_STATE();
    case 18:
      if (lookahead == '"') ADVANCE(154);
      END_STATE();
    case 19:
      if (lookahead == '"') ADVANCE(130);
      if (lookahead == '#') ADVANCE(126);
      if (lookahead == '\\') ADVANCE(4);
      if (lookahead == '{') ADVANCE(129);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(127);
      if (lookahead != 0 &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(128);
      END_STATE();
    case 20:
      if (lookahead == '"') ADVANCE(130);
      if (lookahead == '#') ADVANCE(150);
      if (lookahead == '\\') ADVANCE(9);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(151);
      if (lookahead != 0 &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(152);
      END_STATE();
    case 21:
      if (lookahead == '"') ADVANCE(15);
      if (lookahead == '#') ADVANCE(155);
      if (lookahead == '\\') ADVANCE(12);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(156);
      if (lookahead != 0 &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(157);
      END_STATE();
    case 22:
      if (lookahead == '"') ADVANCE(14);
      if (lookahead == '#') ADVANCE(126);
      if (lookahead == '\\') ADVANCE(4);
      if (lookahead == '{') ADVANCE(129);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(127);
      if (lookahead != 0 &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(128);
      END_STATE();
    case 23:
      if (lookahead == '#') ADVANCE(137);
      if (lookahead == '\'') ADVANCE(140);
      if (lookahead == '\\') ADVANCE(135);
      if (lookahead == '{') ADVANCE(129);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(138);
      if (lookahead != 0 &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(139);
      END_STATE();
    case 24:
      if (lookahead == '#') ADVANCE(137);
      if (lookahead == '\'') ADVANCE(32);
      if (lookahead == '\\') ADVANCE(135);
      if (lookahead == '{') ADVANCE(129);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(138);
      if (lookahead != 0 &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(139);
      END_STATE();
    case 25:
      if (lookahead == '&') ADVANCE(80);
      END_STATE();
    case 26:
      if (lookahead == '\'') ADVANCE(122);
      if (lookahead != 0 &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(28);
      END_STATE();
    case 27:
      if (lookahead == '\'') ADVANCE(141);
      END_STATE();
    case 28:
      if (lookahead == '\'') ADVANCE(121);
      if (lookahead != 0 &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(28);
      END_STATE();
    case 29:
      if (lookahead == '\'') ADVANCE(124);
      if (lookahead != 0 &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(31);
      END_STATE();
    case 30:
      if (lookahead == '\'') ADVANCE(142);
      END_STATE();
    case 31:
      if (lookahead == '\'') ADVANCE(123);
      if (lookahead != 0 &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(31);
      END_STATE();
    case 32:
      if (lookahead == '\'') ADVANCE(30);
      END_STATE();
    case 33:
      if (lookahead == '/') ADVANCE(110);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(33);
      END_STATE();
    case 34:
      if (lookahead == ':') ADVANCE(70);
      END_STATE();
    case 35:
      if (lookahead == ':') ADVANCE(70);
      if (lookahead == '=') ADVANCE(68);
      END_STATE();
    case 36:
      if (lookahead == '=') ADVANCE(77);
      if (lookahead == '~') ADVANCE(79);
      END_STATE();
    case 37:
      if (lookahead == '=') ADVANCE(76);
      if (lookahead == '~') ADVANCE(78);
      END_STATE();
    case 38:
      if (lookahead == '`') ADVANCE(165);
      END_STATE();
    case 39:
      if (lookahead == '{') ADVANCE(166);
      END_STATE();
    case 40:
      if (lookahead == '{') ADVANCE(49);
      END_STATE();
    case 41:
      if (lookahead == '|') ADVANCE(81);
      END_STATE();
    case 42:
      if (lookahead == '}') ADVANCE(167);
      END_STATE();
    case 43:
      if (lookahead == '}') ADVANCE(159);
      END_STATE();
    case 44:
      if (lookahead == '}') ADVANCE(159);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(43);
      END_STATE();
    case 45:
      if (lookahead == '}') ADVANCE(159);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(44);
      END_STATE();
    case 46:
      if (lookahead == '}') ADVANCE(159);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(45);
      END_STATE();
    case 47:
      if (lookahead == '}') ADVANCE(159);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(46);
      END_STATE();
    case 48:
      if (lookahead == '}') ADVANCE(159);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(47);
      END_STATE();
    case 49:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(48);
      END_STATE();
    case 50:
      if (eof) ADVANCE(65);
      if (lookahead == '\n') SKIP(60);
      END_STATE();
    case 51:
      if (eof) ADVANCE(65);
      if (lookahead == '\n') SKIP(60);
      if (lookahead == '\r') SKIP(50);
      END_STATE();
    case 52:
      if (eof) ADVANCE(65);
      if (lookahead == '\n') SKIP(64);
      END_STATE();
    case 53:
      if (eof) ADVANCE(65);
      if (lookahead == '\n') SKIP(64);
      if (lookahead == '\r') SKIP(52);
      END_STATE();
    case 54:
      if (eof) ADVANCE(65);
      if (lookahead == '\n') SKIP(61);
      END_STATE();
    case 55:
      if (eof) ADVANCE(65);
      if (lookahead == '\n') SKIP(61);
      if (lookahead == '\r') SKIP(54);
      END_STATE();
    case 56:
      if (eof) ADVANCE(65);
      if (lookahead == '\n') SKIP(62);
      END_STATE();
    case 57:
      if (eof) ADVANCE(65);
      if (lookahead == '\n') SKIP(62);
      if (lookahead == '\r') SKIP(56);
      END_STATE();
    case 58:
      if (eof) ADVANCE(65);
      if (lookahead == '\n') SKIP(63);
      END_STATE();
    case 59:
      if (eof) ADVANCE(65);
      if (lookahead == '\n') SKIP(63);
      if (lookahead == '\r') SKIP(58);
      END_STATE();
    case 60:
      if (eof) ADVANCE(65);
      ADVANCE_MAP(
        '!', 85,
        '"', 131,
        '#', 173,
        '$', 94,
        '&', 25,
        '\'', 140,
        '(', 66,
        ')', 67,
        '*', 95,
        '+', 75,
        ',', 69,
        '-', 108,
        '/', 73,
        ':', 89,
        '=', 91,
        '?', 72,
        '@', 93,
        '[', 86,
      );
      if (lookahead == '\\') SKIP(51);
      if (lookahead == ']') ADVANCE(87);
      if (lookahead == '`') ADVANCE(164);
      if (lookahead == 'f') ADVANCE(168);
      if (lookahead == 'x') ADVANCE(169);
      if (lookahead == '{') ADVANCE(83);
      if (lookahead == '|') ADVANCE(41);
      if (lookahead == '}') ADVANCE(84);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(60);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(171);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(170);
      END_STATE();
    case 61:
      if (eof) ADVANCE(65);
      ADVANCE_MAP(
        '!', 85,
        '"', 131,
        '#', 175,
        '&', 25,
        '\'', 29,
        '(', 66,
        ')', 67,
        '*', 95,
        '+', 75,
        ',', 69,
        '/', 73,
        ':', 35,
        '=', 91,
        '@', 92,
        '[', 86,
      );
      if (lookahead == '\\') SKIP(55);
      if (lookahead == ']') ADVANCE(87);
      if (lookahead == '`') ADVANCE(164);
      if (lookahead == 'f') ADVANCE(168);
      if (lookahead == 'x') ADVANCE(169);
      if (lookahead == '{') ADVANCE(82);
      if (lookahead == '|') ADVANCE(41);
      if (lookahead == '}') ADVANCE(42);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(61);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(171);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(170);
      END_STATE();
    case 62:
      if (eof) ADVANCE(65);
      ADVANCE_MAP(
        '!', 36,
        '#', 175,
        '$', 94,
        '&', 25,
        '(', 66,
        ')', 67,
        '*', 95,
        '+', 75,
        ',', 69,
        '-', 112,
        '/', 73,
        ':', 90,
        '=', 91,
        '?', 71,
        '@', 92,
        '[', 86,
      );
      if (lookahead == '\\') SKIP(57);
      if (lookahead == ']') ADVANCE(87);
      if (lookahead == '`') ADVANCE(163);
      if (lookahead == '{') ADVANCE(82);
      if (lookahead == '|') ADVANCE(41);
      if (lookahead == '}') ADVANCE(42);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(62);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('_' <= lookahead && lookahead <= 'z')) ADVANCE(170);
      END_STATE();
    case 63:
      if (eof) ADVANCE(65);
      ADVANCE_MAP(
        '!', 36,
        '#', 175,
        '&', 25,
        '(', 66,
        '*', 95,
        '+', 75,
        '/', 73,
        ':', 34,
        '=', 37,
        '@', 92,
        '[', 86,
      );
      if (lookahead == '\\') SKIP(59);
      if (lookahead == '|') ADVANCE(41);
      if (lookahead == '}') ADVANCE(84);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(63);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(170);
      END_STATE();
    case 64:
      if (eof) ADVANCE(65);
      if (lookahead == '#') ADVANCE(173);
      if (lookahead == '-') ADVANCE(108);
      if (lookahead == '?') ADVANCE(72);
      if (lookahead == '@') ADVANCE(93);
      if (lookahead == '[') ADVANCE(86);
      if (lookahead == '\\') SKIP(53);
      if (lookahead == '{') ADVANCE(39);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(64);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(170);
      END_STATE();
    case 65:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 66:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 67:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 68:
      ACCEPT_TOKEN(anon_sym_COLON_EQ);
      END_STATE();
    case 69:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 70:
      ACCEPT_TOKEN(anon_sym_COLON_COLON);
      END_STATE();
    case 71:
      ACCEPT_TOKEN(anon_sym_QMARK);
      END_STATE();
    case 72:
      ACCEPT_TOKEN(anon_sym_QMARK);
      if (lookahead == '-') ADVANCE(107);
      if (lookahead == '@') ADVANCE(106);
      END_STATE();
    case 73:
      ACCEPT_TOKEN(anon_sym_SLASH);
      END_STATE();
    case 74:
      ACCEPT_TOKEN(anon_sym_PLUS_PLUS);
      END_STATE();
    case 75:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(74);
      END_STATE();
    case 76:
      ACCEPT_TOKEN(anon_sym_EQ_EQ);
      END_STATE();
    case 77:
      ACCEPT_TOKEN(anon_sym_BANG_EQ);
      END_STATE();
    case 78:
      ACCEPT_TOKEN(anon_sym_EQ_TILDE);
      END_STATE();
    case 79:
      ACCEPT_TOKEN(anon_sym_BANG_TILDE);
      END_STATE();
    case 80:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      END_STATE();
    case 81:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      END_STATE();
    case 82:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 83:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      if (lookahead == '{') ADVANCE(166);
      END_STATE();
    case 84:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 85:
      ACCEPT_TOKEN(anon_sym_BANG);
      if (lookahead == '=') ADVANCE(77);
      if (lookahead == '~') ADVANCE(79);
      END_STATE();
    case 86:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      END_STATE();
    case 87:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 88:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 89:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(70);
      if (lookahead == '=') ADVANCE(68);
      END_STATE();
    case 90:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == '=') ADVANCE(68);
      END_STATE();
    case 91:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '=') ADVANCE(76);
      if (lookahead == '~') ADVANCE(78);
      END_STATE();
    case 92:
      ACCEPT_TOKEN(anon_sym_AT);
      END_STATE();
    case 93:
      ACCEPT_TOKEN(anon_sym_AT);
      if (lookahead == '-') ADVANCE(102);
      if (lookahead == '?') ADVANCE(103);
      END_STATE();
    case 94:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      END_STATE();
    case 95:
      ACCEPT_TOKEN(anon_sym_STAR);
      END_STATE();
    case 96:
      ACCEPT_TOKEN(anon_sym_AT_DASH_QMARK);
      END_STATE();
    case 97:
      ACCEPT_TOKEN(anon_sym_AT_QMARK_DASH);
      END_STATE();
    case 98:
      ACCEPT_TOKEN(anon_sym_DASH_AT_QMARK);
      END_STATE();
    case 99:
      ACCEPT_TOKEN(anon_sym_DASH_QMARK_AT);
      END_STATE();
    case 100:
      ACCEPT_TOKEN(anon_sym_QMARK_AT_DASH);
      END_STATE();
    case 101:
      ACCEPT_TOKEN(anon_sym_QMARK_DASH_AT);
      END_STATE();
    case 102:
      ACCEPT_TOKEN(anon_sym_AT_DASH);
      if (lookahead == '?') ADVANCE(96);
      END_STATE();
    case 103:
      ACCEPT_TOKEN(anon_sym_AT_QMARK);
      if (lookahead == '-') ADVANCE(97);
      END_STATE();
    case 104:
      ACCEPT_TOKEN(anon_sym_DASH_AT);
      if (lookahead == '?') ADVANCE(98);
      END_STATE();
    case 105:
      ACCEPT_TOKEN(anon_sym_DASH_QMARK);
      if (lookahead == '@') ADVANCE(99);
      END_STATE();
    case 106:
      ACCEPT_TOKEN(anon_sym_QMARK_AT);
      if (lookahead == '-') ADVANCE(100);
      END_STATE();
    case 107:
      ACCEPT_TOKEN(anon_sym_QMARK_DASH);
      if (lookahead == '@') ADVANCE(101);
      END_STATE();
    case 108:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '?') ADVANCE(105);
      if (lookahead == '@') ADVANCE(104);
      END_STATE();
    case 109:
      ACCEPT_TOKEN(aux_sym_shebang_token1);
      if (lookahead == '\t' ||
          lookahead == ' ') ADVANCE(109);
      END_STATE();
    case 110:
      ACCEPT_TOKEN(aux_sym__shebang_with_lang_token1);
      if (lookahead == '/') ADVANCE(110);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(33);
      END_STATE();
    case 111:
      ACCEPT_TOKEN(aux_sym__shebang_with_lang_token1);
      if (lookahead == '/') ADVANCE(111);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(174);
      END_STATE();
    case 112:
      ACCEPT_TOKEN(aux_sym__shebang_with_lang_token2);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != ' ' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(112);
      END_STATE();
    case 113:
      ACCEPT_TOKEN(aux_sym__shebang_with_lang_token3);
      if (lookahead == '\r') ADVANCE(115);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(115);
      END_STATE();
    case 114:
      ACCEPT_TOKEN(aux_sym__shebang_with_lang_token3);
      if (lookahead == '#') ADVANCE(115);
      if (lookahead == '\\') ADVANCE(113);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(114);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(115);
      END_STATE();
    case 115:
      ACCEPT_TOKEN(aux_sym__shebang_with_lang_token3);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(115);
      END_STATE();
    case 116:
      ACCEPT_TOKEN(sym__opaque_shebang);
      if (lookahead == '\r') ADVANCE(120);
      if (lookahead == '\t' ||
          lookahead == 0x0b ||
          lookahead == '\f' ||
          lookahead == ' ') ADVANCE(120);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '/' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(118);
      END_STATE();
    case 117:
      ACCEPT_TOKEN(sym__opaque_shebang);
      if (lookahead == '#') ADVANCE(119);
      if (lookahead == '\\') ADVANCE(116);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(117);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '/' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(118);
      END_STATE();
    case 118:
      ACCEPT_TOKEN(sym__opaque_shebang);
      if (lookahead == '/') ADVANCE(110);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(120);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(118);
      END_STATE();
    case 119:
      ACCEPT_TOKEN(sym__opaque_shebang);
      if (lookahead == '/') ADVANCE(111);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(120);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(119);
      END_STATE();
    case 120:
      ACCEPT_TOKEN(sym__opaque_shebang);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '/' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(120);
      END_STATE();
    case 121:
      ACCEPT_TOKEN(aux_sym_string_token1);
      END_STATE();
    case 122:
      ACCEPT_TOKEN(aux_sym_string_token1);
      if (lookahead == '\'') ADVANCE(158);
      END_STATE();
    case 123:
      ACCEPT_TOKEN(aux_sym_string_token2);
      END_STATE();
    case 124:
      ACCEPT_TOKEN(aux_sym_string_token2);
      if (lookahead == '\'') ADVANCE(142);
      END_STATE();
    case 125:
      ACCEPT_TOKEN(anon_sym_f_DQUOTE);
      if (lookahead == '"') ADVANCE(17);
      END_STATE();
    case 126:
      ACCEPT_TOKEN(aux_sym__format_string_token1);
      if (lookahead == '\n') ADVANCE(128);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '\\' &&
          lookahead != '{' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(126);
      END_STATE();
    case 127:
      ACCEPT_TOKEN(aux_sym__format_string_token1);
      if (lookahead == '#') ADVANCE(126);
      if (lookahead == '\\') ADVANCE(4);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(127);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '#' &&
          lookahead != '{' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(128);
      END_STATE();
    case 128:
      ACCEPT_TOKEN(aux_sym__format_string_token1);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '\\' &&
          lookahead != '{' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(128);
      END_STATE();
    case 129:
      ACCEPT_TOKEN(aux_sym__format_string_token2);
      if (lookahead == '{') ADVANCE(166);
      END_STATE();
    case 130:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      END_STATE();
    case 131:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      if (lookahead == '"') ADVANCE(16);
      END_STATE();
    case 132:
      ACCEPT_TOKEN(anon_sym_f_DQUOTE_DQUOTE_DQUOTE);
      END_STATE();
    case 133:
      ACCEPT_TOKEN(anon_sym_DQUOTE_DQUOTE_DQUOTE);
      END_STATE();
    case 134:
      ACCEPT_TOKEN(anon_sym_f_SQUOTE);
      if (lookahead == '\'') ADVANCE(27);
      END_STATE();
    case 135:
      ACCEPT_TOKEN(aux_sym__format_raw_string_token1);
      if (lookahead == '\n') ADVANCE(138);
      if (lookahead == '\r') ADVANCE(136);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '{' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(139);
      END_STATE();
    case 136:
      ACCEPT_TOKEN(aux_sym__format_raw_string_token1);
      if (lookahead == '\n') ADVANCE(138);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '{' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(139);
      END_STATE();
    case 137:
      ACCEPT_TOKEN(aux_sym__format_raw_string_token1);
      if (lookahead == '\n') ADVANCE(139);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '{' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(137);
      END_STATE();
    case 138:
      ACCEPT_TOKEN(aux_sym__format_raw_string_token1);
      if (lookahead == '#') ADVANCE(137);
      if (lookahead == '\\') ADVANCE(135);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(138);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '{' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(139);
      END_STATE();
    case 139:
      ACCEPT_TOKEN(aux_sym__format_raw_string_token1);
      if (lookahead != 0 &&
          lookahead != '\'' &&
          lookahead != '{' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(139);
      END_STATE();
    case 140:
      ACCEPT_TOKEN(anon_sym_SQUOTE);
      END_STATE();
    case 141:
      ACCEPT_TOKEN(anon_sym_f_SQUOTE_SQUOTE_SQUOTE);
      END_STATE();
    case 142:
      ACCEPT_TOKEN(anon_sym_SQUOTE_SQUOTE_SQUOTE);
      END_STATE();
    case 143:
      ACCEPT_TOKEN(aux_sym__raw_string_indented_token1);
      END_STATE();
    case 144:
      ACCEPT_TOKEN(aux_sym__raw_string_indented_token1);
      if (lookahead == '#') ADVANCE(143);
      if (lookahead == '\'') ADVANCE(147);
      if (lookahead == '\\') ADVANCE(143);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(144);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(143);
      END_STATE();
    case 145:
      ACCEPT_TOKEN(aux_sym__raw_string_indented_token1);
      if (lookahead == '#') ADVANCE(143);
      if (lookahead == '\\') ADVANCE(143);
      if (lookahead == '`') ADVANCE(163);
      if (lookahead == '{') ADVANCE(149);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(145);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(143);
      END_STATE();
    case 146:
      ACCEPT_TOKEN(aux_sym__raw_string_indented_token1);
      if (lookahead == '#') ADVANCE(143);
      if (lookahead == '\\') ADVANCE(143);
      if (lookahead == '`') ADVANCE(148);
      if (lookahead == '{') ADVANCE(149);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(146);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(143);
      END_STATE();
    case 147:
      ACCEPT_TOKEN(aux_sym__raw_string_indented_token1);
      if (lookahead == '\'') ADVANCE(30);
      END_STATE();
    case 148:
      ACCEPT_TOKEN(aux_sym__raw_string_indented_token1);
      if (lookahead == '`') ADVANCE(38);
      END_STATE();
    case 149:
      ACCEPT_TOKEN(aux_sym__raw_string_indented_token1);
      if (lookahead == '{') ADVANCE(166);
      END_STATE();
    case 150:
      ACCEPT_TOKEN(aux_sym__string_token1);
      if (lookahead == '\n') ADVANCE(152);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '\\' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(150);
      END_STATE();
    case 151:
      ACCEPT_TOKEN(aux_sym__string_token1);
      if (lookahead == '#') ADVANCE(150);
      if (lookahead == '\\') ADVANCE(9);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(151);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '#' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(152);
      END_STATE();
    case 152:
      ACCEPT_TOKEN(aux_sym__string_token1);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '\\' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(152);
      END_STATE();
    case 153:
      ACCEPT_TOKEN(anon_sym_x_DQUOTE);
      if (lookahead == '"') ADVANCE(18);
      END_STATE();
    case 154:
      ACCEPT_TOKEN(anon_sym_x_DQUOTE_DQUOTE_DQUOTE);
      END_STATE();
    case 155:
      ACCEPT_TOKEN(aux_sym__shell_expanded_string_indented_token1);
      if (lookahead == '\n') ADVANCE(157);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '\\' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(155);
      END_STATE();
    case 156:
      ACCEPT_TOKEN(aux_sym__shell_expanded_string_indented_token1);
      if (lookahead == '"') ADVANCE(15);
      if (lookahead == '#') ADVANCE(155);
      if (lookahead == '\\') ADVANCE(12);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(156);
      if (lookahead != 0 &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(157);
      END_STATE();
    case 157:
      ACCEPT_TOKEN(aux_sym__shell_expanded_string_indented_token1);
      if (lookahead != 0 &&
          lookahead != '"' &&
          lookahead != '\\' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(157);
      END_STATE();
    case 158:
      ACCEPT_TOKEN(anon_sym_x_SQUOTE_SQUOTE_SQUOTE);
      END_STATE();
    case 159:
      ACCEPT_TOKEN(sym_escape_sequence);
      END_STATE();
    case 160:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (lookahead == '\\') ADVANCE(4);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(127);
      END_STATE();
    case 161:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (lookahead == '\\') ADVANCE(9);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(151);
      END_STATE();
    case 162:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (lookahead == '\\') ADVANCE(12);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(156);
      END_STATE();
    case 163:
      ACCEPT_TOKEN(anon_sym_BQUOTE);
      END_STATE();
    case 164:
      ACCEPT_TOKEN(anon_sym_BQUOTE);
      if (lookahead == '`') ADVANCE(38);
      END_STATE();
    case 165:
      ACCEPT_TOKEN(anon_sym_BQUOTE_BQUOTE_BQUOTE);
      END_STATE();
    case 166:
      ACCEPT_TOKEN(anon_sym_LBRACE_LBRACE);
      END_STATE();
    case 167:
      ACCEPT_TOKEN(anon_sym_RBRACE_RBRACE);
      END_STATE();
    case 168:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '"') ADVANCE(125);
      if (lookahead == '\'') ADVANCE(134);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(170);
      END_STATE();
    case 169:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '"') ADVANCE(153);
      if (lookahead == '\'') ADVANCE(26);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(170);
      END_STATE();
    case 170:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(170);
      END_STATE();
    case 171:
      ACCEPT_TOKEN(sym_numeric_error);
      if (lookahead == '.') ADVANCE(172);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(171);
      END_STATE();
    case 172:
      ACCEPT_TOKEN(sym_numeric_error);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(172);
      END_STATE();
    case 173:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '!') ADVANCE(109);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(175);
      END_STATE();
    case 174:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '/') ADVANCE(111);
      if (lookahead == '\t' ||
          (0x0b <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(175);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(174);
      END_STATE();
    case 175:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != 0x17f &&
          lookahead != 0x212a) ADVANCE(175);
      END_STATE();
    default:
      return false;
  }
}

static bool ts_lex_keywords(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (lookahead == '\\') SKIP(1);
      if (lookahead == 'a') ADVANCE(2);
      if (lookahead == 'e') ADVANCE(3);
      if (lookahead == 'f') ADVANCE(4);
      if (lookahead == 'i') ADVANCE(5);
      if (lookahead == 'm') ADVANCE(6);
      if (lookahead == 's') ADVANCE(7);
      if (lookahead == 't') ADVANCE(8);
      if (lookahead == 'u') ADVANCE(9);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(0);
      END_STATE();
    case 1:
      if (lookahead == '\n') SKIP(10);
      if (lookahead == '\r') SKIP(11);
      END_STATE();
    case 2:
      if (lookahead == 'l') ADVANCE(12);
      if (lookahead == 's') ADVANCE(13);
      END_STATE();
    case 3:
      if (lookahead == 'a') ADVANCE(14);
      if (lookahead == 'l') ADVANCE(15);
      if (lookahead == 'n') ADVANCE(16);
      if (lookahead == 'x') ADVANCE(17);
      END_STATE();
    case 4:
      if (lookahead == 'a') ADVANCE(18);
      END_STATE();
    case 5:
      if (lookahead == 'f') ADVANCE(19);
      if (lookahead == 'm') ADVANCE(20);
      END_STATE();
    case 6:
      if (lookahead == 'o') ADVANCE(21);
      END_STATE();
    case 7:
      if (lookahead == 'e') ADVANCE(22);
      END_STATE();
    case 8:
      if (lookahead == 'r') ADVANCE(23);
      END_STATE();
    case 9:
      if (lookahead == 'n') ADVANCE(24);
      END_STATE();
    case 10:
      if (lookahead == '\\') SKIP(1);
      if (lookahead == 'a') ADVANCE(2);
      if (lookahead == 'e') ADVANCE(3);
      if (lookahead == 'f') ADVANCE(4);
      if (lookahead == 'i') ADVANCE(5);
      if (lookahead == 'm') ADVANCE(6);
      if (lookahead == 's') ADVANCE(7);
      if (lookahead == 't') ADVANCE(8);
      if (lookahead == 'u') ADVANCE(9);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') SKIP(10);
      END_STATE();
    case 11:
      if (lookahead == '\n') SKIP(10);
      END_STATE();
    case 12:
      if (lookahead == 'i') ADVANCE(25);
      END_STATE();
    case 13:
      if (lookahead == 's') ADVANCE(26);
      END_STATE();
    case 14:
      if (lookahead == 'g') ADVANCE(27);
      END_STATE();
    case 15:
      if (lookahead == 's') ADVANCE(28);
      END_STATE();
    case 16:
      if (lookahead == 'v') ADVANCE(29);
      END_STATE();
    case 17:
      if (lookahead == 'p') ADVANCE(30);
      END_STATE();
    case 18:
      if (lookahead == 'l') ADVANCE(31);
      END_STATE();
    case 19:
      ACCEPT_TOKEN(anon_sym_if);
      END_STATE();
    case 20:
      if (lookahead == 'p') ADVANCE(32);
      END_STATE();
    case 21:
      if (lookahead == 'd') ADVANCE(33);
      END_STATE();
    case 22:
      if (lookahead == 't') ADVANCE(34);
      END_STATE();
    case 23:
      if (lookahead == 'u') ADVANCE(35);
      END_STATE();
    case 24:
      if (lookahead == 'e') ADVANCE(36);
      END_STATE();
    case 25:
      if (lookahead == 'a') ADVANCE(37);
      END_STATE();
    case 26:
      if (lookahead == 'e') ADVANCE(38);
      END_STATE();
    case 27:
      if (lookahead == 'e') ADVANCE(39);
      END_STATE();
    case 28:
      if (lookahead == 'e') ADVANCE(40);
      END_STATE();
    case 29:
      ACCEPT_TOKEN(anon_sym_env);
      END_STATE();
    case 30:
      if (lookahead == 'o') ADVANCE(41);
      END_STATE();
    case 31:
      if (lookahead == 's') ADVANCE(42);
      END_STATE();
    case 32:
      if (lookahead == 'o') ADVANCE(43);
      END_STATE();
    case 33:
      ACCEPT_TOKEN(anon_sym_mod);
      END_STATE();
    case 34:
      ACCEPT_TOKEN(anon_sym_set);
      END_STATE();
    case 35:
      if (lookahead == 'e') ADVANCE(44);
      END_STATE();
    case 36:
      if (lookahead == 'x') ADVANCE(45);
      END_STATE();
    case 37:
      if (lookahead == 's') ADVANCE(46);
      END_STATE();
    case 38:
      if (lookahead == 'r') ADVANCE(47);
      END_STATE();
    case 39:
      if (lookahead == 'r') ADVANCE(48);
      END_STATE();
    case 40:
      ACCEPT_TOKEN(anon_sym_else);
      END_STATE();
    case 41:
      if (lookahead == 'r') ADVANCE(49);
      END_STATE();
    case 42:
      if (lookahead == 'e') ADVANCE(50);
      END_STATE();
    case 43:
      if (lookahead == 'r') ADVANCE(51);
      END_STATE();
    case 44:
      ACCEPT_TOKEN(anon_sym_true);
      END_STATE();
    case 45:
      if (lookahead == 'p') ADVANCE(52);
      END_STATE();
    case 46:
      ACCEPT_TOKEN(anon_sym_alias);
      END_STATE();
    case 47:
      if (lookahead == 't') ADVANCE(53);
      END_STATE();
    case 48:
      ACCEPT_TOKEN(anon_sym_eager);
      END_STATE();
    case 49:
      if (lookahead == 't') ADVANCE(54);
      END_STATE();
    case 50:
      ACCEPT_TOKEN(anon_sym_false);
      END_STATE();
    case 51:
      if (lookahead == 't') ADVANCE(55);
      END_STATE();
    case 52:
      if (lookahead == 'o') ADVANCE(56);
      END_STATE();
    case 53:
      ACCEPT_TOKEN(anon_sym_assert);
      END_STATE();
    case 54:
      ACCEPT_TOKEN(anon_sym_export);
      END_STATE();
    case 55:
      ACCEPT_TOKEN(anon_sym_import);
      END_STATE();
    case 56:
      if (lookahead == 'r') ADVANCE(57);
      END_STATE();
    case 57:
      if (lookahead == 't') ADVANCE(58);
      END_STATE();
    case 58:
      ACCEPT_TOKEN(anon_sym_unexport);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexerMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0, .external_lex_state = 1},
  [1] = {.lex_state = 64},
  [2] = {.lex_state = 61},
  [3] = {.lex_state = 61},
  [4] = {.lex_state = 61},
  [5] = {.lex_state = 61},
  [6] = {.lex_state = 61},
  [7] = {.lex_state = 61},
  [8] = {.lex_state = 61},
  [9] = {.lex_state = 61},
  [10] = {.lex_state = 61},
  [11] = {.lex_state = 61},
  [12] = {.lex_state = 61},
  [13] = {.lex_state = 61},
  [14] = {.lex_state = 61},
  [15] = {.lex_state = 61},
  [16] = {.lex_state = 61},
  [17] = {.lex_state = 61},
  [18] = {.lex_state = 61},
  [19] = {.lex_state = 61},
  [20] = {.lex_state = 61},
  [21] = {.lex_state = 61},
  [22] = {.lex_state = 61},
  [23] = {.lex_state = 61},
  [24] = {.lex_state = 61},
  [25] = {.lex_state = 61},
  [26] = {.lex_state = 61},
  [27] = {.lex_state = 61},
  [28] = {.lex_state = 61},
  [29] = {.lex_state = 61},
  [30] = {.lex_state = 61},
  [31] = {.lex_state = 61},
  [32] = {.lex_state = 61},
  [33] = {.lex_state = 61},
  [34] = {.lex_state = 61},
  [35] = {.lex_state = 61},
  [36] = {.lex_state = 61},
  [37] = {.lex_state = 61},
  [38] = {.lex_state = 61},
  [39] = {.lex_state = 61},
  [40] = {.lex_state = 61},
  [41] = {.lex_state = 61},
  [42] = {.lex_state = 61},
  [43] = {.lex_state = 61},
  [44] = {.lex_state = 61},
  [45] = {.lex_state = 61},
  [46] = {.lex_state = 61},
  [47] = {.lex_state = 61},
  [48] = {.lex_state = 61},
  [49] = {.lex_state = 61},
  [50] = {.lex_state = 61},
  [51] = {.lex_state = 61},
  [52] = {.lex_state = 61},
  [53] = {.lex_state = 61},
  [54] = {.lex_state = 61},
  [55] = {.lex_state = 61},
  [56] = {.lex_state = 61},
  [57] = {.lex_state = 61},
  [58] = {.lex_state = 61},
  [59] = {.lex_state = 61},
  [60] = {.lex_state = 61},
  [61] = {.lex_state = 61},
  [62] = {.lex_state = 61},
  [63] = {.lex_state = 61},
  [64] = {.lex_state = 61},
  [65] = {.lex_state = 61},
  [66] = {.lex_state = 61},
  [67] = {.lex_state = 61},
  [68] = {.lex_state = 61},
  [69] = {.lex_state = 61},
  [70] = {.lex_state = 61},
  [71] = {.lex_state = 61},
  [72] = {.lex_state = 61},
  [73] = {.lex_state = 61},
  [74] = {.lex_state = 61},
  [75] = {.lex_state = 61},
  [76] = {.lex_state = 61},
  [77] = {.lex_state = 61},
  [78] = {.lex_state = 61},
  [79] = {.lex_state = 61},
  [80] = {.lex_state = 61},
  [81] = {.lex_state = 61},
  [82] = {.lex_state = 61},
  [83] = {.lex_state = 61},
  [84] = {.lex_state = 61},
  [85] = {.lex_state = 13},
  [86] = {.lex_state = 13},
  [87] = {.lex_state = 13},
  [88] = {.lex_state = 13},
  [89] = {.lex_state = 13},
  [90] = {.lex_state = 13},
  [91] = {.lex_state = 61},
  [92] = {.lex_state = 13},
  [93] = {.lex_state = 13},
  [94] = {.lex_state = 13},
  [95] = {.lex_state = 13},
  [96] = {.lex_state = 61},
  [97] = {.lex_state = 61},
  [98] = {.lex_state = 61},
  [99] = {.lex_state = 61},
  [100] = {.lex_state = 13},
  [101] = {.lex_state = 13},
  [102] = {.lex_state = 13},
  [103] = {.lex_state = 13},
  [104] = {.lex_state = 13},
  [105] = {.lex_state = 13},
  [106] = {.lex_state = 13},
  [107] = {.lex_state = 13},
  [108] = {.lex_state = 13},
  [109] = {.lex_state = 61},
  [110] = {.lex_state = 61},
  [111] = {.lex_state = 61},
  [112] = {.lex_state = 61},
  [113] = {.lex_state = 61},
  [114] = {.lex_state = 13},
  [115] = {.lex_state = 61},
  [116] = {.lex_state = 61},
  [117] = {.lex_state = 61},
  [118] = {.lex_state = 61},
  [119] = {.lex_state = 61},
  [120] = {.lex_state = 61},
  [121] = {.lex_state = 61},
  [122] = {.lex_state = 61},
  [123] = {.lex_state = 61},
  [124] = {.lex_state = 61},
  [125] = {.lex_state = 61},
  [126] = {.lex_state = 61},
  [127] = {.lex_state = 61},
  [128] = {.lex_state = 62},
  [129] = {.lex_state = 62},
  [130] = {.lex_state = 62},
  [131] = {.lex_state = 62},
  [132] = {.lex_state = 62},
  [133] = {.lex_state = 62},
  [134] = {.lex_state = 62},
  [135] = {.lex_state = 62},
  [136] = {.lex_state = 62},
  [137] = {.lex_state = 61},
  [138] = {.lex_state = 61},
  [139] = {.lex_state = 64, .external_lex_state = 2},
  [140] = {.lex_state = 61},
  [141] = {.lex_state = 61},
  [142] = {.lex_state = 61},
  [143] = {.lex_state = 61},
  [144] = {.lex_state = 13, .external_lex_state = 2},
  [145] = {.lex_state = 13, .external_lex_state = 2},
  [146] = {.lex_state = 13, .external_lex_state = 2},
  [147] = {.lex_state = 13},
  [148] = {.lex_state = 13, .external_lex_state = 2},
  [149] = {.lex_state = 61},
  [150] = {.lex_state = 13},
  [151] = {.lex_state = 62},
  [152] = {.lex_state = 13, .external_lex_state = 2},
  [153] = {.lex_state = 62},
  [154] = {.lex_state = 62},
  [155] = {.lex_state = 62},
  [156] = {.lex_state = 62},
  [157] = {.lex_state = 62},
  [158] = {.lex_state = 62},
  [159] = {.lex_state = 62},
  [160] = {.lex_state = 62},
  [161] = {.lex_state = 62},
  [162] = {.lex_state = 62},
  [163] = {.lex_state = 62},
  [164] = {.lex_state = 62},
  [165] = {.lex_state = 62},
  [166] = {.lex_state = 62},
  [167] = {.lex_state = 62},
  [168] = {.lex_state = 62},
  [169] = {.lex_state = 61},
  [170] = {.lex_state = 61},
  [171] = {.lex_state = 61},
  [172] = {.lex_state = 61},
  [173] = {.lex_state = 61},
  [174] = {.lex_state = 63},
  [175] = {.lex_state = 62, .external_lex_state = 3},
  [176] = {.lex_state = 61},
  [177] = {.lex_state = 63},
  [178] = {.lex_state = 61},
  [179] = {.lex_state = 62, .external_lex_state = 3},
  [180] = {.lex_state = 63},
  [181] = {.lex_state = 62, .external_lex_state = 4},
  [182] = {.lex_state = 63},
  [183] = {.lex_state = 63},
  [184] = {.lex_state = 63},
  [185] = {.lex_state = 62, .external_lex_state = 4},
  [186] = {.lex_state = 61},
  [187] = {.lex_state = 63},
  [188] = {.lex_state = 62, .external_lex_state = 3},
  [189] = {.lex_state = 62},
  [190] = {.lex_state = 63},
  [191] = {.lex_state = 62},
  [192] = {.lex_state = 62},
  [193] = {.lex_state = 62},
  [194] = {.lex_state = 62},
  [195] = {.lex_state = 62},
  [196] = {.lex_state = 62, .external_lex_state = 3},
  [197] = {.lex_state = 62},
  [198] = {.lex_state = 62},
  [199] = {.lex_state = 62},
  [200] = {.lex_state = 62},
  [201] = {.lex_state = 62},
  [202] = {.lex_state = 62},
  [203] = {.lex_state = 62},
  [204] = {.lex_state = 62},
  [205] = {.lex_state = 62},
  [206] = {.lex_state = 62},
  [207] = {.lex_state = 62},
  [208] = {.lex_state = 62},
  [209] = {.lex_state = 62, .external_lex_state = 3},
  [210] = {.lex_state = 62},
  [211] = {.lex_state = 62},
  [212] = {.lex_state = 62},
  [213] = {.lex_state = 62},
  [214] = {.lex_state = 62},
  [215] = {.lex_state = 63},
  [216] = {.lex_state = 63},
  [217] = {.lex_state = 63},
  [218] = {.lex_state = 13},
  [219] = {.lex_state = 62},
  [220] = {.lex_state = 62},
  [221] = {.lex_state = 62},
  [222] = {.lex_state = 62},
  [223] = {.lex_state = 62},
  [224] = {.lex_state = 62, .external_lex_state = 3},
  [225] = {.lex_state = 61, .external_lex_state = 3},
  [226] = {.lex_state = 62},
  [227] = {.lex_state = 62},
  [228] = {.lex_state = 62},
  [229] = {.lex_state = 62},
  [230] = {.lex_state = 61, .external_lex_state = 3},
  [231] = {.lex_state = 62},
  [232] = {.lex_state = 61, .external_lex_state = 3},
  [233] = {.lex_state = 61, .external_lex_state = 3},
  [234] = {.lex_state = 62},
  [235] = {.lex_state = 62, .external_lex_state = 3},
  [236] = {.lex_state = 62, .external_lex_state = 3},
  [237] = {.lex_state = 61, .external_lex_state = 3},
  [238] = {.lex_state = 61, .external_lex_state = 3},
  [239] = {.lex_state = 61, .external_lex_state = 3},
  [240] = {.lex_state = 61, .external_lex_state = 3},
  [241] = {.lex_state = 61, .external_lex_state = 3},
  [242] = {.lex_state = 61, .external_lex_state = 3},
  [243] = {.lex_state = 61, .external_lex_state = 3},
  [244] = {.lex_state = 61, .external_lex_state = 3},
  [245] = {.lex_state = 61, .external_lex_state = 3},
  [246] = {.lex_state = 62, .external_lex_state = 3},
  [247] = {.lex_state = 13},
  [248] = {.lex_state = 13},
  [249] = {.lex_state = 62, .external_lex_state = 3},
  [250] = {.lex_state = 13},
  [251] = {.lex_state = 62},
  [252] = {.lex_state = 62},
  [253] = {.lex_state = 61, .external_lex_state = 3},
  [254] = {.lex_state = 62, .external_lex_state = 3},
  [255] = {.lex_state = 61, .external_lex_state = 3},
  [256] = {.lex_state = 61, .external_lex_state = 3},
  [257] = {.lex_state = 13},
  [258] = {.lex_state = 13},
  [259] = {.lex_state = 13},
  [260] = {.lex_state = 61, .external_lex_state = 3},
  [261] = {.lex_state = 62},
  [262] = {.lex_state = 13},
  [263] = {.lex_state = 61, .external_lex_state = 3},
  [264] = {.lex_state = 61, .external_lex_state = 3},
  [265] = {.lex_state = 61, .external_lex_state = 3},
  [266] = {.lex_state = 13},
  [267] = {.lex_state = 13},
  [268] = {.lex_state = 61, .external_lex_state = 3},
  [269] = {.lex_state = 61, .external_lex_state = 3},
  [270] = {.lex_state = 61, .external_lex_state = 3},
  [271] = {.lex_state = 62, .external_lex_state = 3},
  [272] = {.lex_state = 61, .external_lex_state = 3},
  [273] = {.lex_state = 13},
  [274] = {.lex_state = 13},
  [275] = {.lex_state = 13},
  [276] = {.lex_state = 61, .external_lex_state = 3},
  [277] = {.lex_state = 62, .external_lex_state = 3},
  [278] = {.lex_state = 62},
  [279] = {.lex_state = 62},
  [280] = {.lex_state = 61, .external_lex_state = 3},
  [281] = {.lex_state = 61, .external_lex_state = 3},
  [282] = {.lex_state = 61, .external_lex_state = 3},
  [283] = {.lex_state = 61, .external_lex_state = 3},
  [284] = {.lex_state = 61, .external_lex_state = 3},
  [285] = {.lex_state = 61, .external_lex_state = 3},
  [286] = {.lex_state = 61, .external_lex_state = 3},
  [287] = {.lex_state = 61, .external_lex_state = 3},
  [288] = {.lex_state = 61, .external_lex_state = 3},
  [289] = {.lex_state = 62, .external_lex_state = 3},
  [290] = {.lex_state = 62},
  [291] = {.lex_state = 62},
  [292] = {.lex_state = 61, .external_lex_state = 3},
  [293] = {.lex_state = 62},
  [294] = {.lex_state = 62},
  [295] = {.lex_state = 62, .external_lex_state = 3},
  [296] = {.lex_state = 62, .external_lex_state = 3},
  [297] = {.lex_state = 62},
  [298] = {.lex_state = 62},
  [299] = {.lex_state = 22},
  [300] = {.lex_state = 63, .external_lex_state = 3},
  [301] = {.lex_state = 19},
  [302] = {.lex_state = 22},
  [303] = {.lex_state = 19},
  [304] = {.lex_state = 22},
  [305] = {.lex_state = 22},
  [306] = {.lex_state = 63, .external_lex_state = 3},
  [307] = {.lex_state = 62},
  [308] = {.lex_state = 19},
  [309] = {.lex_state = 22},
  [310] = {.lex_state = 19},
  [311] = {.lex_state = 22},
  [312] = {.lex_state = 63, .external_lex_state = 3},
  [313] = {.lex_state = 19},
  [314] = {.lex_state = 22},
  [315] = {.lex_state = 62},
  [316] = {.lex_state = 62},
  [317] = {.lex_state = 19},
  [318] = {.lex_state = 19},
  [319] = {.lex_state = 62},
  [320] = {.lex_state = 24},
  [321] = {.lex_state = 23},
  [322] = {.lex_state = 24},
  [323] = {.lex_state = 24},
  [324] = {.lex_state = 62},
  [325] = {.lex_state = 5},
  [326] = {.lex_state = 10},
  [327] = {.lex_state = 62},
  [328] = {.lex_state = 63, .external_lex_state = 3},
  [329] = {.lex_state = 23},
  [330] = {.lex_state = 24},
  [331] = {.lex_state = 23},
  [332] = {.lex_state = 24},
  [333] = {.lex_state = 5},
  [334] = {.lex_state = 5},
  [335] = {.lex_state = 10},
  [336] = {.lex_state = 10},
  [337] = {.lex_state = 23},
  [338] = {.lex_state = 62},
  [339] = {.lex_state = 23},
  [340] = {.lex_state = 24},
  [341] = {.lex_state = 62},
  [342] = {.lex_state = 24},
  [343] = {.lex_state = 23},
  [344] = {.lex_state = 23},
  [345] = {.lex_state = 62},
  [346] = {.lex_state = 62, .external_lex_state = 3},
  [347] = {.lex_state = 22},
  [348] = {.lex_state = 62},
  [349] = {.lex_state = 13},
  [350] = {.lex_state = 13, .external_lex_state = 5},
  [351] = {.lex_state = 10},
  [352] = {.lex_state = 10},
  [353] = {.lex_state = 5},
  [354] = {.lex_state = 62},
  [355] = {.lex_state = 62, .external_lex_state = 3},
  [356] = {.lex_state = 19},
  [357] = {.lex_state = 62, .external_lex_state = 3},
  [358] = {.lex_state = 62, .external_lex_state = 3},
  [359] = {.lex_state = 62, .external_lex_state = 3},
  [360] = {.lex_state = 62},
  [361] = {.lex_state = 5},
  [362] = {.lex_state = 62, .external_lex_state = 3},
  [363] = {.lex_state = 13, .external_lex_state = 5},
  [364] = {.lex_state = 62, .external_lex_state = 3},
  [365] = {.lex_state = 13, .external_lex_state = 5},
  [366] = {.lex_state = 20},
  [367] = {.lex_state = 13, .external_lex_state = 6},
  [368] = {.lex_state = 62},
  [369] = {.lex_state = 20},
  [370] = {.lex_state = 21},
  [371] = {.lex_state = 20},
  [372] = {.lex_state = 21},
  [373] = {.lex_state = 62},
  [374] = {.lex_state = 23},
  [375] = {.lex_state = 13},
  [376] = {.lex_state = 62},
  [377] = {.lex_state = 61},
  [378] = {.lex_state = 24},
  [379] = {.lex_state = 20},
  [380] = {.lex_state = 21},
  [381] = {.lex_state = 20},
  [382] = {.lex_state = 21},
  [383] = {.lex_state = 62},
  [384] = {.lex_state = 62},
  [385] = {.lex_state = 20},
  [386] = {.lex_state = 21},
  [387] = {.lex_state = 20},
  [388] = {.lex_state = 21},
  [389] = {.lex_state = 21},
  [390] = {.lex_state = 6},
  [391] = {.lex_state = 6},
  [392] = {.lex_state = 61},
  [393] = {.lex_state = 61},
  [394] = {.lex_state = 62},
  [395] = {.lex_state = 61},
  [396] = {.lex_state = 61},
  [397] = {.lex_state = 61},
  [398] = {.lex_state = 61},
  [399] = {.lex_state = 61},
  [400] = {.lex_state = 61},
  [401] = {.lex_state = 62},
  [402] = {.lex_state = 61},
  [403] = {.lex_state = 62},
  [404] = {.lex_state = 61},
  [405] = {.lex_state = 61},
  [406] = {.lex_state = 61},
  [407] = {.lex_state = 5},
  [408] = {.lex_state = 10},
  [409] = {.lex_state = 13, .external_lex_state = 5},
  [410] = {.lex_state = 62},
  [411] = {.lex_state = 61},
  [412] = {.lex_state = 61},
  [413] = {.lex_state = 6},
  [414] = {.lex_state = 7},
  [415] = {.lex_state = 61},
  [416] = {.lex_state = 6},
  [417] = {.lex_state = 61},
  [418] = {.lex_state = 61},
  [419] = {.lex_state = 61},
  [420] = {.lex_state = 62},
  [421] = {.lex_state = 61},
  [422] = {.lex_state = 61},
  [423] = {.lex_state = 6},
  [424] = {.lex_state = 6},
  [425] = {.lex_state = 62},
  [426] = {.lex_state = 62},
  [427] = {.lex_state = 6},
  [428] = {.lex_state = 61},
  [429] = {.lex_state = 62},
  [430] = {.lex_state = 62},
  [431] = {.lex_state = 61},
  [432] = {.lex_state = 61},
  [433] = {.lex_state = 61},
  [434] = {.lex_state = 61},
  [435] = {.lex_state = 61},
  [436] = {.lex_state = 61},
  [437] = {.lex_state = 61},
  [438] = {.lex_state = 61},
  [439] = {.lex_state = 61},
  [440] = {.lex_state = 62},
  [441] = {.lex_state = 61},
  [442] = {.lex_state = 61},
  [443] = {.lex_state = 61, .external_lex_state = 3},
  [444] = {.lex_state = 62},
  [445] = {.lex_state = 61},
  [446] = {.lex_state = 61},
  [447] = {.lex_state = 61},
  [448] = {.lex_state = 61},
  [449] = {.lex_state = 61},
  [450] = {.lex_state = 62},
  [451] = {.lex_state = 62},
  [452] = {.lex_state = 61},
  [453] = {.lex_state = 61},
  [454] = {.lex_state = 61},
  [455] = {.lex_state = 62},
  [456] = {.lex_state = 62},
  [457] = {.lex_state = 13, .external_lex_state = 6},
  [458] = {.lex_state = 61, .external_lex_state = 3},
  [459] = {.lex_state = 62},
  [460] = {.lex_state = 62},
  [461] = {.lex_state = 62},
  [462] = {.lex_state = 62},
  [463] = {.lex_state = 61, .external_lex_state = 3},
  [464] = {.lex_state = 61, .external_lex_state = 3},
  [465] = {.lex_state = 61},
  [466] = {.lex_state = 61, .external_lex_state = 3},
  [467] = {.lex_state = 114},
  [468] = {.lex_state = 61},
  [469] = {.lex_state = 61},
  [470] = {.lex_state = 61, .external_lex_state = 3},
  [471] = {.lex_state = 61, .external_lex_state = 3},
  [472] = {.lex_state = 61},
  [473] = {.lex_state = 61, .external_lex_state = 3},
  [474] = {.lex_state = 13},
  [475] = {.lex_state = 61, .external_lex_state = 3},
  [476] = {.lex_state = 13},
  [477] = {.lex_state = 61, .external_lex_state = 3},
  [478] = {.lex_state = 61, .external_lex_state = 3},
  [479] = {.lex_state = 61, .external_lex_state = 3},
  [480] = {.lex_state = 61, .external_lex_state = 3},
  [481] = {.lex_state = 61, .external_lex_state = 3},
  [482] = {.lex_state = 62},
  [483] = {.lex_state = 61},
  [484] = {.lex_state = 61},
  [485] = {.lex_state = 13},
  [486] = {.lex_state = 61, .external_lex_state = 3},
  [487] = {.lex_state = 61},
  [488] = {.lex_state = 62},
  [489] = {.lex_state = 61},
  [490] = {.lex_state = 61},
  [491] = {.lex_state = 61},
  [492] = {.lex_state = 61},
  [493] = {.lex_state = 62},
  [494] = {.lex_state = 61},
  [495] = {.lex_state = 61},
  [496] = {.lex_state = 61, .external_lex_state = 3},
  [497] = {.lex_state = 61},
  [498] = {.lex_state = 13},
  [499] = {.lex_state = 61, .external_lex_state = 3},
  [500] = {.lex_state = 62},
  [501] = {.lex_state = 61, .external_lex_state = 3},
  [502] = {.lex_state = 61},
  [503] = {.lex_state = 61, .external_lex_state = 3},
  [504] = {.lex_state = 13},
  [505] = {.lex_state = 62},
  [506] = {.lex_state = 13},
  [507] = {.lex_state = 61, .external_lex_state = 3},
  [508] = {.lex_state = 61, .external_lex_state = 3},
  [509] = {.lex_state = 61, .external_lex_state = 3},
  [510] = {.lex_state = 13},
  [511] = {.lex_state = 62},
  [512] = {.lex_state = 61},
  [513] = {.lex_state = 62},
  [514] = {.lex_state = 61},
  [515] = {.lex_state = 61, .external_lex_state = 3},
  [516] = {.lex_state = 61, .external_lex_state = 3},
  [517] = {.lex_state = 114},
  [518] = {.lex_state = 61, .external_lex_state = 3},
  [519] = {.lex_state = 61},
  [520] = {.lex_state = 61, .external_lex_state = 3},
  [521] = {.lex_state = 61, .external_lex_state = 3},
  [522] = {.lex_state = 62},
  [523] = {.lex_state = 62},
  [524] = {.lex_state = 62},
  [525] = {.lex_state = 61},
  [526] = {.lex_state = 61},
  [527] = {.lex_state = 61, .external_lex_state = 3},
  [528] = {.lex_state = 61},
  [529] = {.lex_state = 62},
  [530] = {.lex_state = 61},
  [531] = {.lex_state = 13},
  [532] = {.lex_state = 61, .external_lex_state = 3},
  [533] = {.lex_state = 61},
  [534] = {.lex_state = 13},
  [535] = {.lex_state = 62},
  [536] = {.lex_state = 61, .external_lex_state = 3},
  [537] = {.lex_state = 62},
  [538] = {.lex_state = 61},
  [539] = {.lex_state = 61},
  [540] = {.lex_state = 61},
  [541] = {.lex_state = 61},
  [542] = {.lex_state = 61},
  [543] = {.lex_state = 61},
  [544] = {.lex_state = 61},
  [545] = {.lex_state = 13},
  [546] = {.lex_state = 61},
  [547] = {.lex_state = 61, .external_lex_state = 3},
  [548] = {.lex_state = 114},
  [549] = {.lex_state = 61},
  [550] = {.lex_state = 62},
  [551] = {.lex_state = 61, .external_lex_state = 3},
  [552] = {.lex_state = 61, .external_lex_state = 3},
  [553] = {.lex_state = 61, .external_lex_state = 3},
  [554] = {.lex_state = 61, .external_lex_state = 3},
  [555] = {.lex_state = 61, .external_lex_state = 3},
  [556] = {.lex_state = 61, .external_lex_state = 3},
  [557] = {.lex_state = 61},
  [558] = {.lex_state = 62},
  [559] = {.lex_state = 62},
  [560] = {.lex_state = 61, .external_lex_state = 3},
  [561] = {.lex_state = 61},
  [562] = {.lex_state = 62},
  [563] = {.lex_state = 61},
  [564] = {.lex_state = 61},
  [565] = {.lex_state = 62},
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [STATE(0)] = {
    [ts_builtin_sym_end] = ACTIONS(1),
    [sym_identifier] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [anon_sym_COLON_EQ] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [anon_sym_alias] = ACTIONS(1),
    [anon_sym_COLON_COLON] = ACTIONS(1),
    [anon_sym_unexport] = ACTIONS(1),
    [anon_sym_eager] = ACTIONS(1),
    [anon_sym_export] = ACTIONS(1),
    [anon_sym_import] = ACTIONS(1),
    [anon_sym_QMARK] = ACTIONS(1),
    [anon_sym_mod] = ACTIONS(1),
    [anon_sym_set] = ACTIONS(1),
    [anon_sym_true] = ACTIONS(1),
    [anon_sym_false] = ACTIONS(1),
    [anon_sym_SLASH] = ACTIONS(1),
    [anon_sym_PLUS_PLUS] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_EQ_EQ] = ACTIONS(1),
    [anon_sym_BANG_EQ] = ACTIONS(1),
    [anon_sym_EQ_TILDE] = ACTIONS(1),
    [anon_sym_BANG_TILDE] = ACTIONS(1),
    [anon_sym_AMP_AMP] = ACTIONS(1),
    [anon_sym_PIPE_PIPE] = ACTIONS(1),
    [anon_sym_if] = ACTIONS(1),
    [anon_sym_else] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [anon_sym_BANG] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [anon_sym_assert] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [anon_sym_AT] = ACTIONS(1),
    [anon_sym_DOLLAR] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [anon_sym_AT_DASH_QMARK] = ACTIONS(1),
    [anon_sym_AT_QMARK_DASH] = ACTIONS(1),
    [anon_sym_DASH_AT_QMARK] = ACTIONS(1),
    [anon_sym_DASH_QMARK_AT] = ACTIONS(1),
    [anon_sym_QMARK_AT_DASH] = ACTIONS(1),
    [anon_sym_QMARK_DASH_AT] = ACTIONS(1),
    [anon_sym_AT_DASH] = ACTIONS(1),
    [anon_sym_AT_QMARK] = ACTIONS(1),
    [anon_sym_DASH_AT] = ACTIONS(1),
    [anon_sym_DASH_QMARK] = ACTIONS(1),
    [anon_sym_QMARK_AT] = ACTIONS(1),
    [anon_sym_QMARK_DASH] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [aux_sym_shebang_token1] = ACTIONS(1),
    [anon_sym_env] = ACTIONS(1),
    [aux_sym_string_token1] = ACTIONS(1),
    [anon_sym_f_DQUOTE] = ACTIONS(1),
    [aux_sym__format_string_token2] = ACTIONS(1),
    [anon_sym_DQUOTE] = ACTIONS(1),
    [anon_sym_f_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1),
    [anon_sym_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1),
    [anon_sym_f_SQUOTE] = ACTIONS(1),
    [anon_sym_SQUOTE] = ACTIONS(1),
    [anon_sym_f_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1),
    [anon_sym_x_DQUOTE] = ACTIONS(1),
    [anon_sym_x_DQUOTE_DQUOTE_DQUOTE] = ACTIONS(1),
    [anon_sym_x_SQUOTE_SQUOTE_SQUOTE] = ACTIONS(1),
    [anon_sym_BQUOTE] = ACTIONS(1),
    [anon_sym_BQUOTE_BQUOTE_BQUOTE] = ACTIONS(1),
    [anon_sym_LBRACE_LBRACE] = ACTIONS(1),
    [sym_numeric_error] = ACTIONS(1),
    [sym_comment] = ACTIONS(3),
    [sym__indent] = ACTIONS(1),
    [sym__dedent] = ACTIONS(1),
    [sym__newline] = ACTIONS(1),
    [sym_text] = ACTIONS(1),
    [sym_error_recovery] = ACTIONS(1),
  },
  [STATE(1)] = {
    [sym_source_file] = STATE(483),
    [sym__item] = STATE(130),
    [sym_function_definition] = STATE(130),
    [sym_alias] = STATE(130),
    [sym__name] = STATE(279),
    [sym_assignment] = STATE(130),
    [sym__assignment] = STATE(207),
    [sym_eager] = STATE(130),
    [sym_export] = STATE(130),
    [sym_unexport] = STATE(130),
    [sym_import] = STATE(130),
    [sym_module] = STATE(130),
    [sym_setting] = STATE(130),
    [sym_attribute] = STATE(168),
    [sym_recipe] = STATE(130),
    [sym_recipe_header] = STATE(507),
    [sym_shebang] = STATE(554),
    [aux_sym_source_file_repeat1] = STATE(130),
    [aux_sym_function_definition_repeat1] = STATE(168),
    [ts_builtin_sym_end] = ACTIONS(5),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_alias] = ACTIONS(9),
    [anon_sym_unexport] = ACTIONS(11),
    [anon_sym_eager] = ACTIONS(13),
    [anon_sym_export] = ACTIONS(15),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_mod] = ACTIONS(19),
    [anon_sym_set] = ACTIONS(21),
    [anon_sym_LBRACK] = ACTIONS(23),
    [anon_sym_AT] = ACTIONS(25),
    [aux_sym_shebang_token1] = ACTIONS(27),
    [sym_comment] = ACTIONS(3),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 29,
    ACTIONS(29), 1,
      sym_identifier,
    ACTIONS(31), 1,
      anon_sym_LPAREN,
    ACTIONS(33), 1,
      anon_sym_RPAREN,
    ACTIONS(35), 1,
      anon_sym_COLON_COLON,
    ACTIONS(37), 1,
      anon_sym_SLASH,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(41), 1,
      anon_sym_BANG,
    ACTIONS(43), 1,
      anon_sym_LBRACK,
    ACTIONS(45), 1,
      anon_sym_assert,
    ACTIONS(47), 1,
      anon_sym_STAR,
    ACTIONS(51), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(55), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(59), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(61), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(69), 1,
      sym_numeric_error,
    ACTIONS(71), 1,
      sym_comment,
    STATE(85), 1,
      sym_format_string,
    STATE(115), 1,
      sym__expression_inner,
    STATE(137), 1,
      aux_sym_module_path_repeat1,
    ACTIONS(49), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(53), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(57), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(63), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(10), 3,
      sym_expression,
      sym_starred_dependency_argument,
      aux_sym_dependency_expression_repeat1,
    STATE(88), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [101] = 29,
    ACTIONS(29), 1,
      sym_identifier,
    ACTIONS(31), 1,
      anon_sym_LPAREN,
    ACTIONS(35), 1,
      anon_sym_COLON_COLON,
    ACTIONS(37), 1,
      anon_sym_SLASH,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(41), 1,
      anon_sym_BANG,
    ACTIONS(43), 1,
      anon_sym_LBRACK,
    ACTIONS(45), 1,
      anon_sym_assert,
    ACTIONS(47), 1,
      anon_sym_STAR,
    ACTIONS(51), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(55), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(59), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(61), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(69), 1,
      sym_numeric_error,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(73), 1,
      anon_sym_RPAREN,
    STATE(85), 1,
      sym_format_string,
    STATE(115), 1,
      sym__expression_inner,
    STATE(137), 1,
      aux_sym_module_path_repeat1,
    ACTIONS(49), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(53), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(57), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(63), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(6), 3,
      sym_expression,
      sym_starred_dependency_argument,
      aux_sym_dependency_expression_repeat1,
    STATE(88), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [202] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(79), 1,
      anon_sym_else,
    STATE(5), 1,
      aux_sym_if_expression_repeat1,
    STATE(17), 1,
      sym_else_if_clause,
    STATE(34), 1,
      sym_else_clause,
    ACTIONS(75), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(77), 24,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      anon_sym_RBRACE_RBRACE,
      sym_numeric_error,
  [258] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(79), 1,
      anon_sym_else,
    STATE(11), 1,
      aux_sym_if_expression_repeat1,
    STATE(17), 1,
      sym_else_if_clause,
    STATE(38), 1,
      sym_else_clause,
    ACTIONS(81), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(83), 24,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      anon_sym_RBRACE_RBRACE,
      sym_numeric_error,
  [314] = 27,
    ACTIONS(29), 1,
      sym_identifier,
    ACTIONS(31), 1,
      anon_sym_LPAREN,
    ACTIONS(37), 1,
      anon_sym_SLASH,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(41), 1,
      anon_sym_BANG,
    ACTIONS(43), 1,
      anon_sym_LBRACK,
    ACTIONS(45), 1,
      anon_sym_assert,
    ACTIONS(47), 1,
      anon_sym_STAR,
    ACTIONS(51), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(55), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(59), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(61), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(69), 1,
      sym_numeric_error,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(85), 1,
      anon_sym_RPAREN,
    STATE(85), 1,
      sym_format_string,
    STATE(115), 1,
      sym__expression_inner,
    ACTIONS(49), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(53), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(57), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(63), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(9), 3,
      sym_expression,
      sym_starred_dependency_argument,
      aux_sym_dependency_expression_repeat1,
    STATE(88), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [409] = 27,
    ACTIONS(29), 1,
      sym_identifier,
    ACTIONS(31), 1,
      anon_sym_LPAREN,
    ACTIONS(37), 1,
      anon_sym_SLASH,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(41), 1,
      anon_sym_BANG,
    ACTIONS(43), 1,
      anon_sym_LBRACK,
    ACTIONS(45), 1,
      anon_sym_assert,
    ACTIONS(47), 1,
      anon_sym_STAR,
    ACTIONS(51), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(55), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(59), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(61), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(69), 1,
      sym_numeric_error,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(73), 1,
      anon_sym_RPAREN,
    STATE(85), 1,
      sym_format_string,
    STATE(115), 1,
      sym__expression_inner,
    ACTIONS(49), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(53), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(57), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(63), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(6), 3,
      sym_expression,
      sym_starred_dependency_argument,
      aux_sym_dependency_expression_repeat1,
    STATE(88), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [504] = 27,
    ACTIONS(29), 1,
      sym_identifier,
    ACTIONS(31), 1,
      anon_sym_LPAREN,
    ACTIONS(33), 1,
      anon_sym_RPAREN,
    ACTIONS(37), 1,
      anon_sym_SLASH,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(41), 1,
      anon_sym_BANG,
    ACTIONS(43), 1,
      anon_sym_LBRACK,
    ACTIONS(45), 1,
      anon_sym_assert,
    ACTIONS(47), 1,
      anon_sym_STAR,
    ACTIONS(51), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(55), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(59), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(61), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(69), 1,
      sym_numeric_error,
    ACTIONS(71), 1,
      sym_comment,
    STATE(85), 1,
      sym_format_string,
    STATE(115), 1,
      sym__expression_inner,
    ACTIONS(49), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(53), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(57), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(63), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(10), 3,
      sym_expression,
      sym_starred_dependency_argument,
      aux_sym_dependency_expression_repeat1,
    STATE(88), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [599] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(87), 1,
      sym_identifier,
    ACTIONS(90), 1,
      anon_sym_LPAREN,
    ACTIONS(93), 1,
      anon_sym_RPAREN,
    ACTIONS(95), 1,
      anon_sym_SLASH,
    ACTIONS(98), 1,
      anon_sym_if,
    ACTIONS(101), 1,
      anon_sym_BANG,
    ACTIONS(104), 1,
      anon_sym_LBRACK,
    ACTIONS(107), 1,
      anon_sym_assert,
    ACTIONS(110), 1,
      anon_sym_STAR,
    ACTIONS(116), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(122), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(128), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(131), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(137), 1,
      anon_sym_BQUOTE,
    ACTIONS(140), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(143), 1,
      sym_numeric_error,
    STATE(85), 1,
      sym_format_string,
    STATE(115), 1,
      sym__expression_inner,
    ACTIONS(113), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(119), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(125), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(134), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(9), 3,
      sym_expression,
      sym_starred_dependency_argument,
      aux_sym_dependency_expression_repeat1,
    STATE(88), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [694] = 27,
    ACTIONS(29), 1,
      sym_identifier,
    ACTIONS(31), 1,
      anon_sym_LPAREN,
    ACTIONS(37), 1,
      anon_sym_SLASH,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(41), 1,
      anon_sym_BANG,
    ACTIONS(43), 1,
      anon_sym_LBRACK,
    ACTIONS(45), 1,
      anon_sym_assert,
    ACTIONS(47), 1,
      anon_sym_STAR,
    ACTIONS(51), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(55), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(59), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(61), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(69), 1,
      sym_numeric_error,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(146), 1,
      anon_sym_RPAREN,
    STATE(85), 1,
      sym_format_string,
    STATE(115), 1,
      sym__expression_inner,
    ACTIONS(49), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(53), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(57), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(63), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(9), 3,
      sym_expression,
      sym_starred_dependency_argument,
      aux_sym_dependency_expression_repeat1,
    STATE(88), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [789] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(152), 1,
      anon_sym_else,
    STATE(11), 1,
      aux_sym_if_expression_repeat1,
    STATE(17), 1,
      sym_else_if_clause,
    ACTIONS(148), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(150), 24,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      anon_sym_RBRACE_RBRACE,
      sym_numeric_error,
  [842] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(155), 1,
      sym_identifier,
    ACTIONS(157), 1,
      anon_sym_LPAREN,
    ACTIONS(161), 1,
      anon_sym_SLASH,
    ACTIONS(163), 1,
      anon_sym_if,
    ACTIONS(165), 1,
      anon_sym_BANG,
    ACTIONS(167), 1,
      anon_sym_LBRACK,
    ACTIONS(169), 1,
      anon_sym_assert,
    ACTIONS(173), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(177), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(181), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(183), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(187), 1,
      anon_sym_BQUOTE,
    ACTIONS(189), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(191), 1,
      sym_numeric_error,
    STATE(232), 1,
      sym_string,
    STATE(243), 1,
      sym_format_string,
    STATE(284), 1,
      sym__expression_inner,
    ACTIONS(159), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(171), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(175), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(179), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(185), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(285), 2,
      sym_if_expression,
      sym_value,
    STATE(287), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(479), 2,
      sym_boolean,
      sym_expression,
    STATE(282), 5,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
  [936] = 27,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(155), 1,
      sym_identifier,
    ACTIONS(157), 1,
      anon_sym_LPAREN,
    ACTIONS(161), 1,
      anon_sym_SLASH,
    ACTIONS(163), 1,
      anon_sym_if,
    ACTIONS(165), 1,
      anon_sym_BANG,
    ACTIONS(167), 1,
      anon_sym_LBRACK,
    ACTIONS(169), 1,
      anon_sym_assert,
    ACTIONS(173), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(177), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(181), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(183), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(187), 1,
      anon_sym_BQUOTE,
    ACTIONS(189), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(191), 1,
      sym_numeric_error,
    STATE(243), 1,
      sym_format_string,
    STATE(276), 1,
      sym_string,
    STATE(284), 1,
      sym__expression_inner,
    ACTIONS(159), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(171), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(175), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(179), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(185), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(285), 2,
      sym_if_expression,
      sym_value,
    STATE(287), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(481), 2,
      sym_boolean,
      sym_expression,
    STATE(282), 5,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
  [1030] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(193), 13,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_else,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(195), 24,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      anon_sym_RBRACE_RBRACE,
      sym_numeric_error,
  [1075] = 27,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(201), 1,
      anon_sym_RPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(422), 1,
      sym_expression,
    STATE(549), 1,
      sym_sequence,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [1168] = 27,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    ACTIONS(233), 1,
      anon_sym_RBRACK,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(419), 1,
      sym_expression,
    STATE(495), 1,
      sym_list_elements,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [1261] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(235), 13,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_else,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(237), 24,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      anon_sym_RBRACE_RBRACE,
      sym_numeric_error,
  [1306] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(239), 13,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_else,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(241), 24,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      anon_sym_RBRACE_RBRACE,
      sym_numeric_error,
  [1351] = 27,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    ACTIONS(243), 1,
      anon_sym_RPAREN,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(422), 1,
      sym_expression,
    STATE(497), 1,
      sym_sequence,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [1444] = 27,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    ACTIONS(245), 1,
      anon_sym_RPAREN,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(422), 1,
      sym_expression,
    STATE(533), 1,
      sym_sequence,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [1537] = 27,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    ACTIONS(247), 1,
      anon_sym_RBRACK,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(419), 1,
      sym_expression,
    STATE(492), 1,
      sym_list_elements,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [1630] = 27,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    ACTIONS(249), 1,
      anon_sym_RBRACK,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(419), 1,
      sym_expression,
    STATE(528), 1,
      sym_list_elements,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [1723] = 26,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    ACTIONS(251), 1,
      anon_sym_RBRACK,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(402), 1,
      sym_expression,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [1813] = 25,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    ACTIONS(253), 1,
      sym_identifier,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(399), 2,
      sym_expression,
      sym_attribute_named_param,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [1901] = 26,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    ACTIONS(255), 1,
      anon_sym_RBRACK,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(402), 1,
      sym_expression,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [1991] = 26,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    ACTIONS(257), 1,
      anon_sym_RPAREN,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(402), 1,
      sym_expression,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [2081] = 26,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(412), 1,
      sym_expression,
    STATE(448), 1,
      sym_condition,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [2171] = 26,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    ACTIONS(259), 1,
      anon_sym_RPAREN,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(402), 1,
      sym_expression,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [2261] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(261), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(263), 24,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      anon_sym_RBRACE_RBRACE,
      sym_numeric_error,
  [2305] = 25,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    ACTIONS(253), 1,
      sym_identifier,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(392), 2,
      sym_expression,
      sym_attribute_named_param,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [2393] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(265), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(267), 24,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      anon_sym_RBRACE_RBRACE,
      sym_numeric_error,
  [2437] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(269), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(271), 24,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      anon_sym_RBRACE_RBRACE,
      sym_numeric_error,
  [2481] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(273), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(275), 24,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      anon_sym_RBRACE_RBRACE,
      sym_numeric_error,
  [2525] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(277), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(279), 24,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      anon_sym_RBRACE_RBRACE,
      sym_numeric_error,
  [2569] = 26,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(412), 1,
      sym_expression,
    STATE(438), 1,
      sym_condition,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [2659] = 26,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(412), 1,
      sym_expression,
    STATE(439), 1,
      sym_condition,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [2749] = 25,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    ACTIONS(253), 1,
      sym_identifier,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(452), 2,
      sym_expression,
      sym_attribute_named_param,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [2837] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(281), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(283), 24,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      anon_sym_RBRACE_RBRACE,
      sym_numeric_error,
  [2881] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(285), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(287), 24,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      anon_sym_RBRACE_RBRACE,
      sym_numeric_error,
  [2925] = 26,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(412), 1,
      sym_expression,
    STATE(433), 1,
      sym_condition,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [3015] = 26,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(412), 1,
      sym_expression,
    STATE(434), 1,
      sym_condition,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [3105] = 26,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(412), 1,
      sym_expression,
    STATE(437), 1,
      sym_condition,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [3195] = 26,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(412), 1,
      sym_expression,
    STATE(449), 1,
      sym_condition,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [3285] = 26,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(412), 1,
      sym_expression,
    STATE(432), 1,
      sym_condition,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [3375] = 26,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(412), 1,
      sym_expression,
    STATE(454), 1,
      sym_condition,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [3465] = 25,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(526), 1,
      sym_expression,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [3552] = 25,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(441), 1,
      sym_expression,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [3639] = 25,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(442), 1,
      sym_expression,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [3726] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(155), 1,
      sym_identifier,
    ACTIONS(157), 1,
      anon_sym_LPAREN,
    ACTIONS(161), 1,
      anon_sym_SLASH,
    ACTIONS(163), 1,
      anon_sym_if,
    ACTIONS(165), 1,
      anon_sym_BANG,
    ACTIONS(167), 1,
      anon_sym_LBRACK,
    ACTIONS(169), 1,
      anon_sym_assert,
    ACTIONS(173), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(177), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(181), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(183), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(187), 1,
      anon_sym_BQUOTE,
    ACTIONS(189), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(191), 1,
      sym_numeric_error,
    STATE(243), 1,
      sym_format_string,
    STATE(284), 1,
      sym__expression_inner,
    STATE(516), 1,
      sym_expression,
    ACTIONS(171), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(175), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(179), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(185), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(285), 2,
      sym_if_expression,
      sym_value,
    STATE(287), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(282), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [3813] = 25,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(561), 1,
      sym_expression,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [3900] = 25,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(402), 1,
      sym_expression,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [3987] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(155), 1,
      sym_identifier,
    ACTIONS(157), 1,
      anon_sym_LPAREN,
    ACTIONS(161), 1,
      anon_sym_SLASH,
    ACTIONS(163), 1,
      anon_sym_if,
    ACTIONS(165), 1,
      anon_sym_BANG,
    ACTIONS(167), 1,
      anon_sym_LBRACK,
    ACTIONS(169), 1,
      anon_sym_assert,
    ACTIONS(173), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(177), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(181), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(183), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(187), 1,
      anon_sym_BQUOTE,
    ACTIONS(189), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(191), 1,
      sym_numeric_error,
    STATE(243), 1,
      sym_format_string,
    STATE(284), 1,
      sym__expression_inner,
    STATE(527), 1,
      sym_expression,
    ACTIONS(171), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(175), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(179), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(185), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(285), 2,
      sym_if_expression,
      sym_value,
    STATE(287), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(282), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [4074] = 25,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(446), 1,
      sym_expression,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [4161] = 25,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(542), 1,
      sym_expression,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [4248] = 25,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(489), 1,
      sym_expression,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [4335] = 25,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(490), 1,
      sym_expression,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [4422] = 25,
    ACTIONS(31), 1,
      anon_sym_LPAREN,
    ACTIONS(43), 1,
      anon_sym_LBRACK,
    ACTIONS(45), 1,
      anon_sym_assert,
    ACTIONS(51), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(55), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(59), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(61), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(69), 1,
      sym_numeric_error,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(289), 1,
      sym_identifier,
    ACTIONS(291), 1,
      anon_sym_SLASH,
    ACTIONS(293), 1,
      anon_sym_if,
    ACTIONS(295), 1,
      anon_sym_BANG,
    STATE(85), 1,
      sym_format_string,
    STATE(247), 1,
      sym__expression_inner,
    STATE(531), 1,
      sym_expression,
    ACTIONS(49), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(53), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(57), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(63), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(248), 2,
      sym_if_expression,
      sym_value,
    STATE(88), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [4509] = 25,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(395), 1,
      sym_expression,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [4596] = 25,
    ACTIONS(31), 1,
      anon_sym_LPAREN,
    ACTIONS(43), 1,
      anon_sym_LBRACK,
    ACTIONS(45), 1,
      anon_sym_assert,
    ACTIONS(51), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(55), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(59), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(61), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(69), 1,
      sym_numeric_error,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(289), 1,
      sym_identifier,
    ACTIONS(291), 1,
      anon_sym_SLASH,
    ACTIONS(293), 1,
      anon_sym_if,
    ACTIONS(295), 1,
      anon_sym_BANG,
    STATE(85), 1,
      sym_format_string,
    STATE(247), 1,
      sym__expression_inner,
    STATE(498), 1,
      sym_expression,
    ACTIONS(49), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(53), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(57), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(63), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(248), 2,
      sym_if_expression,
      sym_value,
    STATE(88), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [4683] = 25,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(436), 1,
      sym_expression,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [4770] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(155), 1,
      sym_identifier,
    ACTIONS(157), 1,
      anon_sym_LPAREN,
    ACTIONS(161), 1,
      anon_sym_SLASH,
    ACTIONS(163), 1,
      anon_sym_if,
    ACTIONS(165), 1,
      anon_sym_BANG,
    ACTIONS(167), 1,
      anon_sym_LBRACK,
    ACTIONS(169), 1,
      anon_sym_assert,
    ACTIONS(173), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(177), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(181), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(183), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(187), 1,
      anon_sym_BQUOTE,
    ACTIONS(189), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(191), 1,
      sym_numeric_error,
    STATE(243), 1,
      sym_format_string,
    STATE(284), 1,
      sym__expression_inner,
    STATE(471), 1,
      sym_expression,
    ACTIONS(171), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(175), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(179), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(185), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(285), 2,
      sym_if_expression,
      sym_value,
    STATE(287), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(282), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [4857] = 25,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(525), 1,
      sym_expression,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [4944] = 25,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(541), 1,
      sym_expression,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [5031] = 25,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(469), 1,
      sym_expression,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [5118] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(155), 1,
      sym_identifier,
    ACTIONS(157), 1,
      anon_sym_LPAREN,
    ACTIONS(161), 1,
      anon_sym_SLASH,
    ACTIONS(163), 1,
      anon_sym_if,
    ACTIONS(165), 1,
      anon_sym_BANG,
    ACTIONS(167), 1,
      anon_sym_LBRACK,
    ACTIONS(169), 1,
      anon_sym_assert,
    ACTIONS(173), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(177), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(181), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(183), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(187), 1,
      anon_sym_BQUOTE,
    ACTIONS(189), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(191), 1,
      sym_numeric_error,
    STATE(243), 1,
      sym_format_string,
    STATE(284), 1,
      sym__expression_inner,
    STATE(470), 1,
      sym_expression,
    ACTIONS(171), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(175), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(179), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(185), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(285), 2,
      sym_if_expression,
      sym_value,
    STATE(287), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(282), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [5205] = 25,
    ACTIONS(31), 1,
      anon_sym_LPAREN,
    ACTIONS(43), 1,
      anon_sym_LBRACK,
    ACTIONS(45), 1,
      anon_sym_assert,
    ACTIONS(51), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(55), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(59), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(61), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(69), 1,
      sym_numeric_error,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(289), 1,
      sym_identifier,
    ACTIONS(291), 1,
      anon_sym_SLASH,
    ACTIONS(293), 1,
      anon_sym_if,
    ACTIONS(295), 1,
      anon_sym_BANG,
    STATE(85), 1,
      sym_format_string,
    STATE(247), 1,
      sym__expression_inner,
    STATE(534), 1,
      sym_expression,
    ACTIONS(49), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(53), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(57), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(63), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(248), 2,
      sym_if_expression,
      sym_value,
    STATE(88), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [5292] = 25,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(453), 1,
      sym_expression,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [5379] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(155), 1,
      sym_identifier,
    ACTIONS(157), 1,
      anon_sym_LPAREN,
    ACTIONS(161), 1,
      anon_sym_SLASH,
    ACTIONS(163), 1,
      anon_sym_if,
    ACTIONS(165), 1,
      anon_sym_BANG,
    ACTIONS(167), 1,
      anon_sym_LBRACK,
    ACTIONS(169), 1,
      anon_sym_assert,
    ACTIONS(173), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(177), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(181), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(183), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(187), 1,
      anon_sym_BQUOTE,
    ACTIONS(189), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(191), 1,
      sym_numeric_error,
    STATE(243), 1,
      sym_format_string,
    STATE(284), 1,
      sym__expression_inner,
    STATE(509), 1,
      sym_expression,
    ACTIONS(171), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(175), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(179), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(185), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(285), 2,
      sym_if_expression,
      sym_value,
    STATE(287), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(282), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [5466] = 25,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(539), 1,
      sym_expression,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [5553] = 25,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(203), 1,
      anon_sym_SLASH,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(172), 1,
      sym__expression_inner,
    STATE(540), 1,
      sym_expression,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [5640] = 25,
    ACTIONS(31), 1,
      anon_sym_LPAREN,
    ACTIONS(43), 1,
      anon_sym_LBRACK,
    ACTIONS(45), 1,
      anon_sym_assert,
    ACTIONS(51), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(55), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(59), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(61), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(69), 1,
      sym_numeric_error,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(289), 1,
      sym_identifier,
    ACTIONS(293), 1,
      anon_sym_if,
    ACTIONS(295), 1,
      anon_sym_BANG,
    STATE(85), 1,
      sym_format_string,
    STATE(258), 1,
      sym_regex_literal,
    STATE(259), 1,
      sym_string,
    STATE(273), 1,
      sym__expression_inner,
    ACTIONS(49), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(53), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(57), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(63), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(248), 2,
      sym_if_expression,
      sym_value,
    STATE(88), 5,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
  [5726] = 25,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(32), 1,
      sym_regex_literal,
    STATE(33), 1,
      sym_string,
    STATE(129), 1,
      sym_format_string,
    STATE(169), 1,
      sym__expression_inner,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 5,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
  [5812] = 25,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(155), 1,
      sym_identifier,
    ACTIONS(157), 1,
      anon_sym_LPAREN,
    ACTIONS(163), 1,
      anon_sym_if,
    ACTIONS(165), 1,
      anon_sym_BANG,
    ACTIONS(167), 1,
      anon_sym_LBRACK,
    ACTIONS(169), 1,
      anon_sym_assert,
    ACTIONS(173), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(177), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(181), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(183), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(187), 1,
      anon_sym_BQUOTE,
    ACTIONS(189), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(191), 1,
      sym_numeric_error,
    STATE(243), 1,
      sym_format_string,
    STATE(263), 1,
      sym__expression_inner,
    STATE(264), 1,
      sym_regex_literal,
    STATE(265), 1,
      sym_string,
    ACTIONS(171), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(175), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(179), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(185), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(285), 2,
      sym_if_expression,
      sym_value,
    STATE(287), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(282), 5,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
  [5898] = 25,
    ACTIONS(29), 1,
      sym_identifier,
    ACTIONS(31), 1,
      anon_sym_LPAREN,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(41), 1,
      anon_sym_BANG,
    ACTIONS(43), 1,
      anon_sym_LBRACK,
    ACTIONS(45), 1,
      anon_sym_assert,
    ACTIONS(51), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(55), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(59), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(61), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(69), 1,
      sym_numeric_error,
    ACTIONS(71), 1,
      sym_comment,
    STATE(32), 1,
      sym_regex_literal,
    STATE(33), 1,
      sym_string,
    STATE(85), 1,
      sym_format_string,
    STATE(116), 1,
      sym__expression_inner,
    ACTIONS(49), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(53), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(57), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(63), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(88), 5,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
  [5984] = 23,
    ACTIONS(31), 1,
      anon_sym_LPAREN,
    ACTIONS(43), 1,
      anon_sym_LBRACK,
    ACTIONS(45), 1,
      anon_sym_assert,
    ACTIONS(51), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(55), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(59), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(61), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(69), 1,
      sym_numeric_error,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(289), 1,
      sym_identifier,
    ACTIONS(293), 1,
      anon_sym_if,
    ACTIONS(295), 1,
      anon_sym_BANG,
    STATE(85), 1,
      sym_format_string,
    STATE(275), 1,
      sym__expression_inner,
    ACTIONS(49), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(53), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(57), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(63), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(248), 2,
      sym_if_expression,
      sym_value,
    STATE(88), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [6065] = 23,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(173), 1,
      sym__expression_inner,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [6146] = 23,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(31), 1,
      sym__expression_inner,
    STATE(129), 1,
      sym_format_string,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [6227] = 23,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(169), 1,
      sym__expression_inner,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [6308] = 23,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(171), 1,
      sym__expression_inner,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [6389] = 23,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(170), 1,
      sym__expression_inner,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [6470] = 23,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(155), 1,
      sym_identifier,
    ACTIONS(157), 1,
      anon_sym_LPAREN,
    ACTIONS(163), 1,
      anon_sym_if,
    ACTIONS(165), 1,
      anon_sym_BANG,
    ACTIONS(167), 1,
      anon_sym_LBRACK,
    ACTIONS(169), 1,
      anon_sym_assert,
    ACTIONS(173), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(177), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(181), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(183), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(187), 1,
      anon_sym_BQUOTE,
    ACTIONS(189), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(191), 1,
      sym_numeric_error,
    STATE(243), 1,
      sym_format_string,
    STATE(260), 1,
      sym__expression_inner,
    ACTIONS(171), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(175), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(179), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(185), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(285), 2,
      sym_if_expression,
      sym_value,
    STATE(287), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(282), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [6551] = 23,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(155), 1,
      sym_identifier,
    ACTIONS(157), 1,
      anon_sym_LPAREN,
    ACTIONS(163), 1,
      anon_sym_if,
    ACTIONS(165), 1,
      anon_sym_BANG,
    ACTIONS(167), 1,
      anon_sym_LBRACK,
    ACTIONS(169), 1,
      anon_sym_assert,
    ACTIONS(173), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(177), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(181), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(183), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(187), 1,
      anon_sym_BQUOTE,
    ACTIONS(189), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(191), 1,
      sym_numeric_error,
    STATE(243), 1,
      sym_format_string,
    STATE(263), 1,
      sym__expression_inner,
    ACTIONS(171), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(175), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(179), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(185), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(285), 2,
      sym_if_expression,
      sym_value,
    STATE(287), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(282), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [6632] = 23,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(155), 1,
      sym_identifier,
    ACTIONS(157), 1,
      anon_sym_LPAREN,
    ACTIONS(163), 1,
      anon_sym_if,
    ACTIONS(165), 1,
      anon_sym_BANG,
    ACTIONS(167), 1,
      anon_sym_LBRACK,
    ACTIONS(169), 1,
      anon_sym_assert,
    ACTIONS(173), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(177), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(181), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(183), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(187), 1,
      anon_sym_BQUOTE,
    ACTIONS(189), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(191), 1,
      sym_numeric_error,
    STATE(243), 1,
      sym_format_string,
    STATE(268), 1,
      sym__expression_inner,
    ACTIONS(171), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(175), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(179), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(185), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(285), 2,
      sym_if_expression,
      sym_value,
    STATE(287), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(282), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [6713] = 23,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(155), 1,
      sym_identifier,
    ACTIONS(157), 1,
      anon_sym_LPAREN,
    ACTIONS(163), 1,
      anon_sym_if,
    ACTIONS(165), 1,
      anon_sym_BANG,
    ACTIONS(167), 1,
      anon_sym_LBRACK,
    ACTIONS(169), 1,
      anon_sym_assert,
    ACTIONS(173), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(177), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(181), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(183), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(187), 1,
      anon_sym_BQUOTE,
    ACTIONS(189), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(191), 1,
      sym_numeric_error,
    STATE(243), 1,
      sym_format_string,
    STATE(269), 1,
      sym__expression_inner,
    ACTIONS(171), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(175), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(179), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(185), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(285), 2,
      sym_if_expression,
      sym_value,
    STATE(287), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(282), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [6794] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(297), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(299), 21,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [6835] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(301), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(303), 21,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [6876] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(305), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(307), 21,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [6917] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(309), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(311), 21,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [6958] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(313), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(315), 21,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [6999] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(317), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(319), 21,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [7040] = 23,
    ACTIONS(31), 1,
      anon_sym_LPAREN,
    ACTIONS(43), 1,
      anon_sym_LBRACK,
    ACTIONS(45), 1,
      anon_sym_assert,
    ACTIONS(51), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(55), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(59), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(61), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(69), 1,
      sym_numeric_error,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(289), 1,
      sym_identifier,
    ACTIONS(293), 1,
      anon_sym_if,
    ACTIONS(295), 1,
      anon_sym_BANG,
    STATE(85), 1,
      sym_format_string,
    STATE(250), 1,
      sym__expression_inner,
    ACTIONS(49), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(53), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(57), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(63), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(248), 2,
      sym_if_expression,
      sym_value,
    STATE(88), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [7121] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(321), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(323), 21,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [7162] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(325), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(327), 21,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [7203] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(329), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(331), 21,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [7244] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(333), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(335), 21,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [7285] = 23,
    ACTIONS(29), 1,
      sym_identifier,
    ACTIONS(31), 1,
      anon_sym_LPAREN,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(41), 1,
      anon_sym_BANG,
    ACTIONS(43), 1,
      anon_sym_LBRACK,
    ACTIONS(45), 1,
      anon_sym_assert,
    ACTIONS(51), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(55), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(59), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(61), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(69), 1,
      sym_numeric_error,
    ACTIONS(71), 1,
      sym_comment,
    STATE(31), 1,
      sym__expression_inner,
    STATE(85), 1,
      sym_format_string,
    ACTIONS(49), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(53), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(57), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(63), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(88), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [7366] = 23,
    ACTIONS(29), 1,
      sym_identifier,
    ACTIONS(31), 1,
      anon_sym_LPAREN,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(41), 1,
      anon_sym_BANG,
    ACTIONS(43), 1,
      anon_sym_LBRACK,
    ACTIONS(45), 1,
      anon_sym_assert,
    ACTIONS(51), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(55), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(59), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(61), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(69), 1,
      sym_numeric_error,
    ACTIONS(71), 1,
      sym_comment,
    STATE(85), 1,
      sym_format_string,
    STATE(116), 1,
      sym__expression_inner,
    ACTIONS(49), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(53), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(57), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(63), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(88), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [7447] = 23,
    ACTIONS(29), 1,
      sym_identifier,
    ACTIONS(31), 1,
      anon_sym_LPAREN,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(41), 1,
      anon_sym_BANG,
    ACTIONS(43), 1,
      anon_sym_LBRACK,
    ACTIONS(45), 1,
      anon_sym_assert,
    ACTIONS(51), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(55), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(59), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(61), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(69), 1,
      sym_numeric_error,
    ACTIONS(71), 1,
      sym_comment,
    STATE(85), 1,
      sym_format_string,
    STATE(118), 1,
      sym__expression_inner,
    ACTIONS(49), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(53), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(57), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(63), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(88), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [7528] = 23,
    ACTIONS(29), 1,
      sym_identifier,
    ACTIONS(31), 1,
      anon_sym_LPAREN,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(41), 1,
      anon_sym_BANG,
    ACTIONS(43), 1,
      anon_sym_LBRACK,
    ACTIONS(45), 1,
      anon_sym_assert,
    ACTIONS(51), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(55), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(59), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(61), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(69), 1,
      sym_numeric_error,
    ACTIONS(71), 1,
      sym_comment,
    STATE(85), 1,
      sym_format_string,
    STATE(119), 1,
      sym__expression_inner,
    ACTIONS(49), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(53), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(57), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(63), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(88), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [7609] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(337), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(339), 21,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [7650] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(341), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(343), 21,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [7691] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(345), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(347), 21,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [7732] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(349), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(351), 21,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [7773] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(353), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(355), 21,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [7814] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(357), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(359), 21,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [7855] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(361), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(363), 21,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [7896] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(365), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(367), 21,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [7937] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(369), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(371), 21,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [7978] = 23,
    ACTIONS(31), 1,
      anon_sym_LPAREN,
    ACTIONS(43), 1,
      anon_sym_LBRACK,
    ACTIONS(45), 1,
      anon_sym_assert,
    ACTIONS(51), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(55), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(59), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(61), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(69), 1,
      sym_numeric_error,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(289), 1,
      sym_identifier,
    ACTIONS(293), 1,
      anon_sym_if,
    ACTIONS(295), 1,
      anon_sym_BANG,
    STATE(85), 1,
      sym_format_string,
    STATE(257), 1,
      sym__expression_inner,
    ACTIONS(49), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(53), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(57), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(63), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(248), 2,
      sym_if_expression,
      sym_value,
    STATE(88), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [8059] = 23,
    ACTIONS(31), 1,
      anon_sym_LPAREN,
    ACTIONS(43), 1,
      anon_sym_LBRACK,
    ACTIONS(45), 1,
      anon_sym_assert,
    ACTIONS(51), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(55), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(59), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(61), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(69), 1,
      sym_numeric_error,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(289), 1,
      sym_identifier,
    ACTIONS(293), 1,
      anon_sym_if,
    ACTIONS(295), 1,
      anon_sym_BANG,
    STATE(85), 1,
      sym_format_string,
    STATE(273), 1,
      sym__expression_inner,
    ACTIONS(49), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(53), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(57), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(63), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(248), 2,
      sym_if_expression,
      sym_value,
    STATE(88), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [8140] = 23,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(155), 1,
      sym_identifier,
    ACTIONS(157), 1,
      anon_sym_LPAREN,
    ACTIONS(163), 1,
      anon_sym_if,
    ACTIONS(165), 1,
      anon_sym_BANG,
    ACTIONS(167), 1,
      anon_sym_LBRACK,
    ACTIONS(169), 1,
      anon_sym_assert,
    ACTIONS(173), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(177), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(181), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(183), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(187), 1,
      anon_sym_BQUOTE,
    ACTIONS(189), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(191), 1,
      sym_numeric_error,
    STATE(237), 1,
      sym__expression_inner,
    STATE(243), 1,
      sym_format_string,
    ACTIONS(171), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(175), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(179), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(185), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(285), 2,
      sym_if_expression,
      sym_value,
    STATE(287), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(282), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [8221] = 23,
    ACTIONS(31), 1,
      anon_sym_LPAREN,
    ACTIONS(43), 1,
      anon_sym_LBRACK,
    ACTIONS(45), 1,
      anon_sym_assert,
    ACTIONS(51), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(55), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(59), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(61), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(69), 1,
      sym_numeric_error,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(289), 1,
      sym_identifier,
    ACTIONS(293), 1,
      anon_sym_if,
    ACTIONS(295), 1,
      anon_sym_BANG,
    STATE(85), 1,
      sym_format_string,
    STATE(274), 1,
      sym__expression_inner,
    ACTIONS(49), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(53), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(57), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(63), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(248), 2,
      sym_if_expression,
      sym_value,
    STATE(88), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [8302] = 23,
    ACTIONS(29), 1,
      sym_identifier,
    ACTIONS(31), 1,
      anon_sym_LPAREN,
    ACTIONS(39), 1,
      anon_sym_if,
    ACTIONS(41), 1,
      anon_sym_BANG,
    ACTIONS(43), 1,
      anon_sym_LBRACK,
    ACTIONS(45), 1,
      anon_sym_assert,
    ACTIONS(51), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(55), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(59), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(61), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(69), 1,
      sym_numeric_error,
    ACTIONS(71), 1,
      sym_comment,
    STATE(85), 1,
      sym_format_string,
    STATE(120), 1,
      sym__expression_inner,
    ACTIONS(49), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(53), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(57), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(63), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(29), 2,
      sym_if_expression,
      sym_value,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(88), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [8383] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(373), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(375), 21,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [8424] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(383), 1,
      anon_sym_PLUS,
    ACTIONS(389), 1,
      anon_sym_AMP_AMP,
    ACTIONS(391), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(381), 2,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(385), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(387), 2,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
    ACTIONS(377), 11,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(379), 12,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [8476] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(383), 1,
      anon_sym_PLUS,
    ACTIONS(381), 2,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(265), 11,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(267), 18,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [8520] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(309), 12,
      anon_sym_PLUS,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(311), 20,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [8560] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(383), 1,
      anon_sym_PLUS,
    ACTIONS(381), 2,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(385), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(387), 2,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
    ACTIONS(265), 11,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(267), 14,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [8608] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(383), 1,
      anon_sym_PLUS,
    ACTIONS(389), 1,
      anon_sym_AMP_AMP,
    ACTIONS(381), 2,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(385), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(387), 2,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
    ACTIONS(265), 11,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(267), 13,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [8658] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(383), 1,
      anon_sym_PLUS,
    ACTIONS(389), 1,
      anon_sym_AMP_AMP,
    ACTIONS(391), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(381), 2,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(385), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(387), 2,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
    ACTIONS(393), 11,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(395), 12,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [8710] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(360), 1,
      sym_value,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [8784] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(155), 1,
      sym_value,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [8858] = 21,
    ACTIONS(29), 1,
      sym_identifier,
    ACTIONS(31), 1,
      anon_sym_LPAREN,
    ACTIONS(41), 1,
      anon_sym_BANG,
    ACTIONS(43), 1,
      anon_sym_LBRACK,
    ACTIONS(45), 1,
      anon_sym_assert,
    ACTIONS(51), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(55), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(59), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(61), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(69), 1,
      sym_numeric_error,
    ACTIONS(71), 1,
      sym_comment,
    STATE(85), 1,
      sym_format_string,
    STATE(149), 1,
      sym_value,
    ACTIONS(49), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(53), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(57), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(63), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(88), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [8932] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(197), 1,
      sym_identifier,
    ACTIONS(199), 1,
      anon_sym_LPAREN,
    ACTIONS(205), 1,
      anon_sym_BANG,
    ACTIONS(207), 1,
      anon_sym_LBRACK,
    ACTIONS(209), 1,
      anon_sym_assert,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(227), 1,
      anon_sym_BQUOTE,
    ACTIONS(229), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(231), 1,
      sym_numeric_error,
    STATE(129), 1,
      sym_format_string,
    STATE(345), 1,
      sym_value,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(162), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(156), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [9006] = 21,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(155), 1,
      sym_identifier,
    ACTIONS(157), 1,
      anon_sym_LPAREN,
    ACTIONS(165), 1,
      anon_sym_BANG,
    ACTIONS(167), 1,
      anon_sym_LBRACK,
    ACTIONS(169), 1,
      anon_sym_assert,
    ACTIONS(173), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(177), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(181), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(183), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(187), 1,
      anon_sym_BQUOTE,
    ACTIONS(189), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(191), 1,
      sym_numeric_error,
    STATE(238), 1,
      sym_value,
    STATE(243), 1,
      sym_format_string,
    ACTIONS(171), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(175), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(179), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(185), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(287), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(282), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [9080] = 21,
    ACTIONS(31), 1,
      anon_sym_LPAREN,
    ACTIONS(43), 1,
      anon_sym_LBRACK,
    ACTIONS(45), 1,
      anon_sym_assert,
    ACTIONS(51), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(55), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(59), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(61), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(69), 1,
      sym_numeric_error,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(289), 1,
      sym_identifier,
    ACTIONS(295), 1,
      anon_sym_BANG,
    STATE(85), 1,
      sym_format_string,
    STATE(92), 1,
      sym_value,
    ACTIONS(49), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(53), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(57), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(63), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(88), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [9154] = 21,
    ACTIONS(29), 1,
      sym_identifier,
    ACTIONS(31), 1,
      anon_sym_LPAREN,
    ACTIONS(41), 1,
      anon_sym_BANG,
    ACTIONS(43), 1,
      anon_sym_LBRACK,
    ACTIONS(45), 1,
      anon_sym_assert,
    ACTIONS(51), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(55), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(59), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(61), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(65), 1,
      anon_sym_BQUOTE,
    ACTIONS(67), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(69), 1,
      sym_numeric_error,
    ACTIONS(71), 1,
      sym_comment,
    STATE(85), 1,
      sym_format_string,
    STATE(92), 1,
      sym_value,
    ACTIONS(49), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(53), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(57), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(63), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    STATE(89), 2,
      sym__backticked,
      sym__indented_backticked,
    STATE(88), 6,
      sym_not_expression,
      sym_list_literal,
      sym_assert_expression,
      sym_function_call,
      sym_external_command,
      sym_string,
  [9228] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(301), 9,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      anon_sym_PLUS,
      sym_identifier,
    ACTIONS(303), 19,
      ts_builtin_sym_end,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_AT,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      anon_sym_RBRACE_RBRACE,
  [9264] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(297), 9,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      anon_sym_PLUS,
      sym_identifier,
    ACTIONS(299), 19,
      ts_builtin_sym_end,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_AT,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      anon_sym_RBRACE_RBRACE,
  [9300] = 17,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      anon_sym_alias,
    ACTIONS(11), 1,
      anon_sym_unexport,
    ACTIONS(13), 1,
      anon_sym_eager,
    ACTIONS(15), 1,
      anon_sym_export,
    ACTIONS(17), 1,
      anon_sym_import,
    ACTIONS(19), 1,
      anon_sym_mod,
    ACTIONS(21), 1,
      anon_sym_set,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(25), 1,
      anon_sym_AT,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(397), 1,
      ts_builtin_sym_end,
    STATE(207), 1,
      sym__assignment,
    STATE(279), 1,
      sym__name,
    STATE(507), 1,
      sym_recipe_header,
    STATE(168), 2,
      sym_attribute,
      aux_sym_function_definition_repeat1,
    STATE(134), 12,
      sym__item,
      sym_function_definition,
      sym_alias,
      sym_assignment,
      sym_eager,
      sym_export,
      sym_unexport,
      sym_import,
      sym_module,
      sym_setting,
      sym_recipe,
      aux_sym_source_file_repeat1,
  [9364] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(305), 9,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      anon_sym_PLUS,
      sym_identifier,
    ACTIONS(307), 19,
      ts_builtin_sym_end,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_AT,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      anon_sym_RBRACE_RBRACE,
  [9400] = 17,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      anon_sym_alias,
    ACTIONS(11), 1,
      anon_sym_unexport,
    ACTIONS(13), 1,
      anon_sym_eager,
    ACTIONS(15), 1,
      anon_sym_export,
    ACTIONS(17), 1,
      anon_sym_import,
    ACTIONS(19), 1,
      anon_sym_mod,
    ACTIONS(21), 1,
      anon_sym_set,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(25), 1,
      anon_sym_AT,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(399), 1,
      ts_builtin_sym_end,
    STATE(207), 1,
      sym__assignment,
    STATE(279), 1,
      sym__name,
    STATE(507), 1,
      sym_recipe_header,
    STATE(168), 2,
      sym_attribute,
      aux_sym_function_definition_repeat1,
    STATE(135), 12,
      sym__item,
      sym_function_definition,
      sym_alias,
      sym_assignment,
      sym_eager,
      sym_export,
      sym_unexport,
      sym_import,
      sym_module,
      sym_setting,
      sym_recipe,
      aux_sym_source_file_repeat1,
  [9464] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(373), 9,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      anon_sym_PLUS,
      sym_identifier,
    ACTIONS(375), 19,
      ts_builtin_sym_end,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_AT,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      anon_sym_RBRACE_RBRACE,
  [9500] = 17,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(401), 1,
      ts_builtin_sym_end,
    ACTIONS(403), 1,
      sym_identifier,
    ACTIONS(406), 1,
      anon_sym_alias,
    ACTIONS(409), 1,
      anon_sym_unexport,
    ACTIONS(412), 1,
      anon_sym_eager,
    ACTIONS(415), 1,
      anon_sym_export,
    ACTIONS(418), 1,
      anon_sym_import,
    ACTIONS(421), 1,
      anon_sym_mod,
    ACTIONS(424), 1,
      anon_sym_set,
    ACTIONS(427), 1,
      anon_sym_LBRACK,
    ACTIONS(430), 1,
      anon_sym_AT,
    STATE(207), 1,
      sym__assignment,
    STATE(279), 1,
      sym__name,
    STATE(507), 1,
      sym_recipe_header,
    STATE(168), 2,
      sym_attribute,
      aux_sym_function_definition_repeat1,
    STATE(134), 12,
      sym__item,
      sym_function_definition,
      sym_alias,
      sym_assignment,
      sym_eager,
      sym_export,
      sym_unexport,
      sym_import,
      sym_module,
      sym_setting,
      sym_recipe,
      aux_sym_source_file_repeat1,
  [9564] = 17,
    ACTIONS(7), 1,
      sym_identifier,
    ACTIONS(9), 1,
      anon_sym_alias,
    ACTIONS(11), 1,
      anon_sym_unexport,
    ACTIONS(13), 1,
      anon_sym_eager,
    ACTIONS(15), 1,
      anon_sym_export,
    ACTIONS(17), 1,
      anon_sym_import,
    ACTIONS(19), 1,
      anon_sym_mod,
    ACTIONS(21), 1,
      anon_sym_set,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(25), 1,
      anon_sym_AT,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(433), 1,
      ts_builtin_sym_end,
    STATE(207), 1,
      sym__assignment,
    STATE(279), 1,
      sym__name,
    STATE(507), 1,
      sym_recipe_header,
    STATE(168), 2,
      sym_attribute,
      aux_sym_function_definition_repeat1,
    STATE(134), 12,
      sym__item,
      sym_function_definition,
      sym_alias,
      sym_assignment,
      sym_eager,
      sym_export,
      sym_unexport,
      sym_import,
      sym_module,
      sym_setting,
      sym_recipe,
      aux_sym_source_file_repeat1,
  [9628] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(317), 9,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      anon_sym_PLUS,
      sym_identifier,
    ACTIONS(319), 19,
      ts_builtin_sym_end,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_AT,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      anon_sym_RBRACE_RBRACE,
  [9664] = 5,
    ACTIONS(35), 1,
      anon_sym_COLON_COLON,
    ACTIONS(71), 1,
      sym_comment,
    STATE(138), 1,
      aux_sym_module_path_repeat1,
    ACTIONS(435), 10,
      anon_sym_if,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(437), 14,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SLASH,
      anon_sym_BANG,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [9702] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(443), 1,
      anon_sym_COLON_COLON,
    STATE(138), 1,
      aux_sym_module_path_repeat1,
    ACTIONS(439), 10,
      anon_sym_if,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(441), 14,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SLASH,
      anon_sym_BANG,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [9740] = 13,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(27), 1,
      aux_sym_shebang_token1,
    ACTIONS(450), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(452), 1,
      sym__dedent,
    ACTIONS(454), 1,
      sym__newline,
    ACTIONS(456), 1,
      sym_text,
    STATE(146), 1,
      aux_sym_recipe_body_repeat1,
    STATE(367), 1,
      sym_recipe_line_prefix,
    STATE(551), 1,
      sym_recipe_line,
    STATE(553), 1,
      sym_shebang,
    STATE(363), 2,
      sym_interpolation,
      aux_sym_recipe_line_repeat1,
    ACTIONS(448), 6,
      anon_sym_AT_DASH_QMARK,
      anon_sym_AT_QMARK_DASH,
      anon_sym_DASH_AT_QMARK,
      anon_sym_DASH_QMARK_AT,
      anon_sym_QMARK_AT_DASH,
      anon_sym_QMARK_DASH_AT,
    ACTIONS(446), 9,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_AT_DASH,
      anon_sym_AT_QMARK,
      anon_sym_DASH_AT,
      anon_sym_DASH_QMARK,
      anon_sym_QMARK_AT,
      anon_sym_QMARK_DASH,
      anon_sym_DASH,
  [9794] = 13,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    STATE(129), 1,
      sym_format_string,
    STATE(211), 1,
      sym_string,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(458), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(460), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [9847] = 13,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    STATE(129), 1,
      sym_format_string,
    STATE(221), 1,
      sym_string,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(462), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(464), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [9900] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(439), 10,
      anon_sym_if,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(441), 15,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COLON_COLON,
      anon_sym_SLASH,
      anon_sym_BANG,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [9933] = 13,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    STATE(129), 1,
      sym_format_string,
    STATE(202), 1,
      sym_string,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(466), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(468), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [9986] = 11,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(450), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(456), 1,
      sym_text,
    ACTIONS(470), 1,
      sym__dedent,
    ACTIONS(472), 1,
      sym__newline,
    STATE(145), 1,
      aux_sym_recipe_body_repeat1,
    STATE(367), 1,
      sym_recipe_line_prefix,
    STATE(551), 1,
      sym_recipe_line,
    STATE(363), 2,
      sym_interpolation,
      aux_sym_recipe_line_repeat1,
    ACTIONS(448), 6,
      anon_sym_AT_DASH_QMARK,
      anon_sym_AT_QMARK_DASH,
      anon_sym_DASH_AT_QMARK,
      anon_sym_DASH_QMARK_AT,
      anon_sym_QMARK_AT_DASH,
      anon_sym_QMARK_DASH_AT,
    ACTIONS(446), 9,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_AT_DASH,
      anon_sym_AT_QMARK,
      anon_sym_DASH_AT,
      anon_sym_DASH_QMARK,
      anon_sym_QMARK_AT,
      anon_sym_QMARK_DASH,
      anon_sym_DASH,
  [10034] = 11,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(450), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(456), 1,
      sym_text,
    ACTIONS(474), 1,
      sym__dedent,
    ACTIONS(476), 1,
      sym__newline,
    STATE(148), 1,
      aux_sym_recipe_body_repeat1,
    STATE(367), 1,
      sym_recipe_line_prefix,
    STATE(551), 1,
      sym_recipe_line,
    STATE(363), 2,
      sym_interpolation,
      aux_sym_recipe_line_repeat1,
    ACTIONS(448), 6,
      anon_sym_AT_DASH_QMARK,
      anon_sym_AT_QMARK_DASH,
      anon_sym_DASH_AT_QMARK,
      anon_sym_DASH_QMARK_AT,
      anon_sym_QMARK_AT_DASH,
      anon_sym_QMARK_DASH_AT,
    ACTIONS(446), 9,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_AT_DASH,
      anon_sym_AT_QMARK,
      anon_sym_DASH_AT,
      anon_sym_DASH_QMARK,
      anon_sym_QMARK_AT,
      anon_sym_QMARK_DASH,
      anon_sym_DASH,
  [10082] = 11,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(450), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(456), 1,
      sym_text,
    ACTIONS(476), 1,
      sym__newline,
    ACTIONS(478), 1,
      sym__dedent,
    STATE(148), 1,
      aux_sym_recipe_body_repeat1,
    STATE(367), 1,
      sym_recipe_line_prefix,
    STATE(551), 1,
      sym_recipe_line,
    STATE(363), 2,
      sym_interpolation,
      aux_sym_recipe_line_repeat1,
    ACTIONS(448), 6,
      anon_sym_AT_DASH_QMARK,
      anon_sym_AT_QMARK_DASH,
      anon_sym_DASH_AT_QMARK,
      anon_sym_DASH_QMARK_AT,
      anon_sym_QMARK_AT_DASH,
      anon_sym_QMARK_DASH_AT,
    ACTIONS(446), 9,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_AT_DASH,
      anon_sym_AT_QMARK,
      anon_sym_DASH_AT,
      anon_sym_DASH_QMARK,
      anon_sym_QMARK_AT,
      anon_sym_QMARK_DASH,
      anon_sym_DASH,
  [10130] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(480), 1,
      sym_identifier,
    ACTIONS(482), 1,
      anon_sym_QMARK,
    ACTIONS(486), 1,
      anon_sym_COLON,
    ACTIONS(488), 1,
      anon_sym_DOLLAR,
    STATE(129), 1,
      sym_format_string,
    STATE(213), 1,
      sym_string,
    STATE(315), 1,
      aux_sym_parameters_repeat1,
    STATE(348), 1,
      sym_parameter,
    STATE(485), 1,
      sym_variadic_parameter,
    STATE(545), 1,
      sym_parameters,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(484), 2,
      anon_sym_PLUS,
      anon_sym_STAR,
  [10196] = 11,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(496), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(499), 1,
      sym__dedent,
    ACTIONS(501), 1,
      sym__newline,
    ACTIONS(504), 1,
      sym_text,
    STATE(148), 1,
      aux_sym_recipe_body_repeat1,
    STATE(367), 1,
      sym_recipe_line_prefix,
    STATE(551), 1,
      sym_recipe_line,
    STATE(363), 2,
      sym_interpolation,
      aux_sym_recipe_line_repeat1,
    ACTIONS(493), 6,
      anon_sym_AT_DASH_QMARK,
      anon_sym_AT_QMARK_DASH,
      anon_sym_DASH_AT_QMARK,
      anon_sym_DASH_QMARK_AT,
      anon_sym_QMARK_AT_DASH,
      anon_sym_QMARK_DASH_AT,
    ACTIONS(490), 9,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_AT_DASH,
      anon_sym_AT_QMARK,
      anon_sym_DASH_AT,
      anon_sym_DASH_QMARK,
      anon_sym_QMARK_AT,
      anon_sym_QMARK_DASH,
      anon_sym_DASH,
  [10244] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(507), 10,
      anon_sym_if,
      anon_sym_assert,
      aux_sym_string_token1,
      aux_sym_string_token2,
      anon_sym_f_DQUOTE,
      anon_sym_DQUOTE,
      anon_sym_f_SQUOTE,
      anon_sym_x_DQUOTE,
      anon_sym_BQUOTE,
      sym_identifier,
    ACTIONS(509), 14,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_SLASH,
      anon_sym_BANG,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      sym_numeric_error,
  [10276] = 20,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(480), 1,
      sym_identifier,
    ACTIONS(486), 1,
      anon_sym_COLON,
    ACTIONS(488), 1,
      anon_sym_DOLLAR,
    ACTIONS(511), 1,
      anon_sym_QMARK,
    STATE(129), 1,
      sym_format_string,
    STATE(201), 1,
      sym_string,
    STATE(315), 1,
      aux_sym_parameters_repeat1,
    STATE(348), 1,
      sym_parameter,
    STATE(485), 1,
      sym_variadic_parameter,
    STATE(545), 1,
      sym_parameters,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(484), 2,
      anon_sym_PLUS,
      anon_sym_STAR,
  [10342] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(309), 1,
      anon_sym_PLUS,
    ACTIONS(513), 1,
      anon_sym_LPAREN,
    ACTIONS(311), 17,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      anon_sym_RBRACE_RBRACE,
      sym_identifier,
  [10371] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(515), 9,
      anon_sym_QMARK,
      anon_sym_AT,
      anon_sym_AT_DASH,
      anon_sym_AT_QMARK,
      anon_sym_DASH_AT,
      anon_sym_DASH_QMARK,
      anon_sym_QMARK_AT,
      anon_sym_QMARK_DASH,
      anon_sym_DASH,
    ACTIONS(499), 10,
      sym__dedent,
      sym__newline,
      sym_text,
      anon_sym_AT_DASH_QMARK,
      anon_sym_AT_QMARK_DASH,
      anon_sym_DASH_AT_QMARK,
      anon_sym_DASH_QMARK_AT,
      anon_sym_QMARK_AT_DASH,
      anon_sym_QMARK_DASH_AT,
      anon_sym_LBRACE_LBRACE,
  [10398] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(333), 1,
      anon_sym_PLUS,
    ACTIONS(335), 17,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      anon_sym_RBRACE_RBRACE,
      sym_identifier,
  [10424] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(329), 1,
      anon_sym_PLUS,
    ACTIONS(331), 17,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      anon_sym_RBRACE_RBRACE,
      sym_identifier,
  [10450] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(321), 1,
      anon_sym_PLUS,
    ACTIONS(323), 17,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      anon_sym_RBRACE_RBRACE,
      sym_identifier,
  [10476] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(309), 1,
      anon_sym_PLUS,
    ACTIONS(311), 17,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      anon_sym_RBRACE_RBRACE,
      sym_identifier,
  [10502] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(325), 1,
      anon_sym_PLUS,
    ACTIONS(327), 17,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      anon_sym_RBRACE_RBRACE,
      sym_identifier,
  [10528] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(369), 1,
      anon_sym_PLUS,
    ACTIONS(371), 17,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      anon_sym_RBRACE_RBRACE,
      sym_identifier,
  [10554] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(361), 1,
      anon_sym_PLUS,
    ACTIONS(363), 17,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      anon_sym_RBRACE_RBRACE,
      sym_identifier,
  [10580] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(357), 1,
      anon_sym_PLUS,
    ACTIONS(359), 17,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      anon_sym_RBRACE_RBRACE,
      sym_identifier,
  [10606] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(365), 1,
      anon_sym_PLUS,
    ACTIONS(367), 17,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      anon_sym_RBRACE_RBRACE,
      sym_identifier,
  [10632] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(313), 1,
      anon_sym_PLUS,
    ACTIONS(315), 17,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      anon_sym_RBRACE_RBRACE,
      sym_identifier,
  [10658] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(337), 1,
      anon_sym_PLUS,
    ACTIONS(339), 17,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      anon_sym_RBRACE_RBRACE,
      sym_identifier,
  [10684] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(353), 1,
      anon_sym_PLUS,
    ACTIONS(355), 17,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      anon_sym_RBRACE_RBRACE,
      sym_identifier,
  [10710] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(345), 1,
      anon_sym_PLUS,
    ACTIONS(347), 17,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      anon_sym_RBRACE_RBRACE,
      sym_identifier,
  [10736] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(341), 1,
      anon_sym_PLUS,
    ACTIONS(343), 17,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      anon_sym_RBRACE_RBRACE,
      sym_identifier,
  [10762] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(349), 1,
      anon_sym_PLUS,
    ACTIONS(351), 17,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      anon_sym_RBRACE_RBRACE,
      sym_identifier,
  [10788] = 15,
    ACTIONS(23), 1,
      anon_sym_LBRACK,
    ACTIONS(25), 1,
      anon_sym_AT,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(517), 1,
      sym_identifier,
    ACTIONS(519), 1,
      anon_sym_alias,
    ACTIONS(521), 1,
      anon_sym_unexport,
    ACTIONS(523), 1,
      anon_sym_eager,
    ACTIONS(525), 1,
      anon_sym_export,
    ACTIONS(527), 1,
      anon_sym_import,
    ACTIONS(529), 1,
      anon_sym_mod,
    ACTIONS(531), 1,
      anon_sym_set,
    STATE(199), 1,
      sym__assignment,
    STATE(279), 1,
      sym__name,
    STATE(555), 1,
      sym_recipe_header,
    STATE(189), 2,
      sym_attribute,
      aux_sym_function_definition_repeat1,
  [10835] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(535), 1,
      anon_sym_PLUS,
    ACTIONS(533), 2,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(267), 11,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_RBRACK,
      anon_sym_RBRACE_RBRACE,
  [10859] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(535), 1,
      anon_sym_PLUS,
    ACTIONS(541), 1,
      anon_sym_AMP_AMP,
    ACTIONS(533), 2,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(537), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(539), 2,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
    ACTIONS(267), 6,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_RBRACK,
      anon_sym_RBRACE_RBRACE,
  [10889] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(535), 1,
      anon_sym_PLUS,
    ACTIONS(533), 2,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(537), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(539), 2,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
    ACTIONS(267), 7,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_LBRACE,
      anon_sym_RBRACK,
      anon_sym_RBRACE_RBRACE,
  [10917] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(535), 1,
      anon_sym_PLUS,
    ACTIONS(541), 1,
      anon_sym_AMP_AMP,
    ACTIONS(543), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(533), 2,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(537), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(539), 2,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
    ACTIONS(379), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RBRACK,
      anon_sym_RBRACE_RBRACE,
  [10949] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(535), 1,
      anon_sym_PLUS,
    ACTIONS(541), 1,
      anon_sym_AMP_AMP,
    ACTIONS(543), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(533), 2,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(537), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(539), 2,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
    ACTIONS(395), 5,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
      anon_sym_RBRACK,
      anon_sym_RBRACE_RBRACE,
  [10981] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(81), 1,
      anon_sym_PLUS,
    ACTIONS(545), 1,
      anon_sym_else,
    STATE(183), 1,
      aux_sym_if_expression_repeat1,
    STATE(215), 1,
      sym_else_if_clause,
    STATE(267), 1,
      sym_else_clause,
    ACTIONS(83), 9,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
  [11011] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(81), 1,
      anon_sym_PLUS,
    ACTIONS(547), 1,
      anon_sym_else,
    STATE(188), 1,
      aux_sym_if_expression_repeat1,
    STATE(224), 1,
      sym_else_if_clause,
    STATE(288), 1,
      sym_else_clause,
    ACTIONS(83), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [11041] = 11,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    STATE(129), 1,
      sym_format_string,
    STATE(201), 1,
      sym_string,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
  [11079] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(75), 1,
      anon_sym_PLUS,
    ACTIONS(545), 1,
      anon_sym_else,
    STATE(174), 1,
      aux_sym_if_expression_repeat1,
    STATE(215), 1,
      sym_else_if_clause,
    STATE(262), 1,
      sym_else_clause,
    ACTIONS(77), 9,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
  [11109] = 11,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(213), 1,
      anon_sym_f_DQUOTE,
    ACTIONS(217), 1,
      anon_sym_f_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(221), 1,
      anon_sym_f_SQUOTE,
    ACTIONS(223), 1,
      anon_sym_f_SQUOTE_SQUOTE_SQUOTE,
    STATE(129), 1,
      sym_format_string,
    STATE(229), 1,
      sym_string,
    ACTIONS(211), 2,
      aux_sym_string_token1,
      aux_sym_string_token2,
    ACTIONS(215), 2,
      anon_sym_DQUOTE,
      anon_sym_x_DQUOTE,
    ACTIONS(219), 2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      anon_sym_x_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(225), 2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_x_SQUOTE_SQUOTE_SQUOTE,
  [11147] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(75), 1,
      anon_sym_PLUS,
    ACTIONS(547), 1,
      anon_sym_else,
    STATE(175), 1,
      aux_sym_if_expression_repeat1,
    STATE(224), 1,
      sym_else_if_clause,
    STATE(281), 1,
      sym_else_clause,
    ACTIONS(77), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [11177] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(553), 1,
      anon_sym_COLON_COLON,
    STATE(187), 1,
      aux_sym_module_path_repeat1,
    ACTIONS(549), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(551), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11202] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(559), 1,
      sym__indent,
    STATE(203), 1,
      sym_recipe_body,
    ACTIONS(555), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(557), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11227] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(553), 1,
      anon_sym_COLON_COLON,
    STATE(187), 1,
      aux_sym_module_path_repeat1,
    ACTIONS(561), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(563), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11252] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(148), 1,
      anon_sym_PLUS,
    ACTIONS(565), 1,
      anon_sym_else,
    STATE(183), 1,
      aux_sym_if_expression_repeat1,
    STATE(215), 1,
      sym_else_if_clause,
    ACTIONS(150), 9,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
  [11279] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(568), 1,
      anon_sym_COLON_COLON,
    STATE(184), 1,
      aux_sym_module_path_repeat1,
    ACTIONS(441), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(439), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11304] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(559), 1,
      sym__indent,
    STATE(191), 1,
      sym_recipe_body,
    ACTIONS(571), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(573), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11329] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(309), 1,
      anon_sym_PLUS,
    ACTIONS(513), 1,
      anon_sym_LPAREN,
    ACTIONS(578), 1,
      anon_sym_EQ,
    ACTIONS(575), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    ACTIONS(311), 8,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [11356] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(553), 1,
      anon_sym_COLON_COLON,
    STATE(184), 1,
      aux_sym_module_path_repeat1,
    ACTIONS(437), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(435), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11381] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(148), 1,
      anon_sym_PLUS,
    ACTIONS(580), 1,
      anon_sym_else,
    STATE(188), 1,
      aux_sym_if_expression_repeat1,
    STATE(224), 1,
      sym_else_if_clause,
    ACTIONS(150), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [11408] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(585), 1,
      anon_sym_LBRACK,
    ACTIONS(588), 1,
      anon_sym_AT,
    STATE(189), 2,
      sym_attribute,
      aux_sym_function_definition_repeat1,
    ACTIONS(583), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11432] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(441), 4,
      ts_builtin_sym_end,
      anon_sym_COLON_COLON,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(439), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11452] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(590), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(592), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11471] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(594), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(596), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11490] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(598), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(600), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11509] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(602), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(604), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11528] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(606), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(608), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11547] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(193), 1,
      anon_sym_PLUS,
    ACTIONS(195), 10,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_else,
  [11566] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(610), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(612), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11585] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(549), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(551), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11604] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(614), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(616), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11623] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(618), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(620), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11642] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(622), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(624), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11661] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(626), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(628), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11680] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(571), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(573), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11699] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(630), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(632), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11718] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(634), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(636), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11737] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(638), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(640), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11756] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(642), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(644), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11775] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(646), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(648), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11794] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(239), 1,
      anon_sym_PLUS,
    ACTIONS(241), 10,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_else,
  [11813] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(650), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(652), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11832] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(654), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(656), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11851] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(658), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(660), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11870] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(662), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(664), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11889] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(666), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(668), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [11908] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(235), 1,
      anon_sym_PLUS,
    ACTIONS(237), 10,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_else,
      anon_sym_RBRACE,
  [11927] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(239), 1,
      anon_sym_PLUS,
    ACTIONS(241), 10,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_else,
      anon_sym_RBRACE,
  [11946] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(193), 1,
      anon_sym_PLUS,
    ACTIONS(195), 10,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_else,
      anon_sym_RBRACE,
  [11965] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(309), 1,
      anon_sym_PLUS,
    ACTIONS(670), 1,
      anon_sym_LPAREN,
    ACTIONS(311), 9,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
  [11986] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(672), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(674), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [12005] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(561), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(563), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [12024] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(676), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(678), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [12043] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(680), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(682), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [12062] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(684), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(686), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [12081] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(235), 1,
      anon_sym_PLUS,
    ACTIONS(237), 10,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_else,
  [12100] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(309), 1,
      anon_sym_PLUS,
    ACTIONS(688), 1,
      anon_sym_LPAREN,
    ACTIONS(311), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [12121] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(690), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(692), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [12140] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(694), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(696), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [12159] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(698), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(700), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [12178] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(702), 3,
      ts_builtin_sym_end,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(704), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [12197] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(349), 1,
      anon_sym_PLUS,
    ACTIONS(351), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [12215] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(708), 2,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(706), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [12233] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(309), 1,
      anon_sym_PLUS,
    ACTIONS(311), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [12251] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(285), 1,
      anon_sym_PLUS,
    ACTIONS(287), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [12269] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(712), 2,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(710), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [12287] = 10,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(714), 1,
      sym_identifier,
    ACTIONS(716), 1,
      anon_sym_LPAREN,
    ACTIONS(718), 1,
      anon_sym_AMP_AMP,
    ACTIONS(720), 1,
      anon_sym_STAR,
    ACTIONS(722), 1,
      sym__newline,
    STATE(355), 1,
      sym_module_path,
    STATE(358), 1,
      sym_dependency_expression,
    STATE(463), 1,
      sym_dependencies,
    STATE(295), 2,
      sym_dependency,
      aux_sym_dependencies_repeat1,
  [12319] = 10,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(714), 1,
      sym_identifier,
    ACTIONS(716), 1,
      anon_sym_LPAREN,
    ACTIONS(718), 1,
      anon_sym_AMP_AMP,
    ACTIONS(720), 1,
      anon_sym_STAR,
    ACTIONS(724), 1,
      sym__newline,
    STATE(355), 1,
      sym_module_path,
    STATE(358), 1,
      sym_dependency_expression,
    STATE(464), 1,
      sym_dependencies,
    STATE(295), 2,
      sym_dependency,
      aux_sym_dependencies_repeat1,
  [12351] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(395), 1,
      sym__newline,
    ACTIONS(728), 1,
      anon_sym_PLUS,
    ACTIONS(734), 1,
      anon_sym_AMP_AMP,
    ACTIONS(736), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(726), 2,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(730), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(732), 2,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
  [12379] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(321), 1,
      anon_sym_PLUS,
    ACTIONS(323), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [12397] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(325), 1,
      anon_sym_PLUS,
    ACTIONS(327), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [12415] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(329), 1,
      anon_sym_PLUS,
    ACTIONS(331), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [12433] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(333), 1,
      anon_sym_PLUS,
    ACTIONS(335), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [12451] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(369), 1,
      anon_sym_PLUS,
    ACTIONS(371), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [12469] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(297), 1,
      anon_sym_PLUS,
    ACTIONS(299), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [12487] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(301), 1,
      anon_sym_PLUS,
    ACTIONS(303), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [12505] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(305), 1,
      anon_sym_PLUS,
    ACTIONS(307), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [12523] = 10,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(714), 1,
      sym_identifier,
    ACTIONS(716), 1,
      anon_sym_LPAREN,
    ACTIONS(718), 1,
      anon_sym_AMP_AMP,
    ACTIONS(720), 1,
      anon_sym_STAR,
    ACTIONS(738), 1,
      sym__newline,
    STATE(355), 1,
      sym_module_path,
    STATE(358), 1,
      sym_dependency_expression,
    STATE(486), 1,
      sym_dependencies,
    STATE(295), 2,
      sym_dependency,
      aux_sym_dependencies_repeat1,
  [12555] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(379), 1,
      anon_sym_RBRACE,
    ACTIONS(742), 1,
      anon_sym_PLUS,
    ACTIONS(748), 1,
      anon_sym_AMP_AMP,
    ACTIONS(750), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(740), 2,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(744), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(746), 2,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
  [12583] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(261), 1,
      anon_sym_PLUS,
    ACTIONS(263), 9,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
  [12601] = 10,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(714), 1,
      sym_identifier,
    ACTIONS(716), 1,
      anon_sym_LPAREN,
    ACTIONS(718), 1,
      anon_sym_AMP_AMP,
    ACTIONS(720), 1,
      anon_sym_STAR,
    ACTIONS(752), 1,
      sym__newline,
    STATE(355), 1,
      sym_module_path,
    STATE(358), 1,
      sym_dependency_expression,
    STATE(475), 1,
      sym_dependencies,
    STATE(295), 2,
      sym_dependency,
      aux_sym_dependencies_repeat1,
  [12633] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(395), 1,
      anon_sym_RBRACE,
    ACTIONS(742), 1,
      anon_sym_PLUS,
    ACTIONS(748), 1,
      anon_sym_AMP_AMP,
    ACTIONS(750), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(740), 2,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(744), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(746), 2,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
  [12661] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(756), 2,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(754), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [12679] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(760), 2,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(758), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [12697] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(337), 1,
      anon_sym_PLUS,
    ACTIONS(339), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [12715] = 10,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(714), 1,
      sym_identifier,
    ACTIONS(716), 1,
      anon_sym_LPAREN,
    ACTIONS(718), 1,
      anon_sym_AMP_AMP,
    ACTIONS(720), 1,
      anon_sym_STAR,
    ACTIONS(762), 1,
      sym__newline,
    STATE(355), 1,
      sym_module_path,
    STATE(358), 1,
      sym_dependency_expression,
    STATE(521), 1,
      sym_dependencies,
    STATE(295), 2,
      sym_dependency,
      aux_sym_dependencies_repeat1,
  [12747] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(341), 1,
      anon_sym_PLUS,
    ACTIONS(343), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [12765] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(345), 1,
      anon_sym_PLUS,
    ACTIONS(347), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [12783] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(265), 1,
      anon_sym_PLUS,
    ACTIONS(267), 9,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
  [12801] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(269), 1,
      anon_sym_PLUS,
    ACTIONS(271), 9,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
  [12819] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(273), 1,
      anon_sym_PLUS,
    ACTIONS(275), 9,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
  [12837] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(265), 1,
      anon_sym_PLUS,
    ACTIONS(267), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [12855] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(766), 2,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(764), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [12873] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(277), 1,
      anon_sym_PLUS,
    ACTIONS(279), 9,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
  [12891] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(728), 1,
      anon_sym_PLUS,
    ACTIONS(726), 2,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(267), 7,
      sym__newline,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [12911] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(269), 1,
      anon_sym_PLUS,
    ACTIONS(271), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [12929] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(273), 1,
      anon_sym_PLUS,
    ACTIONS(275), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [12947] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(285), 1,
      anon_sym_PLUS,
    ACTIONS(287), 9,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
  [12965] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(281), 1,
      anon_sym_PLUS,
    ACTIONS(283), 9,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
  [12983] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(728), 1,
      anon_sym_PLUS,
    ACTIONS(726), 2,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(730), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(732), 2,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
    ACTIONS(267), 3,
      sym__newline,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [13007] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(728), 1,
      anon_sym_PLUS,
    ACTIONS(734), 1,
      anon_sym_AMP_AMP,
    ACTIONS(267), 2,
      sym__newline,
      anon_sym_PIPE_PIPE,
    ACTIONS(726), 2,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(730), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(732), 2,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
  [13033] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(317), 1,
      anon_sym_PLUS,
    ACTIONS(319), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [13051] = 10,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(714), 1,
      sym_identifier,
    ACTIONS(716), 1,
      anon_sym_LPAREN,
    ACTIONS(718), 1,
      anon_sym_AMP_AMP,
    ACTIONS(720), 1,
      anon_sym_STAR,
    ACTIONS(768), 1,
      sym__newline,
    STATE(355), 1,
      sym_module_path,
    STATE(358), 1,
      sym_dependency_expression,
    STATE(536), 1,
      sym_dependencies,
    STATE(295), 2,
      sym_dependency,
      aux_sym_dependencies_repeat1,
  [13083] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(373), 1,
      anon_sym_PLUS,
    ACTIONS(375), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [13101] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(742), 1,
      anon_sym_PLUS,
    ACTIONS(740), 2,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(267), 7,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
  [13121] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(742), 1,
      anon_sym_PLUS,
    ACTIONS(740), 2,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(744), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(746), 2,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
    ACTIONS(267), 3,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
  [13145] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(742), 1,
      anon_sym_PLUS,
    ACTIONS(748), 1,
      anon_sym_AMP_AMP,
    ACTIONS(267), 2,
      anon_sym_PIPE_PIPE,
      anon_sym_RBRACE,
    ACTIONS(740), 2,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(744), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(746), 2,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
  [13171] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(309), 1,
      anon_sym_PLUS,
    ACTIONS(311), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [13189] = 10,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(714), 1,
      sym_identifier,
    ACTIONS(716), 1,
      anon_sym_LPAREN,
    ACTIONS(718), 1,
      anon_sym_AMP_AMP,
    ACTIONS(720), 1,
      anon_sym_STAR,
    ACTIONS(770), 1,
      sym__newline,
    STATE(355), 1,
      sym_module_path,
    STATE(358), 1,
      sym_dependency_expression,
    STATE(552), 1,
      sym_dependencies,
    STATE(295), 2,
      sym_dependency,
      aux_sym_dependencies_repeat1,
  [13221] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(774), 2,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(772), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [13239] = 10,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(488), 1,
      anon_sym_DOLLAR,
    ACTIONS(776), 1,
      sym_identifier,
    ACTIONS(778), 1,
      anon_sym_COLON_EQ,
    ACTIONS(780), 1,
      anon_sym_COLON,
    STATE(315), 1,
      aux_sym_parameters_repeat1,
    STATE(348), 1,
      sym_parameter,
    STATE(476), 1,
      sym_parameters,
    STATE(485), 1,
      sym_variadic_parameter,
    ACTIONS(484), 2,
      anon_sym_PLUS,
      anon_sym_STAR,
  [13271] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(365), 1,
      anon_sym_PLUS,
    ACTIONS(367), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [13289] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(277), 1,
      anon_sym_PLUS,
    ACTIONS(279), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [13307] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(309), 1,
      anon_sym_PLUS,
    ACTIONS(311), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [13325] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(357), 1,
      anon_sym_PLUS,
    ACTIONS(359), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [13343] = 8,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(379), 1,
      sym__newline,
    ACTIONS(728), 1,
      anon_sym_PLUS,
    ACTIONS(734), 1,
      anon_sym_AMP_AMP,
    ACTIONS(736), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(726), 2,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
    ACTIONS(730), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(732), 2,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
  [13371] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(261), 1,
      anon_sym_PLUS,
    ACTIONS(263), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [13389] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(361), 1,
      anon_sym_PLUS,
    ACTIONS(363), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [13407] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(313), 1,
      anon_sym_PLUS,
    ACTIONS(315), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [13425] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(281), 1,
      anon_sym_PLUS,
    ACTIONS(283), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [13443] = 10,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(714), 1,
      sym_identifier,
    ACTIONS(716), 1,
      anon_sym_LPAREN,
    ACTIONS(718), 1,
      anon_sym_AMP_AMP,
    ACTIONS(720), 1,
      anon_sym_STAR,
    ACTIONS(782), 1,
      sym__newline,
    STATE(355), 1,
      sym_module_path,
    STATE(358), 1,
      sym_dependency_expression,
    STATE(547), 1,
      sym_dependencies,
    STATE(295), 2,
      sym_dependency,
      aux_sym_dependencies_repeat1,
  [13475] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(786), 2,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(784), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [13493] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(790), 2,
      anon_sym_LBRACK,
      anon_sym_AT,
    ACTIONS(788), 8,
      anon_sym_alias,
      anon_sym_unexport,
      anon_sym_eager,
      anon_sym_export,
      anon_sym_import,
      anon_sym_mod,
      anon_sym_set,
      sym_identifier,
  [13511] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(353), 1,
      anon_sym_PLUS,
    ACTIONS(355), 9,
      sym__newline,
      anon_sym_SLASH,
      anon_sym_PLUS_PLUS,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_EQ_TILDE,
      anon_sym_BANG_TILDE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
  [13529] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(488), 1,
      anon_sym_DOLLAR,
    ACTIONS(776), 1,
      sym_identifier,
    ACTIONS(792), 1,
      anon_sym_COLON,
    STATE(315), 1,
      aux_sym_parameters_repeat1,
    STATE(348), 1,
      sym_parameter,
    STATE(485), 1,
      sym_variadic_parameter,
    STATE(504), 1,
      sym_parameters,
    ACTIONS(484), 2,
      anon_sym_PLUS,
      anon_sym_STAR,
  [13558] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(488), 1,
      anon_sym_DOLLAR,
    ACTIONS(776), 1,
      sym_identifier,
    ACTIONS(794), 1,
      anon_sym_COLON,
    STATE(315), 1,
      aux_sym_parameters_repeat1,
    STATE(348), 1,
      sym_parameter,
    STATE(485), 1,
      sym_variadic_parameter,
    STATE(510), 1,
      sym_parameters,
    ACTIONS(484), 2,
      anon_sym_PLUS,
      anon_sym_STAR,
  [13587] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(714), 1,
      sym_identifier,
    ACTIONS(716), 1,
      anon_sym_LPAREN,
    ACTIONS(718), 1,
      anon_sym_AMP_AMP,
    ACTIONS(720), 1,
      anon_sym_STAR,
    ACTIONS(796), 1,
      sym__newline,
    STATE(355), 1,
      sym_module_path,
    STATE(358), 1,
      sym_dependency_expression,
    STATE(296), 2,
      sym_dependency,
      aux_sym_dependencies_repeat1,
  [13616] = 9,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(798), 1,
      sym_identifier,
    ACTIONS(801), 1,
      anon_sym_LPAREN,
    ACTIONS(804), 1,
      anon_sym_AMP_AMP,
    ACTIONS(807), 1,
      anon_sym_STAR,
    ACTIONS(810), 1,
      sym__newline,
    STATE(355), 1,
      sym_module_path,
    STATE(358), 1,
      sym_dependency_expression,
    STATE(296), 2,
      sym_dependency,
      aux_sym_dependencies_repeat1,
  [13645] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(812), 1,
      sym_identifier,
    ACTIONS(817), 1,
      anon_sym_unexport,
    ACTIONS(819), 1,
      anon_sym_COLON,
    STATE(508), 1,
      sym__name,
    ACTIONS(815), 4,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DOLLAR,
      anon_sym_STAR,
  [13667] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(817), 1,
      anon_sym_unexport,
    ACTIONS(819), 1,
      anon_sym_COLON,
    ACTIONS(821), 1,
      sym_identifier,
    STATE(496), 1,
      sym__name,
    ACTIONS(815), 4,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DOLLAR,
      anon_sym_STAR,
  [13689] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(827), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(829), 1,
      anon_sym_LBRACE_LBRACE,
    STATE(299), 2,
      sym_interpolation,
      aux_sym__format_string_repeat1,
    ACTIONS(824), 3,
      aux_sym__format_string_token1,
      aux_sym__format_string_token2,
      sym_escape_sequence,
  [13708] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(832), 1,
      anon_sym_COLON_COLON,
    STATE(300), 1,
      aux_sym_module_path_repeat1,
    ACTIONS(441), 5,
      sym__newline,
      anon_sym_LPAREN,
      anon_sym_AMP_AMP,
      anon_sym_STAR,
      sym_identifier,
  [13725] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(837), 1,
      anon_sym_DQUOTE,
    ACTIONS(839), 1,
      anon_sym_LBRACE_LBRACE,
    STATE(303), 2,
      sym_interpolation,
      aux_sym__format_string_repeat1,
    ACTIONS(835), 3,
      aux_sym__format_string_token1,
      aux_sym__format_string_token2,
      sym_escape_sequence,
  [13744] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(837), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    ACTIONS(843), 1,
      anon_sym_LBRACE_LBRACE,
    STATE(304), 2,
      sym_interpolation,
      aux_sym__format_string_repeat1,
    ACTIONS(841), 3,
      aux_sym__format_string_token1,
      aux_sym__format_string_token2,
      sym_escape_sequence,
  [13763] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(839), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(847), 1,
      anon_sym_DQUOTE,
    STATE(318), 2,
      sym_interpolation,
      aux_sym__format_string_repeat1,
    ACTIONS(845), 3,
      aux_sym__format_string_token1,
      aux_sym__format_string_token2,
      sym_escape_sequence,
  [13782] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(843), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(847), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(299), 2,
      sym_interpolation,
      aux_sym__format_string_repeat1,
    ACTIONS(849), 3,
      aux_sym__format_string_token1,
      aux_sym__format_string_token2,
      sym_escape_sequence,
  [13801] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(843), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(851), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(299), 2,
      sym_interpolation,
      aux_sym__format_string_repeat1,
    ACTIONS(849), 3,
      aux_sym__format_string_token1,
      aux_sym__format_string_token2,
      sym_escape_sequence,
  [13820] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(853), 1,
      anon_sym_COLON_COLON,
    STATE(300), 1,
      aux_sym_module_path_repeat1,
    ACTIONS(437), 5,
      sym__newline,
      anon_sym_LPAREN,
      anon_sym_AMP_AMP,
      anon_sym_STAR,
      sym_identifier,
  [13837] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(857), 1,
      anon_sym_LPAREN,
    ACTIONS(859), 1,
      anon_sym_COLON,
    ACTIONS(855), 5,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      sym_identifier,
  [13854] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(839), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(863), 1,
      anon_sym_DQUOTE,
    STATE(310), 2,
      sym_interpolation,
      aux_sym__format_string_repeat1,
    ACTIONS(861), 3,
      aux_sym__format_string_token1,
      aux_sym__format_string_token2,
      sym_escape_sequence,
  [13873] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(843), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(863), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(311), 2,
      sym_interpolation,
      aux_sym__format_string_repeat1,
    ACTIONS(865), 3,
      aux_sym__format_string_token1,
      aux_sym__format_string_token2,
      sym_escape_sequence,
  [13892] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(839), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(867), 1,
      anon_sym_DQUOTE,
    STATE(318), 2,
      sym_interpolation,
      aux_sym__format_string_repeat1,
    ACTIONS(845), 3,
      aux_sym__format_string_token1,
      aux_sym__format_string_token2,
      sym_escape_sequence,
  [13911] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(843), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(867), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(299), 2,
      sym_interpolation,
      aux_sym__format_string_repeat1,
    ACTIONS(849), 3,
      aux_sym__format_string_token1,
      aux_sym__format_string_token2,
      sym_escape_sequence,
  [13930] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(853), 1,
      anon_sym_COLON_COLON,
    STATE(306), 1,
      aux_sym_module_path_repeat1,
    ACTIONS(869), 5,
      sym__newline,
      anon_sym_LPAREN,
      anon_sym_AMP_AMP,
      anon_sym_STAR,
      sym_identifier,
  [13947] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(839), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(873), 1,
      anon_sym_DQUOTE,
    STATE(317), 2,
      sym_interpolation,
      aux_sym__format_string_repeat1,
    ACTIONS(871), 3,
      aux_sym__format_string_token1,
      aux_sym__format_string_token2,
      sym_escape_sequence,
  [13966] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(843), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(873), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(305), 2,
      sym_interpolation,
      aux_sym__format_string_repeat1,
    ACTIONS(875), 3,
      aux_sym__format_string_token1,
      aux_sym__format_string_token2,
      sym_escape_sequence,
  [13985] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(488), 1,
      anon_sym_DOLLAR,
    ACTIONS(776), 1,
      sym_identifier,
    STATE(319), 1,
      aux_sym_parameters_repeat1,
    STATE(354), 1,
      sym_parameter,
    STATE(474), 1,
      sym_variadic_parameter,
    ACTIONS(484), 2,
      anon_sym_PLUS,
      anon_sym_STAR,
  [14008] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(859), 1,
      anon_sym_COLON,
    ACTIONS(877), 1,
      anon_sym_LPAREN,
    ACTIONS(855), 5,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      sym_identifier,
  [14025] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(839), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(851), 1,
      anon_sym_DQUOTE,
    STATE(318), 2,
      sym_interpolation,
      aux_sym__format_string_repeat1,
    ACTIONS(845), 3,
      aux_sym__format_string_token1,
      aux_sym__format_string_token2,
      sym_escape_sequence,
  [14044] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(827), 1,
      anon_sym_DQUOTE,
    ACTIONS(882), 1,
      anon_sym_LBRACE_LBRACE,
    STATE(318), 2,
      sym_interpolation,
      aux_sym__format_string_repeat1,
    ACTIONS(879), 3,
      aux_sym__format_string_token1,
      aux_sym__format_string_token2,
      sym_escape_sequence,
  [14063] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(885), 1,
      sym_identifier,
    ACTIONS(890), 1,
      anon_sym_DOLLAR,
    ACTIONS(888), 2,
      anon_sym_PLUS,
      anon_sym_STAR,
    STATE(319), 2,
      sym_parameter,
      aux_sym_parameters_repeat1,
  [14081] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(837), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(893), 1,
      aux_sym__format_string_token2,
    ACTIONS(895), 1,
      aux_sym__format_raw_string_token1,
    ACTIONS(897), 1,
      anon_sym_LBRACE_LBRACE,
    STATE(322), 2,
      sym_interpolation,
      aux_sym__format_raw_string_repeat1,
  [14101] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(847), 1,
      anon_sym_SQUOTE,
    ACTIONS(899), 1,
      aux_sym__format_string_token2,
    ACTIONS(901), 1,
      aux_sym__format_raw_string_token1,
    ACTIONS(903), 1,
      anon_sym_LBRACE_LBRACE,
    STATE(344), 2,
      sym_interpolation,
      aux_sym__format_raw_string_repeat1,
  [14121] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(847), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(897), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(905), 1,
      aux_sym__format_string_token2,
    ACTIONS(907), 1,
      aux_sym__format_raw_string_token1,
    STATE(323), 2,
      sym_interpolation,
      aux_sym__format_raw_string_repeat1,
  [14141] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(909), 1,
      aux_sym__format_string_token2,
    ACTIONS(912), 1,
      aux_sym__format_raw_string_token1,
    ACTIONS(915), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(917), 1,
      anon_sym_LBRACE_LBRACE,
    STATE(323), 2,
      sym_interpolation,
      aux_sym__format_raw_string_repeat1,
  [14161] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(922), 1,
      anon_sym_EQ,
    ACTIONS(920), 5,
      anon_sym_PLUS,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      sym_identifier,
  [14175] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(924), 1,
      aux_sym__raw_string_indented_token1,
    ACTIONS(926), 1,
      anon_sym_BQUOTE,
    ACTIONS(928), 1,
      anon_sym_LBRACE_LBRACE,
    STATE(493), 1,
      sym_command_body,
    STATE(353), 2,
      sym_interpolation,
      aux_sym_command_body_repeat1,
  [14195] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(930), 1,
      aux_sym__raw_string_indented_token1,
    ACTIONS(932), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(934), 1,
      anon_sym_LBRACE_LBRACE,
    STATE(494), 1,
      sym_command_body,
    STATE(352), 2,
      sym_interpolation,
      aux_sym_command_body_repeat1,
  [14215] = 7,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(714), 1,
      sym_identifier,
    ACTIONS(716), 1,
      anon_sym_LPAREN,
    ACTIONS(720), 1,
      anon_sym_STAR,
    STATE(346), 1,
      sym_dependency,
    STATE(355), 1,
      sym_module_path,
    STATE(358), 1,
      sym_dependency_expression,
  [14237] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(441), 6,
      sym__newline,
      anon_sym_LPAREN,
      anon_sym_COLON_COLON,
      anon_sym_AMP_AMP,
      anon_sym_STAR,
      sym_identifier,
  [14249] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(863), 1,
      anon_sym_SQUOTE,
    ACTIONS(903), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(936), 1,
      aux_sym__format_string_token2,
    ACTIONS(938), 1,
      aux_sym__format_raw_string_token1,
    STATE(331), 2,
      sym_interpolation,
      aux_sym__format_raw_string_repeat1,
  [14269] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(863), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(897), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(940), 1,
      aux_sym__format_string_token2,
    ACTIONS(942), 1,
      aux_sym__format_raw_string_token1,
    STATE(332), 2,
      sym_interpolation,
      aux_sym__format_raw_string_repeat1,
  [14289] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(867), 1,
      anon_sym_SQUOTE,
    ACTIONS(899), 1,
      aux_sym__format_string_token2,
    ACTIONS(901), 1,
      aux_sym__format_raw_string_token1,
    ACTIONS(903), 1,
      anon_sym_LBRACE_LBRACE,
    STATE(344), 2,
      sym_interpolation,
      aux_sym__format_raw_string_repeat1,
  [14309] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(867), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(897), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(905), 1,
      aux_sym__format_string_token2,
    ACTIONS(907), 1,
      aux_sym__format_raw_string_token1,
    STATE(323), 2,
      sym_interpolation,
      aux_sym__format_raw_string_repeat1,
  [14329] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(924), 1,
      aux_sym__raw_string_indented_token1,
    ACTIONS(928), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(944), 1,
      anon_sym_BQUOTE,
    STATE(505), 1,
      sym_command_body,
    STATE(353), 2,
      sym_interpolation,
      aux_sym_command_body_repeat1,
  [14349] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(924), 1,
      aux_sym__raw_string_indented_token1,
    ACTIONS(928), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(946), 1,
      anon_sym_BQUOTE,
    STATE(529), 1,
      sym_command_body,
    STATE(353), 2,
      sym_interpolation,
      aux_sym_command_body_repeat1,
  [14369] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(930), 1,
      aux_sym__raw_string_indented_token1,
    ACTIONS(934), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(948), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    STATE(530), 1,
      sym_command_body,
    STATE(352), 2,
      sym_interpolation,
      aux_sym_command_body_repeat1,
  [14389] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(930), 1,
      aux_sym__raw_string_indented_token1,
    ACTIONS(934), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(950), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    STATE(512), 1,
      sym_command_body,
    STATE(352), 2,
      sym_interpolation,
      aux_sym_command_body_repeat1,
  [14409] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(851), 1,
      anon_sym_SQUOTE,
    ACTIONS(899), 1,
      aux_sym__format_string_token2,
    ACTIONS(901), 1,
      aux_sym__format_raw_string_token1,
    ACTIONS(903), 1,
      anon_sym_LBRACE_LBRACE,
    STATE(344), 2,
      sym_interpolation,
      aux_sym__format_raw_string_repeat1,
  [14429] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(954), 1,
      anon_sym_EQ,
    ACTIONS(952), 5,
      anon_sym_PLUS,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      sym_identifier,
  [14443] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(873), 1,
      anon_sym_SQUOTE,
    ACTIONS(903), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(956), 1,
      aux_sym__format_string_token2,
    ACTIONS(958), 1,
      aux_sym__format_raw_string_token1,
    STATE(337), 2,
      sym_interpolation,
      aux_sym__format_raw_string_repeat1,
  [14463] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(851), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(897), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(905), 1,
      aux_sym__format_string_token2,
    ACTIONS(907), 1,
      aux_sym__format_raw_string_token1,
    STATE(323), 2,
      sym_interpolation,
      aux_sym__format_raw_string_repeat1,
  [14483] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(819), 1,
      anon_sym_COLON,
    ACTIONS(815), 5,
      anon_sym_COLON_EQ,
      anon_sym_PLUS,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      sym_identifier,
  [14497] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(873), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(897), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(960), 1,
      aux_sym__format_string_token2,
    ACTIONS(962), 1,
      aux_sym__format_raw_string_token1,
    STATE(340), 2,
      sym_interpolation,
      aux_sym__format_raw_string_repeat1,
  [14517] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(837), 1,
      anon_sym_SQUOTE,
    ACTIONS(903), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(964), 1,
      aux_sym__format_string_token2,
    ACTIONS(966), 1,
      aux_sym__format_raw_string_token1,
    STATE(321), 2,
      sym_interpolation,
      aux_sym__format_raw_string_repeat1,
  [14537] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(915), 1,
      anon_sym_SQUOTE,
    ACTIONS(968), 1,
      aux_sym__format_string_token2,
    ACTIONS(971), 1,
      aux_sym__format_raw_string_token1,
    ACTIONS(974), 1,
      anon_sym_LBRACE_LBRACE,
    STATE(344), 2,
      sym_interpolation,
      aux_sym__format_raw_string_repeat1,
  [14557] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(977), 5,
      anon_sym_PLUS,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      sym_identifier,
  [14568] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(810), 5,
      sym__newline,
      anon_sym_LPAREN,
      anon_sym_AMP_AMP,
      anon_sym_STAR,
      sym_identifier,
  [14579] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(979), 5,
      aux_sym__format_string_token1,
      aux_sym__format_string_token2,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
      sym_escape_sequence,
      anon_sym_LBRACE_LBRACE,
  [14590] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(983), 1,
      anon_sym_COLON,
    ACTIONS(981), 4,
      anon_sym_PLUS,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      sym_identifier,
  [14603] = 6,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(985), 1,
      anon_sym_LPAREN,
    ACTIONS(987), 1,
      anon_sym_COMMA,
    ACTIONS(989), 1,
      anon_sym_RBRACK,
    ACTIONS(991), 1,
      anon_sym_COLON,
    STATE(418), 1,
      aux_sym_attribute_repeat2,
  [14622] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(993), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(996), 1,
      sym__newline,
    ACTIONS(998), 1,
      sym_text,
    STATE(350), 2,
      sym_interpolation,
      aux_sym_recipe_line_repeat1,
  [14639] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1001), 1,
      aux_sym__raw_string_indented_token1,
    ACTIONS(1004), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    ACTIONS(1006), 1,
      anon_sym_LBRACE_LBRACE,
    STATE(351), 2,
      sym_interpolation,
      aux_sym_command_body_repeat1,
  [14656] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(934), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(1009), 1,
      aux_sym__raw_string_indented_token1,
    ACTIONS(1011), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
    STATE(351), 2,
      sym_interpolation,
      aux_sym_command_body_repeat1,
  [14673] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(928), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(1011), 1,
      anon_sym_BQUOTE,
    ACTIONS(1013), 1,
      aux_sym__raw_string_indented_token1,
    STATE(361), 2,
      sym_interpolation,
      aux_sym_command_body_repeat1,
  [14690] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1015), 1,
      anon_sym_COLON,
    ACTIONS(981), 4,
      anon_sym_PLUS,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      sym_identifier,
  [14703] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(869), 5,
      sym__newline,
      anon_sym_LPAREN,
      anon_sym_AMP_AMP,
      anon_sym_STAR,
      sym_identifier,
  [14714] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(979), 5,
      aux_sym__format_string_token1,
      aux_sym__format_string_token2,
      anon_sym_DQUOTE,
      sym_escape_sequence,
      anon_sym_LBRACE_LBRACE,
  [14725] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1017), 5,
      sym__newline,
      anon_sym_LPAREN,
      anon_sym_AMP_AMP,
      anon_sym_STAR,
      sym_identifier,
  [14736] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1019), 5,
      sym__newline,
      anon_sym_LPAREN,
      anon_sym_AMP_AMP,
      anon_sym_STAR,
      sym_identifier,
  [14747] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1021), 5,
      sym__newline,
      anon_sym_LPAREN,
      anon_sym_AMP_AMP,
      anon_sym_STAR,
      sym_identifier,
  [14758] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1023), 5,
      anon_sym_PLUS,
      anon_sym_COLON,
      anon_sym_DOLLAR,
      anon_sym_STAR,
      sym_identifier,
  [14769] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1004), 1,
      anon_sym_BQUOTE,
    ACTIONS(1025), 1,
      aux_sym__raw_string_indented_token1,
    ACTIONS(1028), 1,
      anon_sym_LBRACE_LBRACE,
    STATE(361), 2,
      sym_interpolation,
      aux_sym_command_body_repeat1,
  [14786] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1031), 5,
      sym__newline,
      anon_sym_LPAREN,
      anon_sym_AMP_AMP,
      anon_sym_STAR,
      sym_identifier,
  [14797] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(450), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(1033), 1,
      sym__newline,
    ACTIONS(1035), 1,
      sym_text,
    STATE(350), 2,
      sym_interpolation,
      aux_sym_recipe_line_repeat1,
  [14814] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1037), 5,
      sym__newline,
      anon_sym_LPAREN,
      anon_sym_AMP_AMP,
      anon_sym_STAR,
      sym_identifier,
  [14825] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(450), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(1035), 1,
      sym_text,
    ACTIONS(1039), 1,
      sym__newline,
    STATE(350), 2,
      sym_interpolation,
      aux_sym_recipe_line_repeat1,
  [14842] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1041), 1,
      anon_sym_DQUOTE,
    STATE(369), 1,
      aux_sym__string_repeat1,
    ACTIONS(1043), 2,
      aux_sym__string_token1,
      sym_escape_sequence,
  [14856] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(450), 1,
      anon_sym_LBRACE_LBRACE,
    ACTIONS(1045), 1,
      sym_text,
    STATE(365), 2,
      sym_interpolation,
      aux_sym_recipe_line_repeat1,
  [14870] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1047), 1,
      sym_identifier,
    ACTIONS(1049), 1,
      anon_sym_unexport,
    ACTIONS(1051), 1,
      anon_sym_import,
    STATE(294), 1,
      sym__name,
  [14886] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE,
    STATE(371), 1,
      aux_sym__string_repeat1,
    ACTIONS(1055), 2,
      aux_sym__string_token1,
      sym_escape_sequence,
  [14900] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1053), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(372), 1,
      aux_sym__shell_expanded_string_indented_repeat1,
    ACTIONS(1057), 2,
      aux_sym__shell_expanded_string_indented_token1,
      sym_escape_sequence,
  [14914] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1059), 1,
      anon_sym_DQUOTE,
    STATE(371), 1,
      aux_sym__string_repeat1,
    ACTIONS(1061), 2,
      aux_sym__string_token1,
      sym_escape_sequence,
  [14928] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1064), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(372), 1,
      aux_sym__shell_expanded_string_indented_repeat1,
    ACTIONS(1066), 2,
      aux_sym__shell_expanded_string_indented_token1,
      sym_escape_sequence,
  [14942] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1049), 1,
      anon_sym_unexport,
    ACTIONS(1069), 1,
      sym_identifier,
    STATE(214), 1,
      sym__assignment,
    STATE(491), 1,
      sym__name,
  [14958] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1071), 1,
      aux_sym__format_raw_string_token1,
    ACTIONS(979), 3,
      aux_sym__format_string_token2,
      anon_sym_SQUOTE,
      anon_sym_LBRACE_LBRACE,
  [14970] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1073), 1,
      anon_sym_LPAREN,
    ACTIONS(1077), 1,
      anon_sym_COLON,
    ACTIONS(1075), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [14984] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1049), 1,
      anon_sym_unexport,
    ACTIONS(1069), 1,
      sym_identifier,
    STATE(222), 1,
      sym__assignment,
    STATE(491), 1,
      sym__name,
  [15000] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1081), 1,
      anon_sym_COMMA,
    STATE(377), 1,
      aux_sym_list_elements_repeat1,
    ACTIONS(1079), 2,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
  [15014] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1071), 1,
      aux_sym__format_raw_string_token1,
    ACTIONS(979), 3,
      aux_sym__format_string_token2,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
      anon_sym_LBRACE_LBRACE,
  [15026] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1084), 1,
      anon_sym_DQUOTE,
    STATE(381), 1,
      aux_sym__string_repeat1,
    ACTIONS(1086), 2,
      aux_sym__string_token1,
      sym_escape_sequence,
  [15040] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1084), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(382), 1,
      aux_sym__shell_expanded_string_indented_repeat1,
    ACTIONS(1088), 2,
      aux_sym__shell_expanded_string_indented_token1,
      sym_escape_sequence,
  [15054] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1090), 1,
      anon_sym_DQUOTE,
    STATE(371), 1,
      aux_sym__string_repeat1,
    ACTIONS(1055), 2,
      aux_sym__string_token1,
      sym_escape_sequence,
  [15068] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1090), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(372), 1,
      aux_sym__shell_expanded_string_indented_repeat1,
    ACTIONS(1057), 2,
      aux_sym__shell_expanded_string_indented_token1,
      sym_escape_sequence,
  [15082] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1049), 1,
      anon_sym_unexport,
    ACTIONS(1069), 1,
      sym_identifier,
    STATE(192), 1,
      sym__assignment,
    STATE(491), 1,
      sym__name,
  [15098] = 5,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1049), 1,
      anon_sym_unexport,
    ACTIONS(1069), 1,
      sym_identifier,
    STATE(197), 1,
      sym__assignment,
    STATE(491), 1,
      sym__name,
  [15114] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1092), 1,
      anon_sym_DQUOTE,
    STATE(387), 1,
      aux_sym__string_repeat1,
    ACTIONS(1094), 2,
      aux_sym__string_token1,
      sym_escape_sequence,
  [15128] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1092), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(388), 1,
      aux_sym__shell_expanded_string_indented_repeat1,
    ACTIONS(1096), 2,
      aux_sym__shell_expanded_string_indented_token1,
      sym_escape_sequence,
  [15142] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1098), 1,
      anon_sym_DQUOTE,
    STATE(371), 1,
      aux_sym__string_repeat1,
    ACTIONS(1055), 2,
      aux_sym__string_token1,
      sym_escape_sequence,
  [15156] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1098), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(372), 1,
      aux_sym__shell_expanded_string_indented_repeat1,
    ACTIONS(1057), 2,
      aux_sym__shell_expanded_string_indented_token1,
      sym_escape_sequence,
  [15170] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1041), 1,
      anon_sym_DQUOTE_DQUOTE_DQUOTE,
    STATE(370), 1,
      aux_sym__shell_expanded_string_indented_repeat1,
    ACTIONS(1100), 2,
      aux_sym__shell_expanded_string_indented_token1,
      sym_escape_sequence,
  [15184] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1041), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1102), 1,
      aux_sym__raw_string_indented_token1,
    STATE(424), 1,
      aux_sym__raw_string_indented_repeat1,
  [15197] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1104), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1106), 1,
      aux_sym__raw_string_indented_token1,
    STATE(391), 1,
      aux_sym__raw_string_indented_repeat1,
  [15210] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1109), 1,
      anon_sym_RPAREN,
    ACTIONS(1111), 1,
      anon_sym_COMMA,
    STATE(411), 1,
      aux_sym_attribute_repeat1,
  [15223] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1113), 1,
      anon_sym_RPAREN,
    ACTIONS(1115), 1,
      anon_sym_COMMA,
    STATE(393), 1,
      aux_sym_function_parameters_repeat1,
  [15236] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1118), 1,
      sym_identifier,
    ACTIONS(1120), 1,
      anon_sym_RPAREN,
    STATE(514), 1,
      sym_function_parameters,
  [15249] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(987), 1,
      anon_sym_COMMA,
    ACTIONS(1122), 1,
      anon_sym_RBRACK,
    STATE(421), 1,
      aux_sym_attribute_repeat2,
  [15262] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(987), 1,
      anon_sym_COMMA,
    ACTIONS(1124), 1,
      anon_sym_RBRACK,
    STATE(400), 1,
      aux_sym_attribute_repeat2,
  [15275] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(987), 1,
      anon_sym_COMMA,
    ACTIONS(1126), 1,
      anon_sym_RBRACK,
    STATE(415), 1,
      aux_sym_attribute_repeat2,
  [15288] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1128), 1,
      anon_sym_RPAREN,
    ACTIONS(1130), 1,
      anon_sym_COMMA,
    STATE(398), 1,
      aux_sym_attribute_repeat1,
  [15301] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1111), 1,
      anon_sym_COMMA,
    ACTIONS(1133), 1,
      anon_sym_RPAREN,
    STATE(417), 1,
      aux_sym_attribute_repeat1,
  [15314] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1135), 1,
      anon_sym_COMMA,
    ACTIONS(1138), 1,
      anon_sym_RBRACK,
    STATE(400), 1,
      aux_sym_attribute_repeat2,
  [15327] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1140), 1,
      anon_sym_if,
    ACTIONS(1142), 1,
      anon_sym_LBRACE,
    STATE(233), 1,
      sym__braced_expr,
  [15340] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1079), 3,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [15349] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1144), 1,
      sym_identifier,
    ACTIONS(1146), 1,
      aux_sym__shebang_with_lang_token2,
    STATE(410), 1,
      aux_sym__shebang_with_lang_repeat1,
  [15362] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(257), 1,
      anon_sym_RPAREN,
    ACTIONS(1148), 1,
      anon_sym_COMMA,
    STATE(377), 1,
      aux_sym_list_elements_repeat1,
  [15375] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(987), 1,
      anon_sym_COMMA,
    ACTIONS(1150), 1,
      anon_sym_RBRACK,
    STATE(396), 1,
      aux_sym_attribute_repeat2,
  [15388] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1152), 1,
      anon_sym_RPAREN,
    ACTIONS(1154), 1,
      anon_sym_COMMA,
    STATE(393), 1,
      aux_sym_function_parameters_repeat1,
  [15401] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(979), 3,
      aux_sym__raw_string_indented_token1,
      anon_sym_BQUOTE,
      anon_sym_LBRACE_LBRACE,
  [15410] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(979), 3,
      aux_sym__raw_string_indented_token1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
      anon_sym_LBRACE_LBRACE,
  [15419] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1071), 3,
      sym__newline,
      sym_text,
      anon_sym_LBRACE_LBRACE,
  [15428] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1156), 1,
      sym_identifier,
    ACTIONS(1158), 1,
      aux_sym__shebang_with_lang_token2,
    STATE(410), 1,
      aux_sym__shebang_with_lang_repeat1,
  [15441] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1111), 1,
      anon_sym_COMMA,
    ACTIONS(1161), 1,
      anon_sym_RPAREN,
    STATE(398), 1,
      aux_sym_attribute_repeat1,
  [15454] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1163), 3,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_LBRACE,
  [15463] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1084), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1165), 1,
      aux_sym__raw_string_indented_token1,
    STATE(416), 1,
      aux_sym__raw_string_indented_repeat1,
  [15476] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1167), 1,
      aux_sym__shebang_with_lang_token1,
    ACTIONS(1169), 1,
      sym__opaque_shebang,
    STATE(560), 1,
      sym__shebang_with_lang,
  [15489] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(987), 1,
      anon_sym_COMMA,
    ACTIONS(1171), 1,
      anon_sym_RBRACK,
    STATE(400), 1,
      aux_sym_attribute_repeat2,
  [15502] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1090), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1173), 1,
      aux_sym__raw_string_indented_token1,
    STATE(391), 1,
      aux_sym__raw_string_indented_repeat1,
  [15515] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1111), 1,
      anon_sym_COMMA,
    ACTIONS(1175), 1,
      anon_sym_RPAREN,
    STATE(398), 1,
      aux_sym_attribute_repeat1,
  [15528] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(987), 1,
      anon_sym_COMMA,
    ACTIONS(1177), 1,
      anon_sym_RBRACK,
    STATE(400), 1,
      aux_sym_attribute_repeat2,
  [15541] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1179), 1,
      anon_sym_COMMA,
    ACTIONS(1181), 1,
      anon_sym_RBRACK,
    STATE(431), 1,
      aux_sym_list_elements_repeat1,
  [15554] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1183), 1,
      anon_sym_if,
    ACTIONS(1185), 1,
      anon_sym_LBRACE,
    STATE(39), 1,
      sym__braced_expr,
  [15567] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(987), 1,
      anon_sym_COMMA,
    ACTIONS(1187), 1,
      anon_sym_RBRACK,
    STATE(400), 1,
      aux_sym_attribute_repeat2,
  [15580] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1189), 1,
      anon_sym_RPAREN,
    ACTIONS(1191), 1,
      anon_sym_COMMA,
    STATE(404), 1,
      aux_sym_list_elements_repeat1,
  [15593] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1092), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1193), 1,
      aux_sym__raw_string_indented_token1,
    STATE(427), 1,
      aux_sym__raw_string_indented_repeat1,
  [15606] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1053), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1173), 1,
      aux_sym__raw_string_indented_token1,
    STATE(391), 1,
      aux_sym__raw_string_indented_repeat1,
  [15619] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1195), 1,
      sym_identifier,
    ACTIONS(1197), 1,
      aux_sym__shebang_with_lang_token2,
    STATE(403), 1,
      aux_sym__shebang_with_lang_repeat1,
  [15632] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1118), 1,
      sym_identifier,
    ACTIONS(1199), 1,
      anon_sym_RPAREN,
    STATE(544), 1,
      sym_function_parameters,
  [15645] = 4,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1098), 1,
      anon_sym_SQUOTE_SQUOTE_SQUOTE,
    ACTIONS(1173), 1,
      aux_sym__raw_string_indented_token1,
    STATE(391), 1,
      aux_sym__raw_string_indented_repeat1,
  [15658] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1201), 1,
      anon_sym_RPAREN,
    ACTIONS(1203), 1,
      anon_sym_COMMA,
    STATE(406), 1,
      aux_sym_function_parameters_repeat1,
  [15671] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1205), 1,
      anon_sym_if,
    ACTIONS(1207), 1,
      anon_sym_LBRACE,
    STATE(266), 1,
      sym__braced_expr,
  [15684] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(488), 1,
      anon_sym_DOLLAR,
    ACTIONS(776), 1,
      sym_identifier,
    STATE(506), 1,
      sym_parameter,
  [15697] = 4,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(251), 1,
      anon_sym_RBRACK,
    ACTIONS(1209), 1,
      anon_sym_COMMA,
    STATE(377), 1,
      aux_sym_list_elements_repeat1,
  [15710] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1211), 1,
      anon_sym_RPAREN,
    ACTIONS(1213), 1,
      anon_sym_COMMA,
  [15720] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1185), 1,
      anon_sym_LBRACE,
    STATE(4), 1,
      sym__braced_expr,
  [15730] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1215), 1,
      anon_sym_RPAREN,
    ACTIONS(1217), 1,
      anon_sym_COMMA,
  [15740] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1219), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [15748] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1221), 1,
      anon_sym_RPAREN,
    ACTIONS(1223), 1,
      anon_sym_COMMA,
  [15758] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1185), 1,
      anon_sym_LBRACE,
    STATE(14), 1,
      sym__braced_expr,
  [15768] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1142), 1,
      anon_sym_LBRACE,
    STATE(179), 1,
      sym__braced_expr,
  [15778] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1225), 1,
      anon_sym_RPAREN,
    ACTIONS(1227), 1,
      anon_sym_COMMA,
  [15788] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1229), 1,
      sym_identifier,
    ACTIONS(1231), 1,
      anon_sym_env,
  [15798] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1233), 1,
      anon_sym_RPAREN,
    ACTIONS(1235), 1,
      anon_sym_COMMA,
  [15808] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1237), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [15816] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1239), 1,
      anon_sym_COLON_EQ,
    ACTIONS(1241), 1,
      sym__newline,
  [15826] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1243), 1,
      sym_identifier,
    ACTIONS(1245), 1,
      anon_sym_RPAREN,
  [15836] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1247), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [15844] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1249), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [15852] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1251), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [15860] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1142), 1,
      anon_sym_LBRACE,
    STATE(196), 1,
      sym__braced_expr,
  [15870] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1207), 1,
      anon_sym_LBRACE,
    STATE(177), 1,
      sym__braced_expr,
  [15880] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1253), 1,
      sym_identifier,
    STATE(8), 1,
      sym_module_path,
  [15890] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1255), 1,
      sym_identifier,
    STATE(7), 1,
      sym_module_path,
  [15900] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1128), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [15908] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1257), 1,
      anon_sym_RPAREN,
    ACTIONS(1259), 1,
      anon_sym_COMMA,
  [15918] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1207), 1,
      anon_sym_LBRACE,
    STATE(217), 1,
      sym__braced_expr,
  [15928] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 1,
      sym_identifier,
    ACTIONS(1263), 1,
      anon_sym_QMARK,
  [15938] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1243), 1,
      sym_identifier,
    ACTIONS(1265), 1,
      anon_sym_RPAREN,
  [15948] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1267), 2,
      sym_text,
      anon_sym_LBRACE_LBRACE,
  [15956] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1269), 1,
      anon_sym_COLON_EQ,
    ACTIONS(1271), 1,
      sym__newline,
  [15966] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1273), 1,
      sym_identifier,
    STATE(220), 1,
      sym_module_path,
  [15976] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1275), 1,
      sym_identifier,
    ACTIONS(1277), 1,
      anon_sym_QMARK,
  [15986] = 3,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1279), 1,
      sym_identifier,
    STATE(198), 1,
      sym_module_path,
  [15996] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1281), 1,
      sym_identifier,
  [16003] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1283), 1,
      sym__newline,
  [16010] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1285), 1,
      sym__newline,
  [16017] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1287), 1,
      anon_sym_COLON_EQ,
  [16024] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1289), 1,
      sym__newline,
  [16031] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1291), 1,
      aux_sym__shebang_with_lang_token3,
  [16038] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1293), 1,
      anon_sym_LPAREN,
  [16045] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1295), 1,
      anon_sym_RPAREN,
  [16052] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1297), 1,
      sym__newline,
  [16059] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1299), 1,
      sym__newline,
  [16066] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1301), 1,
      anon_sym_COLON_EQ,
  [16073] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1303), 1,
      sym__newline,
  [16080] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1015), 1,
      anon_sym_COLON,
  [16087] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1305), 1,
      sym__newline,
  [16094] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1307), 1,
      anon_sym_COLON,
  [16101] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1309), 1,
      sym__newline,
  [16108] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1311), 1,
      sym__newline,
  [16115] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1313), 1,
      sym__newline,
  [16122] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1315), 1,
      sym__newline,
  [16129] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1317), 1,
      sym__newline,
  [16136] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1319), 1,
      sym_identifier,
  [16143] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1321), 1,
      ts_builtin_sym_end,
  [16150] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1323), 1,
      anon_sym_COLON_EQ,
  [16157] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(983), 1,
      anon_sym_COLON,
  [16164] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(752), 1,
      sym__newline,
  [16171] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1325), 1,
      anon_sym_COLON_EQ,
  [16178] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1327), 1,
      sym_identifier,
  [16185] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1329), 1,
      anon_sym_RBRACE_RBRACE,
  [16192] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1331), 1,
      anon_sym_RPAREN,
  [16199] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(778), 1,
      anon_sym_COLON_EQ,
  [16206] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1333), 1,
      anon_sym_RBRACK,
  [16213] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1335), 1,
      anon_sym_BQUOTE,
  [16220] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1337), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
  [16227] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1339), 1,
      anon_sym_RBRACK,
  [16234] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1341), 1,
      sym__newline,
  [16241] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1343), 1,
      anon_sym_RPAREN,
  [16248] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1345), 1,
      anon_sym_RBRACE,
  [16255] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1347), 1,
      sym__newline,
  [16262] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1349), 1,
      sym_identifier,
  [16269] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1351), 1,
      sym__newline,
  [16276] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1353), 1,
      anon_sym_RPAREN,
  [16283] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(815), 1,
      sym__newline,
  [16290] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1355), 1,
      anon_sym_COLON,
  [16297] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1357), 1,
      anon_sym_BQUOTE,
  [16304] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1359), 1,
      anon_sym_COLON,
  [16311] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1361), 1,
      sym__newline,
  [16318] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1363), 1,
      sym__newline,
  [16325] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1365), 1,
      sym__newline,
  [16332] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1367), 1,
      anon_sym_COLON,
  [16339] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1261), 1,
      sym_identifier,
  [16346] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1369), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
  [16353] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1371), 1,
      sym_identifier,
  [16360] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1373), 1,
      anon_sym_RPAREN,
  [16367] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1375), 1,
      sym__newline,
  [16374] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1377), 1,
      sym__newline,
  [16381] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1379), 1,
      aux_sym__shebang_with_lang_token3,
  [16388] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1381), 1,
      sym__newline,
  [16395] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1383), 1,
      anon_sym_COLON_EQ,
  [16402] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1385), 1,
      sym__newline,
  [16409] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(722), 1,
      sym__newline,
  [16416] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1243), 1,
      sym_identifier,
  [16423] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1140), 1,
      anon_sym_if,
  [16430] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1387), 1,
      sym_identifier,
  [16437] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1389), 1,
      anon_sym_RBRACE_RBRACE,
  [16444] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1391), 1,
      anon_sym_RPAREN,
  [16451] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1393), 1,
      sym__newline,
  [16458] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1395), 1,
      anon_sym_RBRACK,
  [16465] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1397), 1,
      anon_sym_BQUOTE,
  [16472] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1399), 1,
      anon_sym_BQUOTE_BQUOTE_BQUOTE,
  [16479] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1401), 1,
      anon_sym_RBRACE,
  [16486] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1403), 1,
      sym__newline,
  [16493] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1405), 1,
      anon_sym_RPAREN,
  [16500] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1407), 1,
      anon_sym_RBRACE,
  [16507] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1409), 1,
      sym_identifier,
  [16514] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(724), 1,
      sym__newline,
  [16521] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1411), 1,
      sym_identifier,
  [16528] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1413), 1,
      anon_sym_RPAREN,
  [16535] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1415), 1,
      anon_sym_RBRACE_RBRACE,
  [16542] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1417), 1,
      anon_sym_RBRACE_RBRACE,
  [16549] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1419), 1,
      anon_sym_RBRACE_RBRACE,
  [16556] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1421), 1,
      anon_sym_RBRACE_RBRACE,
  [16563] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1423), 1,
      anon_sym_RPAREN,
  [16570] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1425), 1,
      anon_sym_RPAREN,
  [16577] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1427), 1,
      anon_sym_COLON,
  [16584] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1429), 1,
      anon_sym_COLON_EQ,
  [16591] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1431), 1,
      sym__newline,
  [16598] = 2,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(1433), 1,
      aux_sym__shebang_with_lang_token3,
  [16605] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1435), 1,
      anon_sym_RPAREN,
  [16612] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1183), 1,
      anon_sym_if,
  [16619] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1437), 1,
      sym__newline,
  [16626] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(782), 1,
      sym__newline,
  [16633] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1439), 1,
      sym__newline,
  [16640] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1441), 1,
      sym__newline,
  [16647] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1443), 1,
      sym__newline,
  [16654] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1445), 1,
      sym__newline,
  [16661] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1447), 1,
      anon_sym_LPAREN,
  [16668] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1205), 1,
      anon_sym_if,
  [16675] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1449), 1,
      sym_identifier,
  [16682] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1451), 1,
      sym__newline,
  [16689] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1453), 1,
      anon_sym_RBRACE_RBRACE,
  [16696] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1455), 1,
      sym_identifier,
  [16703] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1457), 1,
      anon_sym_LPAREN,
  [16710] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1459), 1,
      anon_sym_LPAREN,
  [16717] = 2,
    ACTIONS(71), 1,
      sym_comment,
    ACTIONS(1461), 1,
      sym_identifier,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(2)] = 0,
  [SMALL_STATE(3)] = 101,
  [SMALL_STATE(4)] = 202,
  [SMALL_STATE(5)] = 258,
  [SMALL_STATE(6)] = 314,
  [SMALL_STATE(7)] = 409,
  [SMALL_STATE(8)] = 504,
  [SMALL_STATE(9)] = 599,
  [SMALL_STATE(10)] = 694,
  [SMALL_STATE(11)] = 789,
  [SMALL_STATE(12)] = 842,
  [SMALL_STATE(13)] = 936,
  [SMALL_STATE(14)] = 1030,
  [SMALL_STATE(15)] = 1075,
  [SMALL_STATE(16)] = 1168,
  [SMALL_STATE(17)] = 1261,
  [SMALL_STATE(18)] = 1306,
  [SMALL_STATE(19)] = 1351,
  [SMALL_STATE(20)] = 1444,
  [SMALL_STATE(21)] = 1537,
  [SMALL_STATE(22)] = 1630,
  [SMALL_STATE(23)] = 1723,
  [SMALL_STATE(24)] = 1813,
  [SMALL_STATE(25)] = 1901,
  [SMALL_STATE(26)] = 1991,
  [SMALL_STATE(27)] = 2081,
  [SMALL_STATE(28)] = 2171,
  [SMALL_STATE(29)] = 2261,
  [SMALL_STATE(30)] = 2305,
  [SMALL_STATE(31)] = 2393,
  [SMALL_STATE(32)] = 2437,
  [SMALL_STATE(33)] = 2481,
  [SMALL_STATE(34)] = 2525,
  [SMALL_STATE(35)] = 2569,
  [SMALL_STATE(36)] = 2659,
  [SMALL_STATE(37)] = 2749,
  [SMALL_STATE(38)] = 2837,
  [SMALL_STATE(39)] = 2881,
  [SMALL_STATE(40)] = 2925,
  [SMALL_STATE(41)] = 3015,
  [SMALL_STATE(42)] = 3105,
  [SMALL_STATE(43)] = 3195,
  [SMALL_STATE(44)] = 3285,
  [SMALL_STATE(45)] = 3375,
  [SMALL_STATE(46)] = 3465,
  [SMALL_STATE(47)] = 3552,
  [SMALL_STATE(48)] = 3639,
  [SMALL_STATE(49)] = 3726,
  [SMALL_STATE(50)] = 3813,
  [SMALL_STATE(51)] = 3900,
  [SMALL_STATE(52)] = 3987,
  [SMALL_STATE(53)] = 4074,
  [SMALL_STATE(54)] = 4161,
  [SMALL_STATE(55)] = 4248,
  [SMALL_STATE(56)] = 4335,
  [SMALL_STATE(57)] = 4422,
  [SMALL_STATE(58)] = 4509,
  [SMALL_STATE(59)] = 4596,
  [SMALL_STATE(60)] = 4683,
  [SMALL_STATE(61)] = 4770,
  [SMALL_STATE(62)] = 4857,
  [SMALL_STATE(63)] = 4944,
  [SMALL_STATE(64)] = 5031,
  [SMALL_STATE(65)] = 5118,
  [SMALL_STATE(66)] = 5205,
  [SMALL_STATE(67)] = 5292,
  [SMALL_STATE(68)] = 5379,
  [SMALL_STATE(69)] = 5466,
  [SMALL_STATE(70)] = 5553,
  [SMALL_STATE(71)] = 5640,
  [SMALL_STATE(72)] = 5726,
  [SMALL_STATE(73)] = 5812,
  [SMALL_STATE(74)] = 5898,
  [SMALL_STATE(75)] = 5984,
  [SMALL_STATE(76)] = 6065,
  [SMALL_STATE(77)] = 6146,
  [SMALL_STATE(78)] = 6227,
  [SMALL_STATE(79)] = 6308,
  [SMALL_STATE(80)] = 6389,
  [SMALL_STATE(81)] = 6470,
  [SMALL_STATE(82)] = 6551,
  [SMALL_STATE(83)] = 6632,
  [SMALL_STATE(84)] = 6713,
  [SMALL_STATE(85)] = 6794,
  [SMALL_STATE(86)] = 6835,
  [SMALL_STATE(87)] = 6876,
  [SMALL_STATE(88)] = 6917,
  [SMALL_STATE(89)] = 6958,
  [SMALL_STATE(90)] = 6999,
  [SMALL_STATE(91)] = 7040,
  [SMALL_STATE(92)] = 7121,
  [SMALL_STATE(93)] = 7162,
  [SMALL_STATE(94)] = 7203,
  [SMALL_STATE(95)] = 7244,
  [SMALL_STATE(96)] = 7285,
  [SMALL_STATE(97)] = 7366,
  [SMALL_STATE(98)] = 7447,
  [SMALL_STATE(99)] = 7528,
  [SMALL_STATE(100)] = 7609,
  [SMALL_STATE(101)] = 7650,
  [SMALL_STATE(102)] = 7691,
  [SMALL_STATE(103)] = 7732,
  [SMALL_STATE(104)] = 7773,
  [SMALL_STATE(105)] = 7814,
  [SMALL_STATE(106)] = 7855,
  [SMALL_STATE(107)] = 7896,
  [SMALL_STATE(108)] = 7937,
  [SMALL_STATE(109)] = 7978,
  [SMALL_STATE(110)] = 8059,
  [SMALL_STATE(111)] = 8140,
  [SMALL_STATE(112)] = 8221,
  [SMALL_STATE(113)] = 8302,
  [SMALL_STATE(114)] = 8383,
  [SMALL_STATE(115)] = 8424,
  [SMALL_STATE(116)] = 8476,
  [SMALL_STATE(117)] = 8520,
  [SMALL_STATE(118)] = 8560,
  [SMALL_STATE(119)] = 8608,
  [SMALL_STATE(120)] = 8658,
  [SMALL_STATE(121)] = 8710,
  [SMALL_STATE(122)] = 8784,
  [SMALL_STATE(123)] = 8858,
  [SMALL_STATE(124)] = 8932,
  [SMALL_STATE(125)] = 9006,
  [SMALL_STATE(126)] = 9080,
  [SMALL_STATE(127)] = 9154,
  [SMALL_STATE(128)] = 9228,
  [SMALL_STATE(129)] = 9264,
  [SMALL_STATE(130)] = 9300,
  [SMALL_STATE(131)] = 9364,
  [SMALL_STATE(132)] = 9400,
  [SMALL_STATE(133)] = 9464,
  [SMALL_STATE(134)] = 9500,
  [SMALL_STATE(135)] = 9564,
  [SMALL_STATE(136)] = 9628,
  [SMALL_STATE(137)] = 9664,
  [SMALL_STATE(138)] = 9702,
  [SMALL_STATE(139)] = 9740,
  [SMALL_STATE(140)] = 9794,
  [SMALL_STATE(141)] = 9847,
  [SMALL_STATE(142)] = 9900,
  [SMALL_STATE(143)] = 9933,
  [SMALL_STATE(144)] = 9986,
  [SMALL_STATE(145)] = 10034,
  [SMALL_STATE(146)] = 10082,
  [SMALL_STATE(147)] = 10130,
  [SMALL_STATE(148)] = 10196,
  [SMALL_STATE(149)] = 10244,
  [SMALL_STATE(150)] = 10276,
  [SMALL_STATE(151)] = 10342,
  [SMALL_STATE(152)] = 10371,
  [SMALL_STATE(153)] = 10398,
  [SMALL_STATE(154)] = 10424,
  [SMALL_STATE(155)] = 10450,
  [SMALL_STATE(156)] = 10476,
  [SMALL_STATE(157)] = 10502,
  [SMALL_STATE(158)] = 10528,
  [SMALL_STATE(159)] = 10554,
  [SMALL_STATE(160)] = 10580,
  [SMALL_STATE(161)] = 10606,
  [SMALL_STATE(162)] = 10632,
  [SMALL_STATE(163)] = 10658,
  [SMALL_STATE(164)] = 10684,
  [SMALL_STATE(165)] = 10710,
  [SMALL_STATE(166)] = 10736,
  [SMALL_STATE(167)] = 10762,
  [SMALL_STATE(168)] = 10788,
  [SMALL_STATE(169)] = 10835,
  [SMALL_STATE(170)] = 10859,
  [SMALL_STATE(171)] = 10889,
  [SMALL_STATE(172)] = 10917,
  [SMALL_STATE(173)] = 10949,
  [SMALL_STATE(174)] = 10981,
  [SMALL_STATE(175)] = 11011,
  [SMALL_STATE(176)] = 11041,
  [SMALL_STATE(177)] = 11079,
  [SMALL_STATE(178)] = 11109,
  [SMALL_STATE(179)] = 11147,
  [SMALL_STATE(180)] = 11177,
  [SMALL_STATE(181)] = 11202,
  [SMALL_STATE(182)] = 11227,
  [SMALL_STATE(183)] = 11252,
  [SMALL_STATE(184)] = 11279,
  [SMALL_STATE(185)] = 11304,
  [SMALL_STATE(186)] = 11329,
  [SMALL_STATE(187)] = 11356,
  [SMALL_STATE(188)] = 11381,
  [SMALL_STATE(189)] = 11408,
  [SMALL_STATE(190)] = 11432,
  [SMALL_STATE(191)] = 11452,
  [SMALL_STATE(192)] = 11471,
  [SMALL_STATE(193)] = 11490,
  [SMALL_STATE(194)] = 11509,
  [SMALL_STATE(195)] = 11528,
  [SMALL_STATE(196)] = 11547,
  [SMALL_STATE(197)] = 11566,
  [SMALL_STATE(198)] = 11585,
  [SMALL_STATE(199)] = 11604,
  [SMALL_STATE(200)] = 11623,
  [SMALL_STATE(201)] = 11642,
  [SMALL_STATE(202)] = 11661,
  [SMALL_STATE(203)] = 11680,
  [SMALL_STATE(204)] = 11699,
  [SMALL_STATE(205)] = 11718,
  [SMALL_STATE(206)] = 11737,
  [SMALL_STATE(207)] = 11756,
  [SMALL_STATE(208)] = 11775,
  [SMALL_STATE(209)] = 11794,
  [SMALL_STATE(210)] = 11813,
  [SMALL_STATE(211)] = 11832,
  [SMALL_STATE(212)] = 11851,
  [SMALL_STATE(213)] = 11870,
  [SMALL_STATE(214)] = 11889,
  [SMALL_STATE(215)] = 11908,
  [SMALL_STATE(216)] = 11927,
  [SMALL_STATE(217)] = 11946,
  [SMALL_STATE(218)] = 11965,
  [SMALL_STATE(219)] = 11986,
  [SMALL_STATE(220)] = 12005,
  [SMALL_STATE(221)] = 12024,
  [SMALL_STATE(222)] = 12043,
  [SMALL_STATE(223)] = 12062,
  [SMALL_STATE(224)] = 12081,
  [SMALL_STATE(225)] = 12100,
  [SMALL_STATE(226)] = 12121,
  [SMALL_STATE(227)] = 12140,
  [SMALL_STATE(228)] = 12159,
  [SMALL_STATE(229)] = 12178,
  [SMALL_STATE(230)] = 12197,
  [SMALL_STATE(231)] = 12215,
  [SMALL_STATE(232)] = 12233,
  [SMALL_STATE(233)] = 12251,
  [SMALL_STATE(234)] = 12269,
  [SMALL_STATE(235)] = 12287,
  [SMALL_STATE(236)] = 12319,
  [SMALL_STATE(237)] = 12351,
  [SMALL_STATE(238)] = 12379,
  [SMALL_STATE(239)] = 12397,
  [SMALL_STATE(240)] = 12415,
  [SMALL_STATE(241)] = 12433,
  [SMALL_STATE(242)] = 12451,
  [SMALL_STATE(243)] = 12469,
  [SMALL_STATE(244)] = 12487,
  [SMALL_STATE(245)] = 12505,
  [SMALL_STATE(246)] = 12523,
  [SMALL_STATE(247)] = 12555,
  [SMALL_STATE(248)] = 12583,
  [SMALL_STATE(249)] = 12601,
  [SMALL_STATE(250)] = 12633,
  [SMALL_STATE(251)] = 12661,
  [SMALL_STATE(252)] = 12679,
  [SMALL_STATE(253)] = 12697,
  [SMALL_STATE(254)] = 12715,
  [SMALL_STATE(255)] = 12747,
  [SMALL_STATE(256)] = 12765,
  [SMALL_STATE(257)] = 12783,
  [SMALL_STATE(258)] = 12801,
  [SMALL_STATE(259)] = 12819,
  [SMALL_STATE(260)] = 12837,
  [SMALL_STATE(261)] = 12855,
  [SMALL_STATE(262)] = 12873,
  [SMALL_STATE(263)] = 12891,
  [SMALL_STATE(264)] = 12911,
  [SMALL_STATE(265)] = 12929,
  [SMALL_STATE(266)] = 12947,
  [SMALL_STATE(267)] = 12965,
  [SMALL_STATE(268)] = 12983,
  [SMALL_STATE(269)] = 13007,
  [SMALL_STATE(270)] = 13033,
  [SMALL_STATE(271)] = 13051,
  [SMALL_STATE(272)] = 13083,
  [SMALL_STATE(273)] = 13101,
  [SMALL_STATE(274)] = 13121,
  [SMALL_STATE(275)] = 13145,
  [SMALL_STATE(276)] = 13171,
  [SMALL_STATE(277)] = 13189,
  [SMALL_STATE(278)] = 13221,
  [SMALL_STATE(279)] = 13239,
  [SMALL_STATE(280)] = 13271,
  [SMALL_STATE(281)] = 13289,
  [SMALL_STATE(282)] = 13307,
  [SMALL_STATE(283)] = 13325,
  [SMALL_STATE(284)] = 13343,
  [SMALL_STATE(285)] = 13371,
  [SMALL_STATE(286)] = 13389,
  [SMALL_STATE(287)] = 13407,
  [SMALL_STATE(288)] = 13425,
  [SMALL_STATE(289)] = 13443,
  [SMALL_STATE(290)] = 13475,
  [SMALL_STATE(291)] = 13493,
  [SMALL_STATE(292)] = 13511,
  [SMALL_STATE(293)] = 13529,
  [SMALL_STATE(294)] = 13558,
  [SMALL_STATE(295)] = 13587,
  [SMALL_STATE(296)] = 13616,
  [SMALL_STATE(297)] = 13645,
  [SMALL_STATE(298)] = 13667,
  [SMALL_STATE(299)] = 13689,
  [SMALL_STATE(300)] = 13708,
  [SMALL_STATE(301)] = 13725,
  [SMALL_STATE(302)] = 13744,
  [SMALL_STATE(303)] = 13763,
  [SMALL_STATE(304)] = 13782,
  [SMALL_STATE(305)] = 13801,
  [SMALL_STATE(306)] = 13820,
  [SMALL_STATE(307)] = 13837,
  [SMALL_STATE(308)] = 13854,
  [SMALL_STATE(309)] = 13873,
  [SMALL_STATE(310)] = 13892,
  [SMALL_STATE(311)] = 13911,
  [SMALL_STATE(312)] = 13930,
  [SMALL_STATE(313)] = 13947,
  [SMALL_STATE(314)] = 13966,
  [SMALL_STATE(315)] = 13985,
  [SMALL_STATE(316)] = 14008,
  [SMALL_STATE(317)] = 14025,
  [SMALL_STATE(318)] = 14044,
  [SMALL_STATE(319)] = 14063,
  [SMALL_STATE(320)] = 14081,
  [SMALL_STATE(321)] = 14101,
  [SMALL_STATE(322)] = 14121,
  [SMALL_STATE(323)] = 14141,
  [SMALL_STATE(324)] = 14161,
  [SMALL_STATE(325)] = 14175,
  [SMALL_STATE(326)] = 14195,
  [SMALL_STATE(327)] = 14215,
  [SMALL_STATE(328)] = 14237,
  [SMALL_STATE(329)] = 14249,
  [SMALL_STATE(330)] = 14269,
  [SMALL_STATE(331)] = 14289,
  [SMALL_STATE(332)] = 14309,
  [SMALL_STATE(333)] = 14329,
  [SMALL_STATE(334)] = 14349,
  [SMALL_STATE(335)] = 14369,
  [SMALL_STATE(336)] = 14389,
  [SMALL_STATE(337)] = 14409,
  [SMALL_STATE(338)] = 14429,
  [SMALL_STATE(339)] = 14443,
  [SMALL_STATE(340)] = 14463,
  [SMALL_STATE(341)] = 14483,
  [SMALL_STATE(342)] = 14497,
  [SMALL_STATE(343)] = 14517,
  [SMALL_STATE(344)] = 14537,
  [SMALL_STATE(345)] = 14557,
  [SMALL_STATE(346)] = 14568,
  [SMALL_STATE(347)] = 14579,
  [SMALL_STATE(348)] = 14590,
  [SMALL_STATE(349)] = 14603,
  [SMALL_STATE(350)] = 14622,
  [SMALL_STATE(351)] = 14639,
  [SMALL_STATE(352)] = 14656,
  [SMALL_STATE(353)] = 14673,
  [SMALL_STATE(354)] = 14690,
  [SMALL_STATE(355)] = 14703,
  [SMALL_STATE(356)] = 14714,
  [SMALL_STATE(357)] = 14725,
  [SMALL_STATE(358)] = 14736,
  [SMALL_STATE(359)] = 14747,
  [SMALL_STATE(360)] = 14758,
  [SMALL_STATE(361)] = 14769,
  [SMALL_STATE(362)] = 14786,
  [SMALL_STATE(363)] = 14797,
  [SMALL_STATE(364)] = 14814,
  [SMALL_STATE(365)] = 14825,
  [SMALL_STATE(366)] = 14842,
  [SMALL_STATE(367)] = 14856,
  [SMALL_STATE(368)] = 14870,
  [SMALL_STATE(369)] = 14886,
  [SMALL_STATE(370)] = 14900,
  [SMALL_STATE(371)] = 14914,
  [SMALL_STATE(372)] = 14928,
  [SMALL_STATE(373)] = 14942,
  [SMALL_STATE(374)] = 14958,
  [SMALL_STATE(375)] = 14970,
  [SMALL_STATE(376)] = 14984,
  [SMALL_STATE(377)] = 15000,
  [SMALL_STATE(378)] = 15014,
  [SMALL_STATE(379)] = 15026,
  [SMALL_STATE(380)] = 15040,
  [SMALL_STATE(381)] = 15054,
  [SMALL_STATE(382)] = 15068,
  [SMALL_STATE(383)] = 15082,
  [SMALL_STATE(384)] = 15098,
  [SMALL_STATE(385)] = 15114,
  [SMALL_STATE(386)] = 15128,
  [SMALL_STATE(387)] = 15142,
  [SMALL_STATE(388)] = 15156,
  [SMALL_STATE(389)] = 15170,
  [SMALL_STATE(390)] = 15184,
  [SMALL_STATE(391)] = 15197,
  [SMALL_STATE(392)] = 15210,
  [SMALL_STATE(393)] = 15223,
  [SMALL_STATE(394)] = 15236,
  [SMALL_STATE(395)] = 15249,
  [SMALL_STATE(396)] = 15262,
  [SMALL_STATE(397)] = 15275,
  [SMALL_STATE(398)] = 15288,
  [SMALL_STATE(399)] = 15301,
  [SMALL_STATE(400)] = 15314,
  [SMALL_STATE(401)] = 15327,
  [SMALL_STATE(402)] = 15340,
  [SMALL_STATE(403)] = 15349,
  [SMALL_STATE(404)] = 15362,
  [SMALL_STATE(405)] = 15375,
  [SMALL_STATE(406)] = 15388,
  [SMALL_STATE(407)] = 15401,
  [SMALL_STATE(408)] = 15410,
  [SMALL_STATE(409)] = 15419,
  [SMALL_STATE(410)] = 15428,
  [SMALL_STATE(411)] = 15441,
  [SMALL_STATE(412)] = 15454,
  [SMALL_STATE(413)] = 15463,
  [SMALL_STATE(414)] = 15476,
  [SMALL_STATE(415)] = 15489,
  [SMALL_STATE(416)] = 15502,
  [SMALL_STATE(417)] = 15515,
  [SMALL_STATE(418)] = 15528,
  [SMALL_STATE(419)] = 15541,
  [SMALL_STATE(420)] = 15554,
  [SMALL_STATE(421)] = 15567,
  [SMALL_STATE(422)] = 15580,
  [SMALL_STATE(423)] = 15593,
  [SMALL_STATE(424)] = 15606,
  [SMALL_STATE(425)] = 15619,
  [SMALL_STATE(426)] = 15632,
  [SMALL_STATE(427)] = 15645,
  [SMALL_STATE(428)] = 15658,
  [SMALL_STATE(429)] = 15671,
  [SMALL_STATE(430)] = 15684,
  [SMALL_STATE(431)] = 15697,
  [SMALL_STATE(432)] = 15710,
  [SMALL_STATE(433)] = 15720,
  [SMALL_STATE(434)] = 15730,
  [SMALL_STATE(435)] = 15740,
  [SMALL_STATE(436)] = 15748,
  [SMALL_STATE(437)] = 15758,
  [SMALL_STATE(438)] = 15768,
  [SMALL_STATE(439)] = 15778,
  [SMALL_STATE(440)] = 15788,
  [SMALL_STATE(441)] = 15798,
  [SMALL_STATE(442)] = 15808,
  [SMALL_STATE(443)] = 15816,
  [SMALL_STATE(444)] = 15826,
  [SMALL_STATE(445)] = 15836,
  [SMALL_STATE(446)] = 15844,
  [SMALL_STATE(447)] = 15852,
  [SMALL_STATE(448)] = 15860,
  [SMALL_STATE(449)] = 15870,
  [SMALL_STATE(450)] = 15880,
  [SMALL_STATE(451)] = 15890,
  [SMALL_STATE(452)] = 15900,
  [SMALL_STATE(453)] = 15908,
  [SMALL_STATE(454)] = 15918,
  [SMALL_STATE(455)] = 15928,
  [SMALL_STATE(456)] = 15938,
  [SMALL_STATE(457)] = 15948,
  [SMALL_STATE(458)] = 15956,
  [SMALL_STATE(459)] = 15966,
  [SMALL_STATE(460)] = 15976,
  [SMALL_STATE(461)] = 15986,
  [SMALL_STATE(462)] = 15996,
  [SMALL_STATE(463)] = 16003,
  [SMALL_STATE(464)] = 16010,
  [SMALL_STATE(465)] = 16017,
  [SMALL_STATE(466)] = 16024,
  [SMALL_STATE(467)] = 16031,
  [SMALL_STATE(468)] = 16038,
  [SMALL_STATE(469)] = 16045,
  [SMALL_STATE(470)] = 16052,
  [SMALL_STATE(471)] = 16059,
  [SMALL_STATE(472)] = 16066,
  [SMALL_STATE(473)] = 16073,
  [SMALL_STATE(474)] = 16080,
  [SMALL_STATE(475)] = 16087,
  [SMALL_STATE(476)] = 16094,
  [SMALL_STATE(477)] = 16101,
  [SMALL_STATE(478)] = 16108,
  [SMALL_STATE(479)] = 16115,
  [SMALL_STATE(480)] = 16122,
  [SMALL_STATE(481)] = 16129,
  [SMALL_STATE(482)] = 16136,
  [SMALL_STATE(483)] = 16143,
  [SMALL_STATE(484)] = 16150,
  [SMALL_STATE(485)] = 16157,
  [SMALL_STATE(486)] = 16164,
  [SMALL_STATE(487)] = 16171,
  [SMALL_STATE(488)] = 16178,
  [SMALL_STATE(489)] = 16185,
  [SMALL_STATE(490)] = 16192,
  [SMALL_STATE(491)] = 16199,
  [SMALL_STATE(492)] = 16206,
  [SMALL_STATE(493)] = 16213,
  [SMALL_STATE(494)] = 16220,
  [SMALL_STATE(495)] = 16227,
  [SMALL_STATE(496)] = 16234,
  [SMALL_STATE(497)] = 16241,
  [SMALL_STATE(498)] = 16248,
  [SMALL_STATE(499)] = 16255,
  [SMALL_STATE(500)] = 16262,
  [SMALL_STATE(501)] = 16269,
  [SMALL_STATE(502)] = 16276,
  [SMALL_STATE(503)] = 16283,
  [SMALL_STATE(504)] = 16290,
  [SMALL_STATE(505)] = 16297,
  [SMALL_STATE(506)] = 16304,
  [SMALL_STATE(507)] = 16311,
  [SMALL_STATE(508)] = 16318,
  [SMALL_STATE(509)] = 16325,
  [SMALL_STATE(510)] = 16332,
  [SMALL_STATE(511)] = 16339,
  [SMALL_STATE(512)] = 16346,
  [SMALL_STATE(513)] = 16353,
  [SMALL_STATE(514)] = 16360,
  [SMALL_STATE(515)] = 16367,
  [SMALL_STATE(516)] = 16374,
  [SMALL_STATE(517)] = 16381,
  [SMALL_STATE(518)] = 16388,
  [SMALL_STATE(519)] = 16395,
  [SMALL_STATE(520)] = 16402,
  [SMALL_STATE(521)] = 16409,
  [SMALL_STATE(522)] = 16416,
  [SMALL_STATE(523)] = 16423,
  [SMALL_STATE(524)] = 16430,
  [SMALL_STATE(525)] = 16437,
  [SMALL_STATE(526)] = 16444,
  [SMALL_STATE(527)] = 16451,
  [SMALL_STATE(528)] = 16458,
  [SMALL_STATE(529)] = 16465,
  [SMALL_STATE(530)] = 16472,
  [SMALL_STATE(531)] = 16479,
  [SMALL_STATE(532)] = 16486,
  [SMALL_STATE(533)] = 16493,
  [SMALL_STATE(534)] = 16500,
  [SMALL_STATE(535)] = 16507,
  [SMALL_STATE(536)] = 16514,
  [SMALL_STATE(537)] = 16521,
  [SMALL_STATE(538)] = 16528,
  [SMALL_STATE(539)] = 16535,
  [SMALL_STATE(540)] = 16542,
  [SMALL_STATE(541)] = 16549,
  [SMALL_STATE(542)] = 16556,
  [SMALL_STATE(543)] = 16563,
  [SMALL_STATE(544)] = 16570,
  [SMALL_STATE(545)] = 16577,
  [SMALL_STATE(546)] = 16584,
  [SMALL_STATE(547)] = 16591,
  [SMALL_STATE(548)] = 16598,
  [SMALL_STATE(549)] = 16605,
  [SMALL_STATE(550)] = 16612,
  [SMALL_STATE(551)] = 16619,
  [SMALL_STATE(552)] = 16626,
  [SMALL_STATE(553)] = 16633,
  [SMALL_STATE(554)] = 16640,
  [SMALL_STATE(555)] = 16647,
  [SMALL_STATE(556)] = 16654,
  [SMALL_STATE(557)] = 16661,
  [SMALL_STATE(558)] = 16668,
  [SMALL_STATE(559)] = 16675,
  [SMALL_STATE(560)] = 16682,
  [SMALL_STATE(561)] = 16689,
  [SMALL_STATE(562)] = 16696,
  [SMALL_STATE(563)] = 16703,
  [SMALL_STATE(564)] = 16710,
  [SMALL_STATE(565)] = 16717,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = false}}, SHIFT_EXTRA(),
  [5] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 0, 0, 0),
  [7] = {.entry = {.count = 1, .reusable = false}}, SHIFT(307),
  [9] = {.entry = {.count = 1, .reusable = false}}, SHIFT(537),
  [11] = {.entry = {.count = 1, .reusable = false}}, SHIFT(297),
  [13] = {.entry = {.count = 1, .reusable = false}}, SHIFT(373),
  [15] = {.entry = {.count = 1, .reusable = false}}, SHIFT(376),
  [17] = {.entry = {.count = 1, .reusable = false}}, SHIFT(147),
  [19] = {.entry = {.count = 1, .reusable = false}}, SHIFT(460),
  [21] = {.entry = {.count = 1, .reusable = false}}, SHIFT(562),
  [23] = {.entry = {.count = 1, .reusable = true}}, SHIFT(500),
  [25] = {.entry = {.count = 1, .reusable = true}}, SHIFT(368),
  [27] = {.entry = {.count = 1, .reusable = true}}, SHIFT(414),
  [29] = {.entry = {.count = 1, .reusable = false}}, SHIFT(117),
  [31] = {.entry = {.count = 1, .reusable = true}}, SHIFT(46),
  [33] = {.entry = {.count = 1, .reusable = true}}, SHIFT(364),
  [35] = {.entry = {.count = 1, .reusable = true}}, SHIFT(524),
  [37] = {.entry = {.count = 1, .reusable = true}}, SHIFT(113),
  [39] = {.entry = {.count = 1, .reusable = false}}, SHIFT(40),
  [41] = {.entry = {.count = 1, .reusable = true}}, SHIFT(127),
  [43] = {.entry = {.count = 1, .reusable = true}}, SHIFT(22),
  [45] = {.entry = {.count = 1, .reusable = false}}, SHIFT(564),
  [47] = {.entry = {.count = 1, .reusable = true}}, SHIFT(123),
  [49] = {.entry = {.count = 1, .reusable = false}}, SHIFT(85),
  [51] = {.entry = {.count = 1, .reusable = false}}, SHIFT(308),
  [53] = {.entry = {.count = 1, .reusable = false}}, SHIFT(385),
  [55] = {.entry = {.count = 1, .reusable = true}}, SHIFT(309),
  [57] = {.entry = {.count = 1, .reusable = true}}, SHIFT(386),
  [59] = {.entry = {.count = 1, .reusable = false}}, SHIFT(329),
  [61] = {.entry = {.count = 1, .reusable = true}}, SHIFT(330),
  [63] = {.entry = {.count = 1, .reusable = true}}, SHIFT(423),
  [65] = {.entry = {.count = 1, .reusable = false}}, SHIFT(334),
  [67] = {.entry = {.count = 1, .reusable = true}}, SHIFT(335),
  [69] = {.entry = {.count = 1, .reusable = true}}, SHIFT(88),
  [71] = {.entry = {.count = 1, .reusable = true}}, SHIFT_EXTRA(),
  [73] = {.entry = {.count = 1, .reusable = true}}, SHIFT(357),
  [75] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_expression, 3, 0, 28),
  [77] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_expression, 3, 0, 28),
  [79] = {.entry = {.count = 1, .reusable = false}}, SHIFT(420),
  [81] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_expression, 4, 0, 41),
  [83] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_expression, 4, 0, 41),
  [85] = {.entry = {.count = 1, .reusable = true}}, SHIFT(359),
  [87] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_dependency_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(117),
  [90] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_dependency_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(46),
  [93] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_dependency_expression_repeat1, 2, 0, 0),
  [95] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_dependency_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(113),
  [98] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_dependency_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(40),
  [101] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_dependency_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(127),
  [104] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_dependency_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(22),
  [107] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_dependency_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(564),
  [110] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_dependency_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(123),
  [113] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_dependency_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(85),
  [116] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_dependency_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(308),
  [119] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_dependency_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(385),
  [122] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_dependency_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(309),
  [125] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_dependency_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(386),
  [128] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_dependency_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(329),
  [131] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_dependency_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(330),
  [134] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_dependency_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(423),
  [137] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_dependency_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(334),
  [140] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_dependency_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(335),
  [143] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_dependency_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(88),
  [146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(362),
  [148] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_if_expression_repeat1, 2, 0, 51),
  [150] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_if_expression_repeat1, 2, 0, 51),
  [152] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_if_expression_repeat1, 2, 0, 51), SHIFT_REPEAT(550),
  [155] = {.entry = {.count = 1, .reusable = false}}, SHIFT(225),
  [157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(64),
  [159] = {.entry = {.count = 1, .reusable = false}}, SHIFT(477),
  [161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(111),
  [163] = {.entry = {.count = 1, .reusable = false}}, SHIFT(35),
  [165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(125),
  [167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(16),
  [169] = {.entry = {.count = 1, .reusable = false}}, SHIFT(557),
  [171] = {.entry = {.count = 1, .reusable = false}}, SHIFT(243),
  [173] = {.entry = {.count = 1, .reusable = false}}, SHIFT(301),
  [175] = {.entry = {.count = 1, .reusable = false}}, SHIFT(379),
  [177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(302),
  [179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(380),
  [181] = {.entry = {.count = 1, .reusable = false}}, SHIFT(343),
  [183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(320),
  [185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(413),
  [187] = {.entry = {.count = 1, .reusable = false}}, SHIFT(333),
  [189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(336),
  [191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(282),
  [193] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_if_clause, 4, 0, 58),
  [195] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_if_clause, 4, 0, 58),
  [197] = {.entry = {.count = 1, .reusable = false}}, SHIFT(151),
  [199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(56),
  [201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(292),
  [203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(76),
  [205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(122),
  [207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(21),
  [209] = {.entry = {.count = 1, .reusable = false}}, SHIFT(563),
  [211] = {.entry = {.count = 1, .reusable = false}}, SHIFT(129),
  [213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(313),
  [215] = {.entry = {.count = 1, .reusable = false}}, SHIFT(366),
  [217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(314),
  [219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(389),
  [221] = {.entry = {.count = 1, .reusable = false}}, SHIFT(339),
  [223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(342),
  [225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(390),
  [227] = {.entry = {.count = 1, .reusable = false}}, SHIFT(325),
  [229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(326),
  [231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(156),
  [233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(239),
  [235] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_if_expression_repeat1, 1, 0, 39),
  [237] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_if_expression_repeat1, 1, 0, 39),
  [239] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__braced_expr, 3, 0, 48),
  [241] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__braced_expr, 3, 0, 48),
  [243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(164),
  [245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(104),
  [247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(157),
  [249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(93),
  [251] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_elements, 2, 0, 0),
  [253] = {.entry = {.count = 1, .reusable = false}}, SHIFT(186),
  [255] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_elements, 3, 0, 0),
  [257] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence, 2, 0, 0),
  [259] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence, 3, 0, 0),
  [261] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression_inner, 1, 0, 0),
  [263] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression_inner, 1, 0, 0),
  [265] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression_inner, 3, 0, 30),
  [267] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression_inner, 3, 0, 30),
  [269] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression_inner, 3, 0, 31),
  [271] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression_inner, 3, 0, 31),
  [273] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_regex_literal, 1, 0, 0),
  [275] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_regex_literal, 1, 0, 0),
  [277] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_expression, 4, 0, 40),
  [279] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_expression, 4, 0, 40),
  [281] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_expression, 5, 0, 50),
  [283] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_expression, 5, 0, 50),
  [285] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_clause, 2, 0, 49),
  [287] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_clause, 2, 0, 49),
  [289] = {.entry = {.count = 1, .reusable = false}}, SHIFT(218),
  [291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(91),
  [293] = {.entry = {.count = 1, .reusable = false}}, SHIFT(43),
  [295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(126),
  [297] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 1, 0, 0),
  [299] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 1, 0, 0),
  [301] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_format_string, 2, 0, 0),
  [303] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_format_string, 2, 0, 0),
  [305] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 2, 0, 0),
  [307] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 2, 0, 0),
  [309] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value, 1, 0, 0),
  [311] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value, 1, 0, 0),
  [313] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_external_command, 1, 0, 0),
  [315] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_external_command, 1, 0, 0),
  [317] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_format_string, 3, 0, 0),
  [319] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_format_string, 3, 0, 0),
  [321] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_not_expression, 2, 0, 0),
  [323] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_not_expression, 2, 0, 0),
  [325] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_literal, 2, 0, 0),
  [327] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_literal, 2, 0, 0),
  [329] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__backticked, 2, 0, 0),
  [331] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__backticked, 2, 0, 0),
  [333] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__indented_backticked, 2, 0, 0),
  [335] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__indented_backticked, 2, 0, 0),
  [337] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_value, 3, 0, 0),
  [339] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_value, 3, 0, 0),
  [341] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_list_literal, 3, 0, 29),
  [343] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_literal, 3, 0, 29),
  [345] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__backticked, 3, 0, 0),
  [347] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__backticked, 3, 0, 0),
  [349] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__indented_backticked, 3, 0, 0),
  [351] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__indented_backticked, 3, 0, 0),
  [353] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_call, 3, 0, 5),
  [355] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_call, 3, 0, 5),
  [357] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assert_expression, 4, 0, 42),
  [359] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assert_expression, 4, 0, 42),
  [361] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_call, 4, 0, 43),
  [363] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_call, 4, 0, 43),
  [365] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assert_expression, 6, 0, 55),
  [367] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assert_expression, 6, 0, 55),
  [369] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assert_expression, 7, 0, 55),
  [371] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assert_expression, 7, 0, 55),
  [373] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 3, 0, 0),
  [375] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 3, 0, 0),
  [377] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression, 1, 0, 0),
  [379] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression, 1, 0, 0),
  [381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(96),
  [383] = {.entry = {.count = 1, .reusable = false}}, SHIFT(96),
  [385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(97),
  [387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(74),
  [389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(98),
  [391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(99),
  [393] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression, 2, 0, 0),
  [395] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression, 2, 0, 0),
  [397] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 1, 0, 0),
  [399] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 2, 0, 0),
  [401] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0),
  [403] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(307),
  [406] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(537),
  [409] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(297),
  [412] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(373),
  [415] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(376),
  [418] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(147),
  [421] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(460),
  [424] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(562),
  [427] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(500),
  [430] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2, 0, 0), SHIFT_REPEAT(368),
  [433] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 3, 0, 0),
  [435] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module_path, 2, 0, 0),
  [437] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module_path, 2, 0, 0),
  [439] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_module_path_repeat1, 2, 0, 0),
  [441] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_module_path_repeat1, 2, 0, 0),
  [443] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_path_repeat1, 2, 0, 0), SHIFT_REPEAT(524),
  [446] = {.entry = {.count = 1, .reusable = false}}, SHIFT(457),
  [448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(457),
  [450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(54),
  [452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(205),
  [454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(146),
  [456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(363),
  [458] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module, 2, 0, 6),
  [460] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module, 2, 0, 6),
  [462] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module, 4, 0, 20),
  [464] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module, 4, 0, 20),
  [466] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module, 3, 0, 9),
  [468] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module, 3, 0, 9),
  [470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(226),
  [472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(145),
  [474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(193),
  [476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(148),
  [478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(219),
  [480] = {.entry = {.count = 1, .reusable = false}}, SHIFT(324),
  [482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(176),
  [484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(430),
  [486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(246),
  [488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(482),
  [490] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_recipe_body_repeat1, 2, 0, 0), SHIFT_REPEAT(457),
  [493] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_recipe_body_repeat1, 2, 0, 0), SHIFT_REPEAT(457),
  [496] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_recipe_body_repeat1, 2, 0, 0), SHIFT_REPEAT(54),
  [499] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_recipe_body_repeat1, 2, 0, 0),
  [501] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_recipe_body_repeat1, 2, 0, 0), SHIFT_REPEAT(148),
  [504] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_recipe_body_repeat1, 2, 0, 0), SHIFT_REPEAT(363),
  [507] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_starred_dependency_argument, 2, 0, 33),
  [509] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_starred_dependency_argument, 2, 0, 33),
  [511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(178),
  [513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(19),
  [515] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_recipe_body_repeat1, 2, 0, 0),
  [517] = {.entry = {.count = 1, .reusable = false}}, SHIFT(316),
  [519] = {.entry = {.count = 1, .reusable = false}}, SHIFT(535),
  [521] = {.entry = {.count = 1, .reusable = false}}, SHIFT(298),
  [523] = {.entry = {.count = 1, .reusable = false}}, SHIFT(383),
  [525] = {.entry = {.count = 1, .reusable = false}}, SHIFT(384),
  [527] = {.entry = {.count = 1, .reusable = false}}, SHIFT(150),
  [529] = {.entry = {.count = 1, .reusable = false}}, SHIFT(455),
  [531] = {.entry = {.count = 1, .reusable = false}}, SHIFT(462),
  [533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(77),
  [535] = {.entry = {.count = 1, .reusable = false}}, SHIFT(77),
  [537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(78),
  [539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(72),
  [541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(79),
  [543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(80),
  [545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(429),
  [547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(401),
  [549] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_alias, 4, 0, 14),
  [551] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_alias, 4, 0, 14),
  [553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(488),
  [555] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recipe, 2, 0, 0),
  [557] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_recipe, 2, 0, 0),
  [559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(139),
  [561] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_alias, 5, 0, 32),
  [563] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_alias, 5, 0, 32),
  [565] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_if_expression_repeat1, 2, 0, 51), SHIFT_REPEAT(558),
  [568] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_path_repeat1, 2, 0, 0), SHIFT_REPEAT(488),
  [571] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recipe, 3, 0, 0),
  [573] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_recipe, 3, 0, 0),
  [575] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_value, 1, 0, 0), REDUCE(sym_attribute_named_param, 1, 0, 5),
  [578] = {.entry = {.count = 1, .reusable = false}}, SHIFT(48),
  [580] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_if_expression_repeat1, 2, 0, 51), SHIFT_REPEAT(523),
  [583] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_function_definition_repeat1, 2, 0, 0),
  [585] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_function_definition_repeat1, 2, 0, 0), SHIFT_REPEAT(500),
  [588] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_function_definition_repeat1, 2, 0, 0),
  [590] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recipe, 4, 0, 0),
  [592] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_recipe, 4, 0, 0),
  [594] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_eager, 3, 0, 13),
  [596] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_eager, 3, 0, 13),
  [598] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recipe_body, 5, 0, 44),
  [600] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_recipe_body, 5, 0, 44),
  [602] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 7, 0, 52),
  [604] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 7, 0, 52),
  [606] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 8, 0, 56),
  [608] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 8, 0, 56),
  [610] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_export, 3, 0, 13),
  [612] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_export, 3, 0, 13),
  [614] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assignment, 2, 0, 7),
  [616] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assignment, 2, 0, 7),
  [618] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unexport, 3, 0, 6),
  [620] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unexport, 3, 0, 6),
  [622] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import, 3, 0, 0),
  [624] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_import, 3, 0, 0),
  [626] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module, 4, 0, 9),
  [628] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module, 4, 0, 9),
  [630] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__assignment, 4, 0, 19),
  [632] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__assignment, 4, 0, 19),
  [634] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recipe_body, 2, 0, 0),
  [636] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_recipe_body, 2, 0, 0),
  [638] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unexport, 4, 0, 9),
  [640] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unexport, 4, 0, 9),
  [642] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assignment, 1, 0, 2),
  [644] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assignment, 1, 0, 2),
  [646] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_setting, 4, 0, 21),
  [648] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_setting, 4, 0, 21),
  [650] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_setting, 5, 0, 23),
  [652] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_setting, 5, 0, 23),
  [654] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module, 3, 0, 6),
  [656] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module, 3, 0, 6),
  [658] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_setting, 3, 0, 10),
  [660] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_setting, 3, 0, 10),
  [662] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import, 2, 0, 0),
  [664] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_import, 2, 0, 0),
  [666] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_eager, 2, 0, 3),
  [668] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_eager, 2, 0, 3),
  [670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(20),
  [672] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recipe_body, 3, 0, 0),
  [674] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_recipe_body, 3, 0, 0),
  [676] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_module, 5, 0, 20),
  [678] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_module, 5, 0, 20),
  [680] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_export, 2, 0, 3),
  [682] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_export, 2, 0, 3),
  [684] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 6, 0, 38),
  [686] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 6, 0, 38),
  [688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(15),
  [690] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recipe_body, 4, 0, 44),
  [692] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_recipe_body, 4, 0, 44),
  [694] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_setting, 6, 0, 45),
  [696] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_setting, 6, 0, 45),
  [698] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_definition, 7, 0, 47),
  [700] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_definition, 7, 0, 47),
  [702] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import, 4, 0, 0),
  [704] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_import, 4, 0, 0),
  [706] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute, 9, 0, 57),
  [708] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 9, 0, 57),
  [710] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute, 4, 0, 0),
  [712] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 4, 0, 0),
  [714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(312),
  [716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(451),
  [718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(327),
  [720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(468),
  [722] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recipe_header, 4, 0, 11),
  [724] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recipe_header, 4, 0, 6),
  [726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(81),
  [728] = {.entry = {.count = 1, .reusable = false}}, SHIFT(81),
  [730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(82),
  [732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(73),
  [734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(83),
  [736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(84),
  [738] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recipe_header, 2, 0, 4),
  [740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(109),
  [742] = {.entry = {.count = 1, .reusable = false}}, SHIFT(109),
  [744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(110),
  [746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(71),
  [748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(112),
  [750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(75),
  [752] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recipe_header, 3, 0, 4),
  [754] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute, 5, 0, 24),
  [756] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 5, 0, 24),
  [758] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute, 8, 0, 53),
  [760] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 8, 0, 53),
  [762] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recipe_header, 3, 0, 11),
  [764] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute, 8, 0, 54),
  [766] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 8, 0, 54),
  [768] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recipe_header, 3, 0, 6),
  [770] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recipe_header, 2, 0, 5),
  [772] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute, 6, 0, 36),
  [774] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 6, 0, 36),
  [776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(324),
  [778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(68),
  [780] = {.entry = {.count = 1, .reusable = false}}, SHIFT(277),
  [782] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recipe_header, 3, 0, 5),
  [784] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute, 7, 0, 36),
  [786] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 7, 0, 36),
  [788] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute, 7, 0, 46),
  [790] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 7, 0, 46),
  [792] = {.entry = {.count = 1, .reusable = true}}, SHIFT(254),
  [794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(271),
  [796] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dependencies, 1, 0, 0),
  [798] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_dependencies_repeat1, 2, 0, 0), SHIFT_REPEAT(312),
  [801] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_dependencies_repeat1, 2, 0, 0), SHIFT_REPEAT(451),
  [804] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_dependencies_repeat1, 2, 0, 0), SHIFT_REPEAT(327),
  [807] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_dependencies_repeat1, 2, 0, 0), SHIFT_REPEAT(468),
  [810] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_dependencies_repeat1, 2, 0, 0),
  [812] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__name, 1, 0, 1), SHIFT(508),
  [815] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__name, 1, 0, 1),
  [817] = {.entry = {.count = 1, .reusable = false}}, SHIFT(503),
  [819] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__name, 1, 0, 1),
  [821] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__name, 1, 0, 1), SHIFT(496),
  [824] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__format_string_repeat1, 2, 0, 0), SHIFT_REPEAT(299),
  [827] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__format_string_repeat1, 2, 0, 0),
  [829] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__format_string_repeat1, 2, 0, 0), SHIFT_REPEAT(55),
  [832] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_module_path_repeat1, 2, 0, 0), SHIFT_REPEAT(565),
  [835] = {.entry = {.count = 1, .reusable = false}}, SHIFT(303),
  [837] = {.entry = {.count = 1, .reusable = false}}, SHIFT(244),
  [839] = {.entry = {.count = 1, .reusable = false}}, SHIFT(50),
  [841] = {.entry = {.count = 1, .reusable = false}}, SHIFT(304),
  [843] = {.entry = {.count = 1, .reusable = false}}, SHIFT(55),
  [845] = {.entry = {.count = 1, .reusable = false}}, SHIFT(318),
  [847] = {.entry = {.count = 1, .reusable = false}}, SHIFT(270),
  [849] = {.entry = {.count = 1, .reusable = false}}, SHIFT(299),
  [851] = {.entry = {.count = 1, .reusable = false}}, SHIFT(136),
  [853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(565),
  [855] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__name, 1, 0, 0),
  [857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(426),
  [859] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__name, 1, 0, 0),
  [861] = {.entry = {.count = 1, .reusable = false}}, SHIFT(310),
  [863] = {.entry = {.count = 1, .reusable = false}}, SHIFT(86),
  [865] = {.entry = {.count = 1, .reusable = false}}, SHIFT(311),
  [867] = {.entry = {.count = 1, .reusable = false}}, SHIFT(90),
  [869] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dependency, 1, 0, 5),
  [871] = {.entry = {.count = 1, .reusable = false}}, SHIFT(317),
  [873] = {.entry = {.count = 1, .reusable = false}}, SHIFT(128),
  [875] = {.entry = {.count = 1, .reusable = false}}, SHIFT(305),
  [877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(394),
  [879] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__format_string_repeat1, 2, 0, 0), SHIFT_REPEAT(318),
  [882] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__format_string_repeat1, 2, 0, 0), SHIFT_REPEAT(50),
  [885] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_parameters_repeat1, 2, 0, 0), SHIFT_REPEAT(324),
  [888] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_parameters_repeat1, 2, 0, 0),
  [890] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_parameters_repeat1, 2, 0, 0), SHIFT_REPEAT(482),
  [893] = {.entry = {.count = 1, .reusable = false}}, SHIFT(322),
  [895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(322),
  [897] = {.entry = {.count = 1, .reusable = false}}, SHIFT(69),
  [899] = {.entry = {.count = 1, .reusable = false}}, SHIFT(344),
  [901] = {.entry = {.count = 1, .reusable = true}}, SHIFT(344),
  [903] = {.entry = {.count = 1, .reusable = false}}, SHIFT(62),
  [905] = {.entry = {.count = 1, .reusable = false}}, SHIFT(323),
  [907] = {.entry = {.count = 1, .reusable = true}}, SHIFT(323),
  [909] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__format_raw_string_repeat1, 2, 0, 0), SHIFT_REPEAT(323),
  [912] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__format_raw_string_repeat1, 2, 0, 0), SHIFT_REPEAT(323),
  [915] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__format_raw_string_repeat1, 2, 0, 0),
  [917] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__format_raw_string_repeat1, 2, 0, 0), SHIFT_REPEAT(69),
  [920] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 1, 0, 5),
  [922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(121),
  [924] = {.entry = {.count = 1, .reusable = false}}, SHIFT(353),
  [926] = {.entry = {.count = 1, .reusable = false}}, SHIFT(154),
  [928] = {.entry = {.count = 1, .reusable = false}}, SHIFT(70),
  [930] = {.entry = {.count = 1, .reusable = false}}, SHIFT(352),
  [932] = {.entry = {.count = 1, .reusable = false}}, SHIFT(153),
  [934] = {.entry = {.count = 1, .reusable = false}}, SHIFT(63),
  [936] = {.entry = {.count = 1, .reusable = false}}, SHIFT(331),
  [938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(331),
  [940] = {.entry = {.count = 1, .reusable = false}}, SHIFT(332),
  [942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(332),
  [944] = {.entry = {.count = 1, .reusable = false}}, SHIFT(240),
  [946] = {.entry = {.count = 1, .reusable = false}}, SHIFT(94),
  [948] = {.entry = {.count = 1, .reusable = false}}, SHIFT(95),
  [950] = {.entry = {.count = 1, .reusable = false}}, SHIFT(241),
  [952] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 2, 0, 6),
  [954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(124),
  [956] = {.entry = {.count = 1, .reusable = false}}, SHIFT(337),
  [958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(337),
  [960] = {.entry = {.count = 1, .reusable = false}}, SHIFT(340),
  [962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(340),
  [964] = {.entry = {.count = 1, .reusable = false}}, SHIFT(321),
  [966] = {.entry = {.count = 1, .reusable = true}}, SHIFT(321),
  [968] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__format_raw_string_repeat1, 2, 0, 0), SHIFT_REPEAT(344),
  [971] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__format_raw_string_repeat1, 2, 0, 0), SHIFT_REPEAT(344),
  [974] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__format_raw_string_repeat1, 2, 0, 0), SHIFT_REPEAT(62),
  [977] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 4, 0, 22),
  [979] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_interpolation, 3, 0, 0),
  [981] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_parameters_repeat1, 1, 0, 0),
  [983] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameters, 1, 0, 0),
  [985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(30),
  [987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(559),
  [989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(473),
  [991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(58),
  [993] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_recipe_line_repeat1, 2, 0, 0), SHIFT_REPEAT(54),
  [996] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_recipe_line_repeat1, 2, 0, 0),
  [998] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_recipe_line_repeat1, 2, 0, 0), SHIFT_REPEAT(350),
  [1001] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_body_repeat1, 2, 0, 0), SHIFT_REPEAT(351),
  [1004] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_command_body_repeat1, 2, 0, 0),
  [1006] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_body_repeat1, 2, 0, 0), SHIFT_REPEAT(63),
  [1009] = {.entry = {.count = 1, .reusable = false}}, SHIFT(351),
  [1011] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_command_body, 1, 0, 0),
  [1013] = {.entry = {.count = 1, .reusable = false}}, SHIFT(361),
  [1015] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameters, 2, 0, 0),
  [1017] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dependency_expression, 3, 0, 6),
  [1019] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dependency, 1, 0, 0),
  [1021] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dependency_expression, 4, 0, 6),
  [1023] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 3, 0, 15),
  [1025] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_body_repeat1, 2, 0, 0), SHIFT_REPEAT(361),
  [1028] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_command_body_repeat1, 2, 0, 0), SHIFT_REPEAT(70),
  [1031] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dependency_expression, 5, 0, 34),
  [1033] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recipe_line, 1, 0, 0),
  [1035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(350),
  [1037] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dependency_expression, 4, 0, 34),
  [1039] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recipe_line, 2, 0, 0),
  [1041] = {.entry = {.count = 1, .reusable = false}}, SHIFT(131),
  [1043] = {.entry = {.count = 1, .reusable = false}}, SHIFT(369),
  [1045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(365),
  [1047] = {.entry = {.count = 1, .reusable = false}}, SHIFT(294),
  [1049] = {.entry = {.count = 1, .reusable = false}}, SHIFT(341),
  [1051] = {.entry = {.count = 1, .reusable = false}}, SHIFT(293),
  [1053] = {.entry = {.count = 1, .reusable = false}}, SHIFT(133),
  [1055] = {.entry = {.count = 1, .reusable = false}}, SHIFT(371),
  [1057] = {.entry = {.count = 1, .reusable = false}}, SHIFT(372),
  [1059] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__string_repeat1, 2, 0, 0),
  [1061] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__string_repeat1, 2, 0, 0), SHIFT_REPEAT(371),
  [1064] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__shell_expanded_string_indented_repeat1, 2, 0, 0),
  [1066] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__shell_expanded_string_indented_repeat1, 2, 0, 0), SHIFT_REPEAT(372),
  [1069] = {.entry = {.count = 1, .reusable = false}}, SHIFT(491),
  [1071] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_interpolation, 3, 0, 0),
  [1073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(24),
  [1075] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attribute_repeat2, 2, 0, 0),
  [1077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(53),
  [1079] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_list_elements_repeat1, 2, 0, 0),
  [1081] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_list_elements_repeat1, 2, 0, 0), SHIFT_REPEAT(51),
  [1084] = {.entry = {.count = 1, .reusable = false}}, SHIFT(245),
  [1086] = {.entry = {.count = 1, .reusable = false}}, SHIFT(381),
  [1088] = {.entry = {.count = 1, .reusable = false}}, SHIFT(382),
  [1090] = {.entry = {.count = 1, .reusable = false}}, SHIFT(272),
  [1092] = {.entry = {.count = 1, .reusable = false}}, SHIFT(87),
  [1094] = {.entry = {.count = 1, .reusable = false}}, SHIFT(387),
  [1096] = {.entry = {.count = 1, .reusable = false}}, SHIFT(388),
  [1098] = {.entry = {.count = 1, .reusable = false}}, SHIFT(114),
  [1100] = {.entry = {.count = 1, .reusable = false}}, SHIFT(370),
  [1102] = {.entry = {.count = 1, .reusable = false}}, SHIFT(424),
  [1104] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__raw_string_indented_repeat1, 2, 0, 0),
  [1106] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__raw_string_indented_repeat1, 2, 0, 0), SHIFT_REPEAT(391),
  [1109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(405),
  [1111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(37),
  [1113] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_function_parameters_repeat1, 2, 0, 27),
  [1115] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_function_parameters_repeat1, 2, 0, 27), SHIFT_REPEAT(522),
  [1118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(428),
  [1120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(546),
  [1122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(532),
  [1124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(478),
  [1126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(480),
  [1128] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attribute_repeat1, 2, 0, 0),
  [1130] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_repeat1, 2, 0, 0), SHIFT_REPEAT(37),
  [1133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(445),
  [1135] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_attribute_repeat2, 2, 0, 16), SHIFT_REPEAT(559),
  [1138] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attribute_repeat2, 2, 0, 16),
  [1140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(27),
  [1142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(57),
  [1144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(467),
  [1146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(410),
  [1148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(28),
  [1150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(501),
  [1152] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_parameters, 2, 0, 18),
  [1154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(444),
  [1156] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym__shebang_with_lang_repeat1, 2, 0, 0),
  [1158] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__shebang_with_lang_repeat1, 2, 0, 0), SHIFT_REPEAT(410),
  [1161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(397),
  [1163] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_condition, 1, 0, 0),
  [1165] = {.entry = {.count = 1, .reusable = false}}, SHIFT(416),
  [1167] = {.entry = {.count = 1, .reusable = false}}, SHIFT(440),
  [1169] = {.entry = {.count = 1, .reusable = false}}, SHIFT(560),
  [1171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(520),
  [1173] = {.entry = {.count = 1, .reusable = false}}, SHIFT(391),
  [1175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(447),
  [1177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(499),
  [1179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(23),
  [1181] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_list_elements, 1, 0, 0),
  [1183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(42),
  [1185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(59),
  [1187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(515),
  [1189] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence, 1, 0, 0),
  [1191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(26),
  [1193] = {.entry = {.count = 1, .reusable = false}}, SHIFT(427),
  [1195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(548),
  [1197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(403),
  [1199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(519),
  [1201] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_parameters, 1, 0, 12),
  [1203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(456),
  [1205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(45),
  [1207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(66),
  [1209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(25),
  [1211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(105),
  [1213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(67),
  [1215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(160),
  [1217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(60),
  [1219] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_function_parameters_repeat1, 2, 0, 26),
  [1221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(161),
  [1223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(502),
  [1225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(283),
  [1227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(47),
  [1229] = {.entry = {.count = 1, .reusable = false}}, SHIFT(517),
  [1231] = {.entry = {.count = 1, .reusable = false}}, SHIFT(425),
  [1233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(280),
  [1235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(543),
  [1237] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_named_param, 3, 0, 35),
  [1239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(12),
  [1241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(212),
  [1243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(435),
  [1245] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_parameters, 3, 0, 18),
  [1247] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attribute_repeat2, 5, 0, 36),
  [1249] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attribute_repeat2, 4, 0, 36),
  [1251] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_attribute_repeat2, 6, 0, 54),
  [1253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2),
  [1255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3),
  [1257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(107),
  [1259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(538),
  [1261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(143),
  [1263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(513),
  [1265] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_parameters, 2, 0, 12),
  [1267] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recipe_line_prefix, 1, 0, 0),
  [1269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(13),
  [1271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(208),
  [1273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(182),
  [1275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(140),
  [1277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(511),
  [1279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(180),
  [1281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(458),
  [1283] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recipe_header, 5, 0, 11),
  [1285] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recipe_header, 5, 0, 6),
  [1287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(52),
  [1289] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__shebang_with_lang, 4, 0, 25),
  [1291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(518),
  [1293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(450),
  [1295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(253),
  [1297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(223),
  [1299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(194),
  [1301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(461),
  [1303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(234),
  [1305] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recipe_header, 4, 0, 4),
  [1307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(289),
  [1309] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_boolean, 1, 0, 0),
  [1311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(252),
  [1313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(210),
  [1315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(261),
  [1317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(227),
  [1319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(338),
  [1321] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [1323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(49),
  [1325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(459),
  [1327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(190),
  [1329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(347),
  [1331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(163),
  [1333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(166),
  [1335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(165),
  [1337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(167),
  [1339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(255),
  [1341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(206),
  [1343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(159),
  [1345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(18),
  [1347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(251),
  [1349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(349),
  [1351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(290),
  [1353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(158),
  [1355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(235),
  [1357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(256),
  [1359] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_parameter, 2, 0, 8),
  [1361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(181),
  [1363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(200),
  [1365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(204),
  [1367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(236),
  [1369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(230),
  [1371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(141),
  [1373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(484),
  [1375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(291),
  [1377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(195),
  [1379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(556),
  [1381] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__shebang_with_lang, 5, 0, 37),
  [1383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(65),
  [1385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(231),
  [1387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(142),
  [1389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(374),
  [1391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(100),
  [1393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(228),
  [1395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(101),
  [1397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(102),
  [1399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(103),
  [1401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(209),
  [1403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(278),
  [1405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(106),
  [1407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(216),
  [1409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(487),
  [1411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(472),
  [1413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(108),
  [1415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(378),
  [1417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(407),
  [1419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(408),
  [1421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(409),
  [1423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(242),
  [1425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(465),
  [1427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(249),
  [1429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(61),
  [1431] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_recipe_header, 4, 0, 5),
  [1433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(466),
  [1435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(286),
  [1437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(152),
  [1439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(144),
  [1441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(132),
  [1443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(185),
  [1445] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__shebang_with_lang, 3, 0, 17),
  [1447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(36),
  [1449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(375),
  [1451] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_shebang, 2, 0, 0),
  [1453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(356),
  [1455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(443),
  [1457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(41),
  [1459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(44),
  [1461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(328),
};

enum ts_external_scanner_symbol_identifiers {
  ts_external_token__indent = 0,
  ts_external_token__dedent = 1,
  ts_external_token__newline = 2,
  ts_external_token_text = 3,
  ts_external_token_error_recovery = 4,
};

static const TSSymbol ts_external_scanner_symbol_map[EXTERNAL_TOKEN_COUNT] = {
  [ts_external_token__indent] = sym__indent,
  [ts_external_token__dedent] = sym__dedent,
  [ts_external_token__newline] = sym__newline,
  [ts_external_token_text] = sym_text,
  [ts_external_token_error_recovery] = sym_error_recovery,
};

static const bool ts_external_scanner_states[7][EXTERNAL_TOKEN_COUNT] = {
  [1] = {
    [ts_external_token__indent] = true,
    [ts_external_token__dedent] = true,
    [ts_external_token__newline] = true,
    [ts_external_token_text] = true,
    [ts_external_token_error_recovery] = true,
  },
  [2] = {
    [ts_external_token__dedent] = true,
    [ts_external_token__newline] = true,
    [ts_external_token_text] = true,
  },
  [3] = {
    [ts_external_token__newline] = true,
  },
  [4] = {
    [ts_external_token__indent] = true,
  },
  [5] = {
    [ts_external_token__newline] = true,
    [ts_external_token_text] = true,
  },
  [6] = {
    [ts_external_token_text] = true,
  },
};

#ifdef __cplusplus
extern "C" {
#endif
void *tree_sitter_just_external_scanner_create(void);
void tree_sitter_just_external_scanner_destroy(void *);
bool tree_sitter_just_external_scanner_scan(void *, TSLexer *, const bool *);
unsigned tree_sitter_just_external_scanner_serialize(void *, char *);
void tree_sitter_just_external_scanner_deserialize(void *, const char *, unsigned);

#ifdef TREE_SITTER_HIDE_SYMBOLS
#define TS_PUBLIC
#elif defined(_WIN32)
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_just(void) {
  static const TSLanguage language = {
    .abi_version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .supertype_count = SUPERTYPE_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = (const void*)ts_lex_modes,
    .lex_fn = ts_lex,
    .keyword_lex_fn = ts_lex_keywords,
    .keyword_capture_token = sym_identifier,
    .external_scanner = {
      &ts_external_scanner_states[0][0],
      ts_external_scanner_symbol_map,
      tree_sitter_just_external_scanner_create,
      tree_sitter_just_external_scanner_destroy,
      tree_sitter_just_external_scanner_scan,
      tree_sitter_just_external_scanner_serialize,
      tree_sitter_just_external_scanner_deserialize,
    },
    .primary_state_ids = ts_primary_state_ids,
    .name = "just",
    .max_reserved_word_set_size = 0,
    .metadata = {
      .major_version = 0,
      .minor_version = 1,
      .patch_version = 0,
    },
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
